/* Generated by re2c 2.2 on Mon Oct 17 19:16:17 2022 */
#line 1 "tokenizer.re"
#include <limits>

#include <lfortran/parser/parser_exception.h>
#include <lfortran/parser/tokenizer.h>
#include <lfortran/parser/parser.tab.hh>
#include <libasr/bigint.h>

namespace LFortran
{

void Tokenizer::set_string(const std::string &str)
{
    // The input string must be NULL terminated, otherwise the tokenizer will
    // not detect the end of string. After C++11, the std::string is guaranteed
    // to end with \0, but we check this here just in case.
    LFORTRAN_ASSERT(str[str.size()] == '\0');
    cur = (unsigned char *)(&str[0]);
    string_start = cur;
    cur_line = cur;
    line_num = 1;
}

template<int base>
bool adddgt(uint64_t &u, uint64_t d)
{
    if (u > (std::numeric_limits<uint64_t>::max() - d) / base) {
        return false;
    }
    u = u * base + d;
    return true;
}

bool lex_dec(const unsigned char *s, const unsigned char *e, uint64_t &u)
{
    for (u = 0; s < e; ++s) {
        if (!adddgt<10>(u, *s - 0x30u)) {
            return false;
        }
    }
    return true;
}

// Tokenizes integer of the kind 1234_ikind into `u` and `suffix`
// s ... the start of the integer
// e ... the character after the end
bool lex_int(const unsigned char *s, const unsigned char *e, uint64_t &u,
    Str &suffix)
{
    for (u = 0; s < e; ++s) {
        if (*s == '_') {
            s++;
            suffix.p = (char*) s;
            suffix.n = e-s;
            return true;
        } else if (!adddgt<10>(u, *s - 0x30u)) {
            return false;
        }
    }
    suffix.p = nullptr;
    suffix.n = 0;
    return true;
}

void lex_int_large(Allocator &al, const unsigned char *s,
    const unsigned char *e, BigInt::BigInt &u, Str &suffix)
{
    uint64_t ui;
    if (lex_int(s, e, ui, suffix)) {
        if (ui <= BigInt::MAX_SMALL_INT) {
            u.from_smallint(ui);
            return;
        }
    }
    const unsigned char *start = s;
    for (; s < e; ++s) {
        if (*s == '_') {
            s++;
            suffix.p = (char*) s;
            suffix.n = e-s;

            Str num;
            num.p = (char*)start;
            num.n = s-start-1;
            u.from_largeint(al, num);
            return;
        }
    }
    suffix.p = nullptr;
    suffix.n = 0;
    Str num;
    num.p = (char*)start;
    num.n = e-start;
    u.from_largeint(al, num);
}

uint64_t parse_int(const unsigned char *s)
{
    while (*s == ' ') s++;
    uint64_t u;
    for (u = 0; ; ++s) {
        if (*s >= '0' && *s <= '9') {
            if (!adddgt<10>(u, *s - 0x30u)) {
                return false;
            }
        } else {
            return u;
        }
    }
}

#define KW(x) token(yylval.string); RET(KW_##x);
#define RET(x) token_loc(loc); last_token=yytokentype::x; return yytokentype::x;
#define WARN_REL(x) add_rel_warning(diagnostics, fixed_form, yytokentype::TK_##x);

void Tokenizer::add_rel_warning(diag::Diagnostics &diagnostics, bool fixed_form, int rel_token) const {
    if (!fixed_form) {
        static const std::map<int, std::pair<std::string, std::string>> m = {
            {yytokentype::TK_EQ, {"==", ".eq."}},
            {yytokentype::TK_NE, {"/=", ".ne."}},
            {yytokentype::TK_LT, {"<",  ".lt."}},
            {yytokentype::TK_GT, {">",  ".gt."}},
            {yytokentype::TK_LE, {"<=", ".le."}},
            {yytokentype::TK_GE, {">=", ".ge."}},
        };
        const std::string rel_new = m.at(rel_token).first;
        const std::string rel_old = m.at(rel_token).second;
        Location loc;
        token_loc(loc);
        diagnostics.tokenizer_style_label(
            "Use '" + rel_new + "' instead of '" + rel_old + "'",
            {loc},
            "help: write this as '" + rel_new + "'");
    }
}

int Tokenizer::lex(Allocator &al, YYSTYPE &yylval, Location &loc, diag::Diagnostics &diagnostics)
{
    if (enddo_state == 1) {
        enddo_state = 2;
        KW(END_DO)
    } else if (enddo_state == 2) {
        enddo_insert_count--;
        if (enddo_insert_count > 0) {
            enddo_state = 1;
        } else {
            enddo_state = 0;
            token_loc(loc); line_num++; cur_line=cur;
            last_token = yytokentype::TK_NEWLINE;
        }
        return yytokentype::TK_NEWLINE;
    }
    for (;;) {
        tok = cur;

        /*
        Re2c has excellent documentation at:

        https://re2c.org/manual/manual_c.html

        The first paragraph there explains the basics:

        * If multiple rules match, the longest match takes precedence
        * If multiple rules match the same string, the earlier rule takes
          precedence
        * Default rule `*` should always be defined, it has the lowest priority
          regardless of its place and matches any code unit
        * We use the "Sentinel character" method for end of input:
            * The end of the input text is denoted with a null character \x00
            * Thus the null character cannot be part of the input otherwise
            * There is one rule to match \x00 to end the parser
            * No other rule is allowed to match \x00, otherwise the re2c block
              would parse past the end of the string and segfaults
            * A special case of the previous point are negated character
              ranges, such as [^"\x00], where one must include \x00 in it to
              ensure this rule does not match \x00 (all other rules simply do
              not mention \x00)
            * See the "Handling the end of input" section in the re2c
              documentation for more info

        The re2c block interacts with the rest of the code via just one pointer
        variable `cur`. On entering the re2c block, the `cur` variable must
        point to the first character of the token to be tokenized by the block.
        The re2c block below then executes on its own until a rule is matched:
        the action in {} is then executed. In that action `cur` points to the
        first character of the next token.

        Before the re2c block we save the current `cur` into `tok`, so that we
        can use `tok` and `cur` in the action in {} to extract the token that
        corresponds to the rule that got matched:

        * `tok` points to the first character of the token
        * `cur-1` points to the last character of the token
        * `cur` points to the first character of the next token
        * `cur-tok` is the length of the token

        In the action, we do one of:

        * call `continue` which executes another cycle in the for loop (which
          will parse the next token); we use this to skip a token
        * call `return` which returns from this function; we return a token
        * throw an exception (terminates the tokenizer)

        In the first two cases, `cur` points to first character of the next
        token, which becomes `tok` at the next iteration of the loop (either
        right away after `continue` or after the `lex` function is called again
        after `return`).

        See the manual for more details.
        */


        // These two variables are needed by the re2c block below internally,
        // initialization is not needed. One can think of them as local
        // variables of the re2c block.
        unsigned char *mar, *ctxmar;
        
#line 220 "tokenizer.cpp"
{
	unsigned char yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm[] = {
		/* table 1 .. 8: 0 */
		  0,   2,   2,   2,   2,   2,   2,   2, 
		  2,   3,   3,   3,   2,   3,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  3,   2,   2,   2,   2,   2,   2,   0, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		242, 242, 210, 210, 210, 210, 210, 210, 
		146, 146,   2,   2,   2,   2,   2,   2, 
		  2, 150, 150, 150, 150, 150, 150,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,   2,   2,   2,   2,  26, 
		  2, 150, 150, 150, 150, 150, 150,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		/* table 9 .. 13: 256 */
		  0, 192, 192, 192, 192, 192, 192, 192, 
		192, 200, 128, 200, 192, 200, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		200, 192,  64, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		240, 240, 240, 240, 240, 240, 240, 240, 
		240, 240, 192, 192, 192, 192, 192, 192, 
		192, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 192, 192, 192, 192, 192, 
		192, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
	};
	yych = *cur;
	if (yybm[256+yych] & 8) {
		goto yy6;
	}
	switch (yych) {
	case 0x00:	goto yy2;
	case '\t':
	case '\n':	goto yy9;
	case ' ':
	case '!':	goto yy11;
	case '"':	goto yy12;
	case '#':	goto yy13;
	case '%':	goto yy14;
	case '&':	goto yy16;
	case '\'':	goto yy17;
	case '(':	goto yy18;
	case ')':	goto yy20;
	case '*':	goto yy22;
	case '+':	goto yy24;
	case ',':	goto yy26;
	case '-':	goto yy28;
	case '.':	goto yy30;
	case '/':	goto yy31;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy33;
	case ':':	goto yy36;
	case ';':	goto yy38;
	case '<':	goto yy40;
	case '=':	goto yy42;
	case '>':	goto yy44;
	case 'A':
	case 'a':	goto yy46;
	case 'B':
	case 'b':	goto yy48;
	case 'C':
	case 'c':	goto yy49;
	case 'D':
	case 'd':	goto yy50;
	case 'E':
	case 'e':	goto yy51;
	case 'F':
	case 'f':	goto yy52;
	case 'G':
	case 'g':	goto yy53;
	case 'H':
	case 'J':
	case 'X':
	case 'Y':
	case '_':
	case 'h':
	case 'j':
	case 'x':
	case 'y':	goto yy54;
	case 'I':
	case 'i':	goto yy56;
	case 'K':
	case 'k':	goto yy57;
	case 'L':
	case 'l':	goto yy58;
	case 'M':
	case 'm':	goto yy59;
	case 'N':
	case 'n':	goto yy60;
	case 'O':
	case 'o':	goto yy61;
	case 'P':
	case 'p':	goto yy62;
	case 'Q':
	case 'q':	goto yy63;
	case 'R':
	case 'r':	goto yy64;
	case 'S':
	case 's':	goto yy65;
	case 'T':
	case 't':	goto yy66;
	case 'U':
	case 'u':	goto yy67;
	case 'V':
	case 'v':	goto yy68;
	case 'W':
	case 'w':	goto yy69;
	case 'Z':
	case 'z':	goto yy70;
	case '[':	goto yy71;
	case ']':	goto yy73;
	case '|':	goto yy75;
	default:	goto yy4;
	}
yy2:
	++cur;
#line 249 "tokenizer.re"
	{ RET(END_OF_FILE); }
#line 392 "tokenizer.cpp"
yy4:
	++cur;
yy5:
#line 240 "tokenizer.re"
	{ token_loc(loc);
                std::string t = token();
                throw parser_local::TokenizerError(diag::Diagnostic(
                    "Token '" + t + "' is not recognized",
                    diag::Level::Error, diag::Stage::Tokenizer, {
                        diag::Label("token not recognized", {loc})
                    })
                );
            }
#line 406 "tokenizer.cpp"
yy6:
	yych = *++cur;
	if (yybm[256+yych] & 8) {
		goto yy6;
	}
#line 250 "tokenizer.re"
	{ continue; }
#line 414 "tokenizer.cpp"
yy9:
	++cur;
#line 504 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    enddo_newline_process = false;
                    enddo_state = 1;
                    return yytokentype::TK_NEWLINE;
                } else {
                    enddo_newline_process = false;
                    enddo_insert_count = 0;
                    token_loc(loc); line_num++; cur_line=cur;
                    last_token = yytokentype::TK_NEWLINE;
                    return yytokentype::TK_NEWLINE;
                }
            }
#line 431 "tokenizer.cpp"
yy11:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy78;
yy12:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy83;
yy13:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy87;
yy14:
	++cur;
#line 537 "tokenizer.re"
	{ RET(TK_PERCENT) }
#line 451 "tokenizer.cpp"
yy16:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= 0x08) goto yy5;
		if (yych <= '\v') goto yy91;
		goto yy5;
	} else {
		if (yych <= '\r') goto yy91;
		if (yych <= 0x1F) goto yy5;
		if (yych <= '!') goto yy91;
		goto yy5;
	}
yy17:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy98;
yy18:
	yych = *++cur;
	if (yych == '/') {
		ctxmar = cur;
		goto yy101;
	}
#line 519 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 478 "tokenizer.cpp"
yy20:
	++cur;
#line 527 "tokenizer.re"
	{ RET(TK_RPAREN) }
#line 483 "tokenizer.cpp"
yy22:
	yych = *++cur;
	if (yych == '*') goto yy102;
#line 539 "tokenizer.re"
	{ RET(TK_STAR) }
#line 489 "tokenizer.cpp"
yy24:
	++cur;
#line 531 "tokenizer.re"
	{ RET(TK_PLUS) }
#line 494 "tokenizer.cpp"
yy26:
	++cur;
#line 538 "tokenizer.re"
	{ RET(TK_COMMA) }
#line 499 "tokenizer.cpp"
yy28:
	++cur;
#line 532 "tokenizer.re"
	{ RET(TK_MINUS) }
#line 504 "tokenizer.cpp"
yy30:
	yyaccept = 0;
	yych = *(mar = ++cur);
	switch (yych) {
	case '.':	goto yy104;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy106;
	case 'A':
	case 'a':	goto yy109;
	case 'B':
	case 'C':
	case 'D':
	case 'H':
	case 'I':
	case 'J':
	case 'K':
	case 'M':
	case 'P':
	case 'Q':
	case 'R':
	case 'S':
	case 'U':
	case 'V':
	case 'W':
	case 'Y':
	case 'Z':
	case 'b':
	case 'c':
	case 'd':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'm':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 'u':
	case 'v':
	case 'w':
	case 'y':
	case 'z':	goto yy110;
	case 'E':
	case 'e':	goto yy112;
	case 'F':
	case 'f':	goto yy113;
	case 'G':
	case 'g':	goto yy114;
	case 'L':
	case 'l':	goto yy115;
	case 'N':
	case 'n':	goto yy116;
	case 'O':
	case 'o':	goto yy117;
	case 'T':
	case 't':	goto yy118;
	case 'X':
	case 'x':	goto yy119;
	default:	goto yy5;
	}
yy31:
	yych = *++cur;
	if (yych <= '.') {
		if (yych == ')') goto yy120;
	} else {
		if (yych <= '/') goto yy122;
		if (yych == '=') goto yy124;
	}
#line 536 "tokenizer.re"
	{ RET(TK_SLASH) }
#line 584 "tokenizer.cpp"
yy33:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 16) {
		goto yy33;
	}
	if (yych <= ' ') {
		if (yych <= '\v') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych >= '\v') {
				ctxmar = cur;
				goto yy126;
			}
		} else {
			if (yych == '\r') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych >= ' ') {
				ctxmar = cur;
				goto yy126;
			}
		}
	} else {
		if (yych <= 'E') {
			if (yych == '.') {
				ctxmar = cur;
				goto yy128;
			}
			if (yych >= 'D') goto yy129;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy130;
			} else {
				if (yych <= 'c') goto yy35;
				if (yych <= 'e') goto yy129;
			}
		}
	}
yy35:
#line 624 "tokenizer.re"
	{
                lex_int_large(al, tok, cur,
                    yylval.int_suffix.int_n,
                    yylval.int_suffix.int_kind);
                RET(TK_INTEGER)
            }
#line 635 "tokenizer.cpp"
yy36:
	yych = *++cur;
	if (yych == ':') goto yy131;
#line 534 "tokenizer.re"
	{ RET(TK_COLON) }
#line 641 "tokenizer.cpp"
yy38:
	++cur;
#line 535 "tokenizer.re"
	{ RET(TK_SEMICOLON) }
#line 646 "tokenizer.cpp"
yy40:
	yych = *++cur;
	if (yych == '=') goto yy133;
#line 556 "tokenizer.re"
	{ RET(TK_LT) }
#line 652 "tokenizer.cpp"
yy42:
	yych = *++cur;
	if (yych <= '<') goto yy43;
	if (yych <= '=') goto yy135;
	if (yych <= '>') goto yy137;
yy43:
#line 533 "tokenizer.re"
	{ RET(TK_EQUAL) }
#line 661 "tokenizer.cpp"
yy44:
	yych = *++cur;
	if (yych == '=') goto yy139;
#line 562 "tokenizer.re"
	{ RET(TK_GT) }
#line 667 "tokenizer.cpp"
yy46:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'K') {
			if (yych == 'B') goto yy141;
			goto yy55;
		} else {
			if (yych <= 'L') goto yy142;
			if (yych <= 'R') goto yy55;
			goto yy143;
		}
	} else {
		if (yych <= 'k') {
			if (yych == 'b') goto yy141;
			goto yy55;
		} else {
			if (yych <= 'l') goto yy142;
			if (yych == 's') goto yy143;
			goto yy55;
		}
	}
yy47:
#line 666 "tokenizer.re"
	{ token(yylval.string); RET(TK_NAME) }
#line 692 "tokenizer.cpp"
yy48:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'I') {
		if (yych <= '\'') {
			if (yych == '"') goto yy146;
			if (yych <= '&') goto yy55;
			goto yy147;
		} else {
			if (yych == 'A') goto yy148;
			if (yych <= 'H') goto yy55;
			goto yy149;
		}
	} else {
		if (yych <= 'a') {
			if (yych == 'L') goto yy150;
			if (yych <= '`') goto yy55;
			goto yy148;
		} else {
			if (yych <= 'i') {
				if (yych <= 'h') goto yy55;
				goto yy149;
			} else {
				if (yych == 'l') goto yy150;
				goto yy55;
			}
		}
	}
yy49:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy151;
	case 'H':
	case 'h':	goto yy152;
	case 'L':
	case 'l':	goto yy153;
	case 'O':
	case 'o':	goto yy154;
	case 'R':
	case 'r':	goto yy155;
	case 'Y':
	case 'y':	goto yy156;
	default:	goto yy55;
	}
yy50:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych <= 'E') {
			if (yych == 'A') goto yy157;
			if (yych <= 'D') goto yy55;
			goto yy158;
		} else {
			if (yych == 'I') goto yy159;
			if (yych <= 'N') goto yy55;
			goto yy160;
		}
	} else {
		if (yych <= 'e') {
			if (yych == 'a') goto yy157;
			if (yych <= 'd') goto yy55;
			goto yy158;
		} else {
			if (yych <= 'i') {
				if (yych <= 'h') goto yy55;
				goto yy159;
			} else {
				if (yych == 'o') goto yy160;
				goto yy55;
			}
		}
	}
yy51:
	yych = *++cur;
	switch (yych) {
	case 'L':
	case 'l':	goto yy162;
	case 'N':
	case 'n':	goto yy163;
	case 'Q':
	case 'q':	goto yy164;
	case 'R':
	case 'r':	goto yy165;
	case 'V':
	case 'v':	goto yy166;
	case 'X':
	case 'x':	goto yy167;
	default:	goto yy55;
	}
yy52:
	yych = *++cur;
	switch (yych) {
	case 'I':
	case 'i':	goto yy168;
	case 'L':
	case 'l':	goto yy169;
	case 'O':
	case 'o':	goto yy170;
	case 'U':
	case 'u':	goto yy171;
	default:	goto yy55;
	}
yy53:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy172;
		if (yych <= 'N') goto yy55;
		goto yy173;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy172;
		} else {
			if (yych == 'o') goto yy173;
			goto yy55;
		}
	}
yy54:
	yych = *++cur;
yy55:
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
	goto yy47;
yy56:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'L') {
			if (yych == 'F') goto yy175;
			goto yy55;
		} else {
			if (yych <= 'M') goto yy177;
			if (yych <= 'N') goto yy178;
			if (yych <= 'R') goto yy55;
			goto yy180;
		}
	} else {
		if (yych <= 'm') {
			if (yych == 'f') goto yy175;
			if (yych <= 'l') goto yy55;
			goto yy177;
		} else {
			if (yych <= 'n') goto yy178;
			if (yych == 's') goto yy180;
			goto yy55;
		}
	}
yy57:
	yych = *++cur;
	if (yych == 'I') goto yy182;
	if (yych == 'i') goto yy182;
	goto yy55;
yy58:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy183;
		if (yych <= 'N') goto yy55;
		goto yy184;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy183;
		} else {
			if (yych == 'o') goto yy184;
			goto yy55;
		}
	}
yy59:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy185;
		if (yych <= 'N') goto yy55;
		goto yy186;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy185;
		} else {
			if (yych == 'o') goto yy186;
			goto yy55;
		}
	}
yy60:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'E') {
			if (yych == 'A') goto yy187;
			if (yych <= 'D') goto yy55;
			goto yy188;
		} else {
			if (yych == 'O') goto yy189;
			if (yych <= 'T') goto yy55;
			goto yy190;
		}
	} else {
		if (yych <= 'e') {
			if (yych == 'a') goto yy187;
			if (yych <= 'd') goto yy55;
			goto yy188;
		} else {
			if (yych <= 'o') {
				if (yych <= 'n') goto yy55;
				goto yy189;
			} else {
				if (yych == 'u') goto yy190;
				goto yy55;
			}
		}
	}
yy61:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'P') {
		if (yych <= '\'') {
			if (yych == '"') goto yy191;
			if (yych <= '&') goto yy55;
			goto yy192;
		} else {
			if (yych == 'N') goto yy193;
			if (yych <= 'O') goto yy55;
			goto yy194;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'U') goto yy195;
			if (yych <= 'm') goto yy55;
			goto yy193;
		} else {
			if (yych <= 'p') {
				if (yych <= 'o') goto yy55;
				goto yy194;
			} else {
				if (yych == 'u') goto yy195;
				goto yy55;
			}
		}
	}
yy62:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'O') {
			if (yych == 'A') goto yy196;
			if (yych <= 'N') goto yy55;
			goto yy197;
		} else {
			if (yych == 'R') goto yy198;
			if (yych <= 'T') goto yy55;
			goto yy199;
		}
	} else {
		if (yych <= 'o') {
			if (yych == 'a') goto yy196;
			if (yych <= 'n') goto yy55;
			goto yy197;
		} else {
			if (yych <= 'r') {
				if (yych <= 'q') goto yy55;
				goto yy198;
			} else {
				if (yych == 'u') goto yy199;
				goto yy55;
			}
		}
	}
yy63:
	yych = *++cur;
	if (yych == 'U') goto yy200;
	if (yych == 'u') goto yy200;
	goto yy55;
yy64:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy201;
		if (yych <= 'D') goto yy55;
		goto yy202;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy201;
		} else {
			if (yych == 'e') goto yy202;
			goto yy55;
		}
	}
yy65:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy203;
	case 'E':
	case 'e':	goto yy204;
	case 'H':
	case 'h':	goto yy205;
	case 'O':
	case 'o':	goto yy206;
	case 'T':
	case 't':	goto yy207;
	case 'U':
	case 'u':	goto yy208;
	case 'Y':
	case 'y':	goto yy209;
	default:	goto yy55;
	}
yy66:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy210;
	case 'E':
	case 'e':	goto yy211;
	case 'H':
	case 'h':	goto yy212;
	case 'O':
	case 'o':	goto yy213;
	case 'Y':
	case 'y':	goto yy215;
	default:	goto yy55;
	}
yy67:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'N') goto yy216;
		if (yych <= 'R') goto yy55;
		goto yy217;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy216;
		} else {
			if (yych == 's') goto yy217;
			goto yy55;
		}
	}
yy68:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy218;
		if (yych <= 'N') goto yy55;
		goto yy219;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy218;
		} else {
			if (yych == 'o') goto yy219;
			goto yy55;
		}
	}
yy69:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych <= 'G') {
			if (yych == 'A') goto yy220;
			goto yy55;
		} else {
			if (yych <= 'H') goto yy221;
			if (yych <= 'Q') goto yy55;
			goto yy222;
		}
	} else {
		if (yych <= 'g') {
			if (yych == 'a') goto yy220;
			goto yy55;
		} else {
			if (yych <= 'h') goto yy221;
			if (yych == 'r') goto yy222;
			goto yy55;
		}
	}
yy70:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy223;
	if (yych == '\'') goto yy224;
	goto yy55;
yy71:
	++cur;
yy72:
#line 528 "tokenizer.re"
	{ RET(TK_LBRACKET) }
#line 1074 "tokenizer.cpp"
yy73:
	++cur;
#line 529 "tokenizer.re"
	{ RET(TK_RBRACKET) }
#line 1079 "tokenizer.cpp"
yy75:
	++cur;
#line 540 "tokenizer.re"
	{ RET(TK_VBAR) }
#line 1084 "tokenizer.cpp"
yy77:
	yych = *++cur;
yy78:
	if (yybm[256+yych] & 64) {
		goto yy77;
	}
	if (yych >= 0x01) goto yy80;
yy79:
	cur = mar;
	if (yyaccept <= 8) {
		if (yyaccept <= 4) {
			if (yyaccept <= 2) {
				if (yyaccept <= 1) {
					if (yyaccept == 0) {
						goto yy5;
					} else {
						goto yy35;
					}
				} else {
					goto yy47;
				}
			} else {
				if (yyaccept == 3) {
					goto yy85;
				} else {
					goto yy94;
				}
			}
		} else {
			if (yyaccept <= 6) {
				if (yyaccept == 5) {
					goto yy100;
				} else {
					goto yy72;
				}
			} else {
				if (yyaccept == 7) {
					goto yy108;
				} else {
					goto yy161;
				}
			}
		}
	} else {
		if (yyaccept <= 13) {
			if (yyaccept <= 11) {
				if (yyaccept <= 10) {
					if (yyaccept == 9) {
						goto yy179;
					} else {
						goto yy294;
					}
				} else {
					goto yy763;
				}
			} else {
				if (yyaccept == 12) {
					goto yy912;
				} else {
					goto yy989;
				}
			}
		} else {
			if (yyaccept <= 15) {
				if (yyaccept == 14) {
					goto yy1091;
				} else {
					goto yy1181;
				}
			} else {
				if (yyaccept == 16) {
					goto yy1301;
				} else {
					goto yy1303;
				}
			}
		}
	}
yy80:
	++cur;
#line 642 "tokenizer.re"
	{
                line_num++; cur_line=cur;
                token(yylval.string);
                yylval.string.n--;
                token_loc(loc);
                if (last_token == yytokentype::TK_NEWLINE) {
                    return yytokentype::TK_COMMENT;
                } else {
                    last_token=yytokentype::TK_NEWLINE;
                    return yytokentype::TK_EOLCOMMENT;
                }
            }
#line 1178 "tokenizer.cpp"
yy82:
	yych = *++cur;
yy83:
	if (yybm[256+yych] & 128) {
		goto yy82;
	}
	if (yych <= 0x00) goto yy79;
	yyaccept = 3;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy82;
yy85:
#line 662 "tokenizer.re"
	{ token_str(yylval.string); RET(TK_STRING) }
#line 1192 "tokenizer.cpp"
yy86:
	yych = *++cur;
yy87:
	if (yych <= 0x00) goto yy79;
	if (yych != '\n') goto yy86;
	++cur;
#line 656 "tokenizer.re"
	{ line_num++; cur_line=cur; continue; }
#line 1201 "tokenizer.cpp"
yy90:
	yych = *++cur;
yy91:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy90;
		} else {
			if (yych <= '\n') goto yy92;
			if (yych <= '\v') goto yy90;
			goto yy79;
		}
	} else {
		if (yych <= 0x1F) {
			if (yych <= '\r') goto yy90;
			goto yy79;
		} else {
			if (yych <= ' ') goto yy90;
			if (yych <= '!') goto yy95;
			goto yy79;
		}
	}
yy92:
	yyaccept = 4;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 1) {
		goto yy92;
	}
	if (yych <= 0x1F) goto yy94;
	if (yych <= '!') goto yy95;
	if (yych == '&') goto yy225;
yy94:
#line 638 "tokenizer.re"
	{
                line_num++; cur_line=cur; continue;
            }
#line 1238 "tokenizer.cpp"
yy95:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '\n') goto yy92;
	goto yy95;
yy97:
	yych = *++cur;
yy98:
	if (yybm[0+yych] & 2) {
		goto yy97;
	}
	if (yych <= 0x00) goto yy79;
	yyaccept = 5;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy97;
yy100:
#line 663 "tokenizer.re"
	{ token_str(yylval.string); RET(TK_STRING) }
#line 1257 "tokenizer.cpp"
yy101:
	yyaccept = 6;
	yych = *(mar = ++cur);
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych == '\t') goto yy226;
			goto yy72;
		} else {
			if (yych == '\f') goto yy72;
			if (yych <= '\r') goto yy226;
			goto yy72;
		}
	} else {
		if (yych <= '.') {
			if (yych <= ' ') goto yy226;
			if (yych == ')') goto yy228;
			goto yy72;
		} else {
			if (yych <= '/') goto yy230;
			if (yych == '=') goto yy231;
			goto yy72;
		}
	}
yy102:
	++cur;
#line 545 "tokenizer.re"
	{ RET(TK_POW) }
#line 1285 "tokenizer.cpp"
yy104:
	++cur;
#line 543 "tokenizer.re"
	{ RET(TK_DBL_DOT) }
#line 1290 "tokenizer.cpp"
yy106:
	yyaccept = 7;
	yych = *(mar = ++cur);
yy107:
	if (yych <= 'E') {
		if (yych <= '/') goto yy108;
		if (yych <= '9') goto yy106;
		if (yych >= 'D') goto yy129;
	} else {
		if (yych <= '_') {
			if (yych >= '_') goto yy233;
		} else {
			if (yych <= 'c') goto yy108;
			if (yych <= 'e') goto yy129;
		}
	}
yy108:
#line 595 "tokenizer.re"
	{ token(yylval.string); RET(TK_REAL) }
#line 1310 "tokenizer.cpp"
yy109:
	yych = *++cur;
	if (yych == 'N') goto yy236;
	if (yych == 'n') goto yy236;
	goto yy111;
yy110:
	yych = *++cur;
yy111:
	if (yybm[0+yych] & 4) {
		goto yy110;
	}
	if (yych == '.') goto yy234;
	goto yy79;
yy112:
	yych = *++cur;
	if (yych == 'Q') goto yy237;
	if (yych == 'q') goto yy237;
	goto yy111;
yy113:
	yych = *++cur;
	if (yych == 'A') goto yy238;
	if (yych == 'a') goto yy238;
	goto yy111;
yy114:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy239;
		if (yych <= 'S') goto yy111;
		goto yy240;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy239;
		} else {
			if (yych == 't') goto yy240;
			goto yy111;
		}
	}
yy115:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy241;
		if (yych <= 'S') goto yy111;
		goto yy242;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy241;
		} else {
			if (yych == 't') goto yy242;
			goto yy111;
		}
	}
yy116:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy243;
		if (yych <= 'N') goto yy111;
		goto yy244;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy243;
		} else {
			if (yych == 'o') goto yy244;
			goto yy111;
		}
	}
yy117:
	yych = *++cur;
	if (yych == 'R') goto yy245;
	if (yych == 'r') goto yy245;
	goto yy111;
yy118:
	yych = *++cur;
	if (yych == 'R') goto yy246;
	if (yych == 'r') goto yy246;
	goto yy111;
yy119:
	yych = *++cur;
	if (yych == 'O') goto yy247;
	if (yych == 'o') goto yy247;
	goto yy111;
yy120:
	++cur;
#line 530 "tokenizer.re"
	{ RET(TK_RBRACKET_OLD) }
#line 1398 "tokenizer.cpp"
yy122:
	++cur;
#line 546 "tokenizer.re"
	{ RET(TK_CONCAT) }
#line 1403 "tokenizer.cpp"
yy124:
	++cur;
#line 553 "tokenizer.re"
	{ RET(TK_NE) }
#line 1408 "tokenizer.cpp"
yy126:
	yych = *++cur;
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych == '\t') goto yy126;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			if (yych <= '\r') goto yy126;
			goto yy79;
		}
	} else {
		if (yych <= '^') {
			if (yych <= ' ') goto yy126;
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy248;
			goto yy79;
		} else {
			if (yych == '`') goto yy79;
			if (yych <= 'z') goto yy248;
			goto yy79;
		}
	}
yy128:
	yyaccept = 7;
	yych = *(mar = ++cur);
	if (yych <= 'Z') {
		if (yych <= '@') goto yy107;
		if (yych <= 'C') goto yy251;
		if (yych <= 'E') goto yy253;
		goto yy251;
	} else {
		if (yych <= 'c') {
			if (yych <= '`') goto yy107;
			goto yy251;
		} else {
			if (yych <= 'e') goto yy253;
			if (yych <= 'z') goto yy251;
			goto yy107;
		}
	}
yy129:
	yych = *++cur;
	if (yych <= ',') {
		if (yych == '+') goto yy254;
		goto yy79;
	} else {
		if (yych <= '-') goto yy254;
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy255;
		goto yy79;
	}
yy130:
	yych = *++cur;
	if (yych <= '9') {
		if (yych <= '&') {
			if (yych == '"') goto yy82;
			goto yy79;
		} else {
			if (yych <= '\'') goto yy97;
			if (yych <= '/') goto yy79;
			goto yy257;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy259;
			goto yy79;
		} else {
			if (yych == '`') goto yy79;
			if (yych <= 'z') goto yy259;
			goto yy79;
		}
	}
yy131:
	++cur;
#line 544 "tokenizer.re"
	{ RET(TK_DBL_COLON) }
#line 1487 "tokenizer.cpp"
yy133:
	++cur;
#line 559 "tokenizer.re"
	{ RET(TK_LE) }
#line 1492 "tokenizer.cpp"
yy135:
	++cur;
#line 550 "tokenizer.re"
	{ RET(TK_EQ) }
#line 1497 "tokenizer.cpp"
yy137:
	++cur;
#line 547 "tokenizer.re"
	{ RET(TK_ARROW) }
#line 1502 "tokenizer.cpp"
yy139:
	++cur;
#line 565 "tokenizer.re"
	{ RET(TK_GE) }
#line 1507 "tokenizer.cpp"
yy141:
	yych = *++cur;
	if (yych == 'S') goto yy261;
	if (yych == 's') goto yy261;
	goto yy55;
yy142:
	yych = *++cur;
	if (yych == 'L') goto yy262;
	if (yych == 'l') goto yy262;
	goto yy55;
yy143:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'S') goto yy264;
		if (yych <= 'X') goto yy55;
		goto yy265;
	} else {
		if (yych <= 's') {
			if (yych <= 'r') goto yy55;
			goto yy264;
		} else {
			if (yych == 'y') goto yy265;
			goto yy55;
		}
	}
yy144:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych <= '&') {
		if (yych == '"') goto yy82;
		goto yy47;
	} else {
		if (yych <= '\'') goto yy97;
		if (yych == '_') goto yy144;
		goto yy47;
	}
yy146:
	yych = *++cur;
	if (yybm[0+yych] & 32) {
		goto yy266;
	}
	goto yy79;
yy147:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '1') goto yy268;
	goto yy79;
yy148:
	yych = *++cur;
	if (yych == 'C') goto yy270;
	if (yych == 'c') goto yy270;
	goto yy55;
yy149:
	yych = *++cur;
	if (yych == 'N') goto yy271;
	if (yych == 'n') goto yy271;
	goto yy55;
yy150:
	yych = *++cur;
	if (yych == 'O') goto yy272;
	if (yych == 'o') goto yy272;
	goto yy55;
yy151:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'L') goto yy273;
		if (yych <= 'R') goto yy55;
		goto yy274;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy273;
		} else {
			if (yych == 's') goto yy274;
			goto yy55;
		}
	}
yy152:
	yych = *++cur;
	if (yych == 'A') goto yy275;
	if (yych == 'a') goto yy275;
	goto yy55;
yy153:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy276;
		if (yych <= 'N') goto yy55;
		goto yy277;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy276;
		} else {
			if (yych == 'o') goto yy277;
			goto yy55;
		}
	}
yy154:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych <= 'D') {
			if (yych <= 'C') goto yy55;
			goto yy278;
		} else {
			if (yych <= 'L') goto yy55;
			if (yych <= 'M') goto yy279;
			goto yy280;
		}
	} else {
		if (yych <= 'l') {
			if (yych == 'd') goto yy278;
			goto yy55;
		} else {
			if (yych <= 'm') goto yy279;
			if (yych <= 'n') goto yy280;
			goto yy55;
		}
	}
yy155:
	yych = *++cur;
	if (yych == 'I') goto yy281;
	if (yych == 'i') goto yy281;
	goto yy55;
yy156:
	yych = *++cur;
	if (yych == 'C') goto yy282;
	if (yych == 'c') goto yy282;
	goto yy55;
yy157:
	yych = *++cur;
	if (yych == 'T') goto yy283;
	if (yych == 't') goto yy283;
	goto yy55;
yy158:
	yych = *++cur;
	if (yych <= 'F') {
		if (yych == 'A') goto yy284;
		if (yych <= 'E') goto yy55;
		goto yy285;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy284;
		} else {
			if (yych == 'f') goto yy285;
			goto yy55;
		}
	}
yy159:
	yych = *++cur;
	if (yych == 'M') goto yy286;
	if (yych == 'm') goto yy286;
	goto yy55;
yy160:
	yyaccept = 8;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= '@') {
		if (yych <= '\f') {
			if (yych <= '\t') {
				if (yych >= '\t') {
					ctxmar = cur;
					goto yy287;
				}
			} else {
				if (yych == '\v') {
					ctxmar = cur;
					goto yy287;
				}
			}
		} else {
			if (yych <= ' ') {
				if (yych <= '\r') {
					ctxmar = cur;
					goto yy287;
				}
				if (yych >= ' ') {
					ctxmar = cur;
					goto yy287;
				}
			} else {
				if (yych <= '/') goto yy161;
				if (yych <= '9') goto yy54;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'V') {
				if (yych == 'U') goto yy289;
				goto yy54;
			} else {
				if (yych <= 'W') goto yy290;
				if (yych <= 'Z') goto yy54;
			}
		} else {
			if (yych <= 'v') {
				if (yych == 'u') goto yy289;
				goto yy54;
			} else {
				if (yych <= 'w') goto yy290;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy161:
#line 292 "tokenizer.re"
	{ KW(DO) }
#line 1720 "tokenizer.cpp"
yy162:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'E') goto yy291;
		if (yych <= 'R') goto yy55;
		goto yy292;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy291;
		} else {
			if (yych == 's') goto yy292;
			goto yy55;
		}
	}
yy163:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'D') {
			if (yych <= 'C') goto yy55;
			goto yy293;
		} else {
			if (yych <= 'S') goto yy55;
			if (yych <= 'T') goto yy295;
			goto yy296;
		}
	} else {
		if (yych <= 's') {
			if (yych == 'd') goto yy293;
			goto yy55;
		} else {
			if (yych <= 't') goto yy295;
			if (yych <= 'u') goto yy296;
			goto yy55;
		}
	}
yy164:
	yych = *++cur;
	if (yych == 'U') goto yy297;
	if (yych == 'u') goto yy297;
	goto yy55;
yy165:
	yych = *++cur;
	if (yych == 'R') goto yy298;
	if (yych == 'r') goto yy298;
	goto yy55;
yy166:
	yych = *++cur;
	if (yych == 'E') goto yy299;
	if (yych == 'e') goto yy299;
	goto yy55;
yy167:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'I') goto yy300;
		if (yych <= 'S') goto yy55;
		goto yy301;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy300;
		} else {
			if (yych == 't') goto yy301;
			goto yy55;
		}
	}
yy168:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'L') goto yy302;
		if (yych <= 'M') goto yy55;
		goto yy303;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy302;
		} else {
			if (yych == 'n') goto yy303;
			goto yy55;
		}
	}
yy169:
	yych = *++cur;
	if (yych == 'U') goto yy304;
	if (yych == 'u') goto yy304;
	goto yy55;
yy170:
	yych = *++cur;
	if (yych == 'R') goto yy305;
	if (yych == 'r') goto yy305;
	goto yy55;
yy171:
	yych = *++cur;
	if (yych == 'N') goto yy306;
	if (yych == 'n') goto yy306;
	goto yy55;
yy172:
	yych = *++cur;
	if (yych == 'N') goto yy307;
	if (yych == 'n') goto yy307;
	goto yy55;
yy173:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy174;
			if (yych <= 'S') goto yy54;
			goto yy308;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy308;
			if (yych <= 'z') goto yy54;
		}
	}
yy174:
#line 409 "tokenizer.re"
	{ KW(GO) }
#line 1847 "tokenizer.cpp"
yy175:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 411 "tokenizer.re"
	{ KW(IF) }
#line 1856 "tokenizer.cpp"
yy177:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'A') goto yy309;
		if (yych <= 'O') goto yy55;
		goto yy310;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy309;
		} else {
			if (yych == 'p') goto yy310;
			goto yy55;
		}
	}
yy178:
	yyaccept = 9;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'P') {
		if (yych <= ' ') {
			if (yych <= '\v') {
				if (yych == '\t') goto yy311;
				if (yych >= '\v') goto yy311;
			} else {
				if (yych == '\r') goto yy311;
				if (yych >= ' ') goto yy311;
			}
		} else {
			if (yych <= 'B') {
				if (yych <= '/') goto yy179;
				if (yych <= '9') goto yy54;
				if (yych >= 'A') goto yy54;
			} else {
				if (yych <= 'C') goto yy313;
				if (yych == 'O') goto yy314;
				goto yy54;
			}
		}
	} else {
		if (yych <= 'c') {
			if (yych <= 'T') {
				if (yych <= 'Q') goto yy315;
				if (yych <= 'R') goto yy54;
				if (yych <= 'S') goto yy316;
				goto yy317;
			} else {
				if (yych <= 'Z') goto yy54;
				if (yych <= '`') goto yy179;
				if (yych <= 'b') goto yy54;
				goto yy313;
			}
		} else {
			if (yych <= 'q') {
				if (yych == 'o') goto yy314;
				if (yych <= 'p') goto yy54;
				goto yy315;
			} else {
				if (yych <= 's') {
					if (yych <= 'r') goto yy54;
					goto yy316;
				} else {
					if (yych <= 't') goto yy317;
					if (yych <= 'z') goto yy54;
				}
			}
		}
	}
yy179:
#line 416 "tokenizer.re"
	{ KW(IN) }
#line 1930 "tokenizer.cpp"
yy180:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 426 "tokenizer.re"
	{ KW(IS) }
#line 1939 "tokenizer.cpp"
yy182:
	yych = *++cur;
	if (yych == 'N') goto yy318;
	if (yych == 'n') goto yy318;
	goto yy55;
yy183:
	yych = *++cur;
	if (yych == 'N') goto yy319;
	if (yych == 'n') goto yy319;
	goto yy55;
yy184:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy321;
		if (yych <= 'F') goto yy55;
		goto yy322;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy321;
		} else {
			if (yych == 'g') goto yy322;
			goto yy55;
		}
	}
yy185:
	yych = *++cur;
	if (yych == 'M') goto yy323;
	if (yych == 'm') goto yy323;
	goto yy55;
yy186:
	yych = *++cur;
	if (yych <= 'L') {
		if (yych == 'D') goto yy324;
		if (yych <= 'K') goto yy55;
		goto yy325;
	} else {
		if (yych <= 'd') {
			if (yych <= 'c') goto yy55;
			goto yy324;
		} else {
			if (yych == 'l') goto yy325;
			goto yy55;
		}
	}
yy187:
	yych = *++cur;
	if (yych == 'M') goto yy326;
	if (yych == 'm') goto yy326;
	goto yy55;
yy188:
	yych = *++cur;
	if (yych == 'W') goto yy327;
	if (yych == 'w') goto yy327;
	goto yy55;
yy189:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'N') goto yy328;
		if (yych <= 'O') goto yy55;
		goto yy329;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy328;
		} else {
			if (yych == 'p') goto yy329;
			goto yy55;
		}
	}
yy190:
	yych = *++cur;
	if (yych == 'L') goto yy330;
	if (yych == 'l') goto yy330;
	goto yy55;
yy191:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy331;
	}
	goto yy79;
yy192:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '7') goto yy333;
	goto yy79;
yy193:
	yych = *++cur;
	if (yych == 'L') goto yy335;
	if (yych == 'l') goto yy335;
	goto yy55;
yy194:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy336;
		if (yych <= 'S') goto yy55;
		goto yy337;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy336;
		} else {
			if (yych == 't') goto yy337;
			goto yy55;
		}
	}
yy195:
	yych = *++cur;
	if (yych == 'T') goto yy338;
	if (yych == 't') goto yy338;
	goto yy55;
yy196:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'Q') goto yy55;
		if (yych <= 'R') goto yy340;
		goto yy341;
	} else {
		if (yych <= 'q') goto yy55;
		if (yych <= 'r') goto yy340;
		if (yych <= 's') goto yy341;
		goto yy55;
	}
yy197:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'I') goto yy342;
		if (yych <= 'R') goto yy55;
		goto yy343;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy342;
		} else {
			if (yych == 's') goto yy343;
			goto yy55;
		}
	}
yy198:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych <= 'H') {
			if (yych == 'E') goto yy344;
			goto yy55;
		} else {
			if (yych <= 'I') goto yy345;
			if (yych <= 'N') goto yy55;
			goto yy346;
		}
	} else {
		if (yych <= 'h') {
			if (yych == 'e') goto yy344;
			goto yy55;
		} else {
			if (yych <= 'i') goto yy345;
			if (yych == 'o') goto yy346;
			goto yy55;
		}
	}
yy199:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'B') goto yy347;
		if (yych <= 'Q') goto yy55;
		goto yy348;
	} else {
		if (yych <= 'b') {
			if (yych <= 'a') goto yy55;
			goto yy347;
		} else {
			if (yych == 'r') goto yy348;
			goto yy55;
		}
	}
yy200:
	yych = *++cur;
	if (yych == 'I') goto yy349;
	if (yych == 'i') goto yy349;
	goto yy55;
yy201:
	yych = *++cur;
	if (yych == 'N') goto yy350;
	if (yych == 'n') goto yy350;
	goto yy55;
yy202:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy351;
	case 'C':
	case 'c':	goto yy352;
	case 'D':
	case 'd':	goto yy353;
	case 'S':
	case 's':	goto yy354;
	case 'T':
	case 't':	goto yy355;
	case 'W':
	case 'w':	goto yy356;
	default:	goto yy55;
	}
yy203:
	yych = *++cur;
	if (yych == 'V') goto yy357;
	if (yych == 'v') goto yy357;
	goto yy55;
yy204:
	yych = *++cur;
	if (yych <= 'Q') {
		if (yych == 'L') goto yy358;
		if (yych <= 'P') goto yy55;
		goto yy359;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy358;
		} else {
			if (yych == 'q') goto yy359;
			goto yy55;
		}
	}
yy205:
	yych = *++cur;
	if (yych == 'A') goto yy360;
	if (yych == 'a') goto yy360;
	goto yy55;
yy206:
	yych = *++cur;
	if (yych == 'U') goto yy361;
	if (yych == 'u') goto yy361;
	goto yy55;
yy207:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy362;
		if (yych <= 'N') goto yy55;
		goto yy363;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy362;
		} else {
			if (yych == 'o') goto yy363;
			goto yy55;
		}
	}
yy208:
	yych = *++cur;
	if (yych == 'B') goto yy364;
	if (yych == 'b') goto yy364;
	goto yy55;
yy209:
	yych = *++cur;
	if (yych == 'N') goto yy365;
	if (yych == 'n') goto yy365;
	goto yy55;
yy210:
	yych = *++cur;
	if (yych == 'R') goto yy366;
	if (yych == 'r') goto yy366;
	goto yy55;
yy211:
	yych = *++cur;
	if (yych <= 'M') {
		if (yych == 'A') goto yy367;
		if (yych <= 'L') goto yy55;
		goto yy368;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy367;
		} else {
			if (yych == 'm') goto yy368;
			goto yy55;
		}
	}
yy212:
	yych = *++cur;
	if (yych == 'E') goto yy369;
	if (yych == 'e') goto yy369;
	goto yy55;
yy213:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 492 "tokenizer.re"
	{ KW(TO) }
#line 2229 "tokenizer.cpp"
yy215:
	yych = *++cur;
	if (yych == 'P') goto yy370;
	if (yych == 'p') goto yy370;
	goto yy55;
yy216:
	yych = *++cur;
	if (yych == 'F') goto yy371;
	if (yych == 'f') goto yy371;
	goto yy55;
yy217:
	yych = *++cur;
	if (yych == 'E') goto yy372;
	if (yych == 'e') goto yy372;
	goto yy55;
yy218:
	yych = *++cur;
	if (yych == 'L') goto yy374;
	if (yych == 'l') goto yy374;
	goto yy55;
yy219:
	yych = *++cur;
	if (yych == 'L') goto yy375;
	if (yych == 'l') goto yy375;
	goto yy55;
yy220:
	yych = *++cur;
	if (yych == 'I') goto yy376;
	if (yych == 'i') goto yy376;
	goto yy55;
yy221:
	yych = *++cur;
	if (yych <= 'I') {
		if (yych == 'E') goto yy377;
		if (yych <= 'H') goto yy55;
		goto yy378;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy377;
		} else {
			if (yych == 'i') goto yy378;
			goto yy55;
		}
	}
yy222:
	yych = *++cur;
	if (yych == 'I') goto yy379;
	if (yych == 'i') goto yy379;
	goto yy55;
yy223:
	yych = *++cur;
	if (yych == '"') goto yy79;
	goto yy381;
yy224:
	yych = *++cur;
	if (yych == '\'') goto yy79;
	goto yy383;
yy225:
	++cur;
	goto yy94;
yy226:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy226;
		} else {
			if (yych == '\v') goto yy226;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy226;
			if (yych <= 0x1F) goto yy79;
			goto yy226;
		} else {
			if (yych == ')') goto yy384;
			goto yy79;
		}
	}
yy228:
	++cur;
	cur -= 2;
#line 521 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 2316 "tokenizer.cpp"
yy230:
	yych = *++cur;
	if (yych == ')') goto yy388;
	goto yy387;
yy231:
	++cur;
	cur -= 2;
#line 520 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 2326 "tokenizer.cpp"
yy233:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy390;
		if (yych <= '@') goto yy79;
		goto yy392;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy392;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy392;
			goto yy79;
		}
	}
yy234:
	++cur;
#line 665 "tokenizer.re"
	{ token(yylval.string); RET(TK_DEF_OP) }
#line 2348 "tokenizer.cpp"
yy236:
	yych = *++cur;
	if (yych == 'D') goto yy394;
	if (yych == 'd') goto yy394;
	goto yy111;
yy237:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == '.') goto yy395;
		goto yy111;
	} else {
		if (yych <= 'V') goto yy397;
		if (yych == 'v') goto yy397;
		goto yy111;
	}
yy238:
	yych = *++cur;
	if (yych == 'L') goto yy398;
	if (yych == 'l') goto yy398;
	goto yy111;
yy239:
	yych = *++cur;
	if (yych == '.') goto yy399;
	goto yy111;
yy240:
	yych = *++cur;
	if (yych == '.') goto yy401;
	goto yy111;
yy241:
	yych = *++cur;
	if (yych == '.') goto yy403;
	goto yy111;
yy242:
	yych = *++cur;
	if (yych == '.') goto yy405;
	goto yy111;
yy243:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == '.') goto yy407;
		goto yy111;
	} else {
		if (yych <= 'Q') goto yy409;
		if (yych == 'q') goto yy409;
		goto yy111;
	}
yy244:
	yych = *++cur;
	if (yych == 'T') goto yy410;
	if (yych == 't') goto yy410;
	goto yy111;
yy245:
	yych = *++cur;
	if (yych == '.') goto yy411;
	goto yy111;
yy246:
	yych = *++cur;
	if (yych == 'U') goto yy413;
	if (yych == 'u') goto yy413;
	goto yy111;
yy247:
	yych = *++cur;
	if (yych == 'R') goto yy414;
	if (yych == 'r') goto yy414;
	goto yy111;
yy248:
	yych = *++cur;
	if (yybm[0+yych] & 16) {
		goto yy248;
	}
	cur = ctxmar;
#line 596 "tokenizer.re"
	{
                if (last_token == yytokentype::TK_NEWLINE) {
                    uint64_t u;
                    if (lex_int(tok, cur, u, yylval.int_suffix.int_kind)) {
                            yylval.n = u;
                            if (enddo_label_stack[enddo_label_stack.size()-1] == u) {
                                while (enddo_label_stack[enddo_label_stack.size()-1] == u) {
                                    enddo_label_stack.pop_back();
                                    enddo_insert_count++;
                                }
                                enddo_newline_process = true;
                            } else {
                                enddo_newline_process = false;
                            }
                            RET(TK_LABEL)
                    } else {
                        token_loc(loc);
                        std::string t = token();
                        throw LFortran::parser_local::TokenizerError("Integer '" + t + "' too large",
                            loc);
                    }
                } else {
                    lex_int_large(al, tok, cur,
                        yylval.int_suffix.int_n,
                        yylval.int_suffix.int_kind);
                    RET(TK_INTEGER)
                }
            }
#line 2449 "tokenizer.cpp"
yy251:
	yych = *++cur;
yy252:
	if (yych <= '@') {
		if (yych == '.') goto yy415;
		goto yy79;
	} else {
		if (yych <= 'Z') goto yy251;
		if (yych <= '`') goto yy79;
		if (yych <= 'z') goto yy251;
		goto yy79;
	}
yy253:
	yych = *++cur;
	if (yych <= ',') {
		if (yych != '+') goto yy252;
	} else {
		if (yych <= '-') goto yy254;
		if (yych <= '/') goto yy252;
		if (yych <= '9') goto yy255;
		goto yy252;
	}
yy254:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych >= ':') goto yy79;
yy255:
	yyaccept = 7;
	yych = *(mar = ++cur);
	if (yych <= '/') goto yy108;
	if (yych <= '9') goto yy255;
	if (yych == '_') goto yy233;
	goto yy108;
yy257:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			goto yy35;
		} else {
			if (yych == '\f') goto yy35;
			ctxmar = cur;
			goto yy126;
		}
	} else {
		if (yych <= '-') {
			if (yych == ' ') {
				ctxmar = cur;
				goto yy126;
			}
			goto yy35;
		} else {
			if (yych <= '.') {
				ctxmar = cur;
				goto yy417;
			}
			if (yych <= '/') goto yy35;
			if (yych <= '9') goto yy257;
			goto yy35;
		}
	}
yy259:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= '-') {
		if (yych <= '\v') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych <= '\n') goto yy35;
			ctxmar = cur;
			goto yy126;
		} else {
			if (yych <= '\r') {
				if (yych <= '\f') goto yy35;
				ctxmar = cur;
				goto yy126;
			} else {
				if (yych == ' ') {
					ctxmar = cur;
					goto yy126;
				}
				goto yy35;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') {
				if (yych <= '.') {
					ctxmar = cur;
					goto yy417;
				}
				goto yy35;
			} else {
				if (yych <= '9') goto yy259;
				if (yych <= '@') goto yy35;
				goto yy259;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy35;
				goto yy259;
			} else {
				if (yych <= '`') goto yy35;
				if (yych <= 'z') goto yy259;
				goto yy35;
			}
		}
	}
yy261:
	yych = *++cur;
	if (yych == 'T') goto yy418;
	if (yych == 't') goto yy418;
	goto yy55;
yy262:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'O') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy263;
			if (yych <= 'N') goto yy54;
			goto yy419;
		}
	} else {
		if (yych <= 'n') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'o') goto yy419;
			if (yych <= 'z') goto yy54;
		}
	}
yy263:
#line 254 "tokenizer.re"
	{ KW(ALL) }
#line 2594 "tokenizer.cpp"
yy264:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'I') goto yy420;
		if (yych <= 'N') goto yy55;
		goto yy421;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy420;
		} else {
			if (yych == 'o') goto yy421;
			goto yy55;
		}
	}
yy265:
	yych = *++cur;
	if (yych == 'N') goto yy422;
	if (yych == 'n') goto yy422;
	goto yy55;
yy266:
	yych = *++cur;
	if (yybm[0+yych] & 32) {
		goto yy266;
	}
	if (yych == '"') goto yy423;
	goto yy79;
yy268:
	yych = *++cur;
	if (yych == '\'') goto yy425;
	if (yych <= '/') goto yy79;
	if (yych <= '1') goto yy268;
	goto yy79;
yy270:
	yych = *++cur;
	if (yych == 'K') goto yy427;
	if (yych == 'k') goto yy427;
	goto yy55;
yy271:
	yych = *++cur;
	if (yych == 'D') goto yy428;
	if (yych == 'd') goto yy428;
	goto yy55;
yy272:
	yych = *++cur;
	if (yych == 'C') goto yy430;
	if (yych == 'c') goto yy430;
	goto yy55;
yy273:
	yych = *++cur;
	if (yych == 'L') goto yy431;
	if (yych == 'l') goto yy431;
	goto yy55;
yy274:
	yych = *++cur;
	if (yych == 'E') goto yy433;
	if (yych == 'e') goto yy433;
	goto yy55;
yy275:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy435;
		if (yych <= 'Q') goto yy55;
		goto yy436;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy435;
		} else {
			if (yych == 'r') goto yy436;
			goto yy55;
		}
	}
yy276:
	yych = *++cur;
	if (yych == 'S') goto yy437;
	if (yych == 's') goto yy437;
	goto yy55;
yy277:
	yych = *++cur;
	if (yych == 'S') goto yy438;
	if (yych == 's') goto yy438;
	goto yy55;
yy278:
	yych = *++cur;
	if (yych == 'I') goto yy439;
	if (yych == 'i') goto yy439;
	goto yy55;
yy279:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'M') goto yy440;
		if (yych <= 'O') goto yy55;
		goto yy441;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy440;
		} else {
			if (yych == 'p') goto yy441;
			goto yy55;
		}
	}
yy280:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'C') goto yy442;
		if (yych <= 'S') goto yy55;
		goto yy443;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy442;
		} else {
			if (yych == 't') goto yy443;
			goto yy55;
		}
	}
yy281:
	yych = *++cur;
	if (yych == 'T') goto yy444;
	if (yych == 't') goto yy444;
	goto yy55;
yy282:
	yych = *++cur;
	if (yych == 'L') goto yy445;
	if (yych == 'l') goto yy445;
	goto yy55;
yy283:
	yych = *++cur;
	if (yych == 'A') goto yy446;
	if (yych == 'a') goto yy446;
	goto yy55;
yy284:
	yych = *++cur;
	if (yych == 'L') goto yy448;
	if (yych == 'l') goto yy448;
	goto yy55;
yy285:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy449;
		if (yych <= 'D') goto yy55;
		goto yy450;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy449;
		} else {
			if (yych == 'e') goto yy450;
			goto yy55;
		}
	}
yy286:
	yych = *++cur;
	if (yych == 'E') goto yy451;
	if (yych == 'e') goto yy451;
	goto yy55;
yy287:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy287;
		} else {
			if (yych == '\v') goto yy287;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy287;
			if (yych <= 0x1F) goto yy79;
			goto yy287;
		} else {
			if (yych <= '/') goto yy79;
			if (yych <= '9') goto yy452;
			goto yy79;
		}
	}
yy289:
	yych = *++cur;
	if (yych == 'B') goto yy455;
	if (yych == 'b') goto yy455;
	goto yy55;
yy290:
	yych = *++cur;
	if (yych == 'H') goto yy456;
	if (yych == 'h') goto yy456;
	goto yy55;
yy291:
	yych = *++cur;
	if (yych == 'M') goto yy457;
	if (yych == 'm') goto yy457;
	goto yy55;
yy292:
	yych = *++cur;
	if (yych == 'E') goto yy458;
	if (yych == 'e') goto yy458;
	goto yy55;
yy293:
	yyaccept = 10;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':	goto yy460;
	case ' ':	goto yy462;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'G':
	case 'H':
	case 'J':
	case 'K':
	case 'L':
	case 'N':
	case 'O':
	case 'Q':
	case 'R':
	case 'U':
	case 'V':
	case 'X':
	case 'Y':
	case 'Z':
	case 'g':
	case 'h':
	case 'j':
	case 'k':
	case 'l':
	case 'n':
	case 'o':
	case 'q':
	case 'r':
	case 'u':
	case 'v':
	case 'x':
	case 'y':
	case 'z':	goto yy54;
	case 'A':
	case 'a':	goto yy463;
	case 'B':
	case 'b':	goto yy464;
	case 'C':
	case 'c':	goto yy465;
	case 'D':
	case 'd':	goto yy466;
	case 'E':
	case 'e':	goto yy467;
	case 'F':
	case 'f':	goto yy468;
	case 'I':
	case 'i':	goto yy469;
	case 'M':
	case 'm':	goto yy470;
	case 'P':
	case 'p':	goto yy471;
	case 'S':
	case 's':	goto yy472;
	case 'T':
	case 't':	goto yy473;
	case 'W':
	case 'w':	goto yy474;
	default:	goto yy294;
	}
yy294:
#line 302 "tokenizer.re"
	{ KW(END) }
#line 2872 "tokenizer.cpp"
yy295:
	yych = *++cur;
	if (yych == 'R') goto yy475;
	if (yych == 'r') goto yy475;
	goto yy55;
yy296:
	yych = *++cur;
	if (yych == 'M') goto yy476;
	if (yych == 'm') goto yy476;
	goto yy55;
yy297:
	yych = *++cur;
	if (yych == 'I') goto yy478;
	if (yych == 'i') goto yy478;
	goto yy55;
yy298:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'M') goto yy479;
		if (yych <= 'N') goto yy55;
		goto yy480;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy479;
		} else {
			if (yych == 'o') goto yy480;
			goto yy55;
		}
	}
yy299:
	yych = *++cur;
	if (yych == 'N') goto yy481;
	if (yych == 'n') goto yy481;
	goto yy55;
yy300:
	yych = *++cur;
	if (yych == 'T') goto yy482;
	if (yych == 't') goto yy482;
	goto yy55;
yy301:
	yych = *++cur;
	if (yych == 'E') goto yy484;
	if (yych == 'e') goto yy484;
	goto yy55;
yy302:
	yych = *++cur;
	if (yych == 'E') goto yy485;
	if (yych == 'e') goto yy485;
	goto yy55;
yy303:
	yych = *++cur;
	if (yych == 'A') goto yy487;
	if (yych == 'a') goto yy487;
	goto yy55;
yy304:
	yych = *++cur;
	if (yych == 'S') goto yy488;
	if (yych == 's') goto yy488;
	goto yy55;
yy305:
	yych = *++cur;
	if (yych <= 'M') {
		if (yych == 'A') goto yy489;
		if (yych <= 'L') goto yy55;
		goto yy490;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy489;
		} else {
			if (yych == 'm') goto yy490;
			goto yy55;
		}
	}
yy306:
	yych = *++cur;
	if (yych == 'C') goto yy492;
	if (yych == 'c') goto yy492;
	goto yy55;
yy307:
	yych = *++cur;
	if (yych == 'E') goto yy493;
	if (yych == 'e') goto yy493;
	goto yy55;
yy308:
	yych = *++cur;
	if (yych == 'O') goto yy494;
	if (yych == 'o') goto yy494;
	goto yy55;
yy309:
	yych = *++cur;
	if (yych == 'G') goto yy496;
	if (yych == 'g') goto yy496;
	goto yy55;
yy310:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'L') goto yy497;
			goto yy55;
		} else {
			if (yych <= 'O') goto yy498;
			if (yych <= 'T') goto yy55;
			goto yy499;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'l') goto yy497;
			goto yy55;
		} else {
			if (yych <= 'o') goto yy498;
			if (yych == 'u') goto yy499;
			goto yy55;
		}
	}
yy311:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy311;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy311;
		}
	} else {
		if (yych <= 'N') {
			if (yych == ' ') goto yy311;
			goto yy79;
		} else {
			if (yych <= 'O') goto yy500;
			if (yych == 'o') goto yy500;
			goto yy79;
		}
	}
yy313:
	yych = *++cur;
	if (yych == 'L') goto yy501;
	if (yych == 'l') goto yy501;
	goto yy55;
yy314:
	yych = *++cur;
	if (yych == 'U') goto yy502;
	if (yych == 'u') goto yy502;
	goto yy55;
yy315:
	yych = *++cur;
	if (yych == 'U') goto yy503;
	if (yych == 'u') goto yy503;
	goto yy55;
yy316:
	yych = *++cur;
	if (yych == 'T') goto yy504;
	if (yych == 't') goto yy504;
	goto yy55;
yy317:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'E') goto yy505;
		if (yych <= 'Q') goto yy55;
		goto yy506;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy505;
		} else {
			if (yych == 'r') goto yy506;
			goto yy55;
		}
	}
yy318:
	yych = *++cur;
	if (yych == 'D') goto yy507;
	if (yych == 'd') goto yy507;
	goto yy55;
yy319:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 428 "tokenizer.re"
	{ KW(LEN) }
#line 3057 "tokenizer.cpp"
yy321:
	yych = *++cur;
	if (yych == 'A') goto yy509;
	if (yych == 'a') goto yy509;
	goto yy55;
yy322:
	yych = *++cur;
	if (yych == 'I') goto yy510;
	if (yych == 'i') goto yy510;
	goto yy55;
yy323:
	yych = *++cur;
	if (yych == 'O') goto yy511;
	if (yych == 'o') goto yy511;
	goto yy55;
yy324:
	yych = *++cur;
	if (yych == 'U') goto yy512;
	if (yych == 'u') goto yy512;
	goto yy55;
yy325:
	yych = *++cur;
	if (yych == 'D') goto yy513;
	if (yych == 'd') goto yy513;
	goto yy55;
yy326:
	yych = *++cur;
	if (yych == 'E') goto yy515;
	if (yych == 'e') goto yy515;
	goto yy55;
yy327:
	yych = *++cur;
	if (yych == '_') goto yy517;
	goto yy55;
yy328:
	yych = *++cur;
	if (yych <= '^') {
		if (yych == 'E') goto yy518;
		goto yy55;
	} else {
		if (yych <= '_') goto yy520;
		if (yych == 'e') goto yy518;
		goto yy55;
	}
yy329:
	yych = *++cur;
	if (yych == 'A') goto yy521;
	if (yych == 'a') goto yy521;
	goto yy55;
yy330:
	yych = *++cur;
	if (yych == 'L') goto yy522;
	if (yych == 'l') goto yy522;
	goto yy55;
yy331:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy331;
	}
	if (yych == '"') goto yy523;
	goto yy79;
yy333:
	yych = *++cur;
	if (yych == '\'') goto yy525;
	if (yych <= '/') goto yy79;
	if (yych <= '7') goto yy333;
	goto yy79;
yy335:
	yych = *++cur;
	if (yych == 'Y') goto yy527;
	if (yych == 'y') goto yy527;
	goto yy55;
yy336:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy529;
		if (yych <= 'Q') goto yy55;
		goto yy531;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy529;
		} else {
			if (yych == 'r') goto yy531;
			goto yy55;
		}
	}
yy337:
	yych = *++cur;
	if (yych == 'I') goto yy532;
	if (yych == 'i') goto yy532;
	goto yy55;
yy338:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 448 "tokenizer.re"
	{ KW(OUT) }
#line 3158 "tokenizer.cpp"
yy340:
	yych = *++cur;
	if (yych == 'A') goto yy533;
	if (yych == 'a') goto yy533;
	goto yy55;
yy341:
	yych = *++cur;
	if (yych == 'S') goto yy534;
	if (yych == 's') goto yy534;
	goto yy55;
yy342:
	yych = *++cur;
	if (yych == 'N') goto yy536;
	if (yych == 'n') goto yy536;
	goto yy55;
yy343:
	yych = *++cur;
	if (yych == 'T') goto yy537;
	if (yych == 't') goto yy537;
	goto yy55;
yy344:
	yych = *++cur;
	if (yych == 'C') goto yy539;
	if (yych == 'c') goto yy539;
	goto yy55;
yy345:
	yych = *++cur;
	if (yych <= 'V') {
		if (yych == 'N') goto yy540;
		if (yych <= 'U') goto yy55;
		goto yy541;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy540;
		} else {
			if (yych == 'v') goto yy541;
			goto yy55;
		}
	}
yy346:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych <= 'F') {
			if (yych == 'C') goto yy542;
			goto yy55;
		} else {
			if (yych <= 'G') goto yy543;
			if (yych <= 'S') goto yy55;
			goto yy544;
		}
	} else {
		if (yych <= 'f') {
			if (yych == 'c') goto yy542;
			goto yy55;
		} else {
			if (yych <= 'g') goto yy543;
			if (yych == 't') goto yy544;
			goto yy55;
		}
	}
yy347:
	yych = *++cur;
	if (yych == 'L') goto yy545;
	if (yych == 'l') goto yy545;
	goto yy55;
yy348:
	yych = *++cur;
	if (yych == 'E') goto yy546;
	if (yych == 'e') goto yy546;
	goto yy55;
yy349:
	yych = *++cur;
	if (yych == 'E') goto yy548;
	if (yych == 'e') goto yy548;
	goto yy55;
yy350:
	yych = *++cur;
	if (yych == 'K') goto yy549;
	if (yych == 'k') goto yy549;
	goto yy55;
yy351:
	yych = *++cur;
	if (yych <= 'L') {
		if (yych == 'D') goto yy551;
		if (yych <= 'K') goto yy55;
		goto yy553;
	} else {
		if (yych <= 'd') {
			if (yych <= 'c') goto yy55;
			goto yy551;
		} else {
			if (yych == 'l') goto yy553;
			goto yy55;
		}
	}
yy352:
	yych = *++cur;
	if (yych == 'U') goto yy555;
	if (yych == 'u') goto yy555;
	goto yy55;
yy353:
	yych = *++cur;
	if (yych == 'U') goto yy556;
	if (yych == 'u') goto yy556;
	goto yy55;
yy354:
	yych = *++cur;
	if (yych == 'U') goto yy557;
	if (yych == 'u') goto yy557;
	goto yy55;
yy355:
	yych = *++cur;
	if (yych == 'U') goto yy558;
	if (yych == 'u') goto yy558;
	goto yy55;
yy356:
	yych = *++cur;
	if (yych == 'I') goto yy559;
	if (yych == 'i') goto yy559;
	goto yy55;
yy357:
	yych = *++cur;
	if (yych == 'E') goto yy560;
	if (yych == 'e') goto yy560;
	goto yy55;
yy358:
	yych = *++cur;
	if (yych == 'E') goto yy562;
	if (yych == 'e') goto yy562;
	goto yy55;
yy359:
	yych = *++cur;
	if (yych == 'U') goto yy563;
	if (yych == 'u') goto yy563;
	goto yy55;
yy360:
	yych = *++cur;
	if (yych == 'R') goto yy564;
	if (yych == 'r') goto yy564;
	goto yy55;
yy361:
	yych = *++cur;
	if (yych == 'R') goto yy565;
	if (yych == 'r') goto yy565;
	goto yy55;
yy362:
	yych = *++cur;
	if (yych == 'T') goto yy566;
	if (yych == 't') goto yy566;
	goto yy55;
yy363:
	yych = *++cur;
	if (yych == 'P') goto yy568;
	if (yych == 'p') goto yy568;
	goto yy55;
yy364:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy570;
		if (yych <= 'Q') goto yy55;
		goto yy571;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy570;
		} else {
			if (yych == 'r') goto yy571;
			goto yy55;
		}
	}
yy365:
	yych = *++cur;
	if (yych == 'C') goto yy572;
	if (yych == 'c') goto yy572;
	goto yy55;
yy366:
	yych = *++cur;
	if (yych == 'G') goto yy574;
	if (yych == 'g') goto yy574;
	goto yy55;
yy367:
	yych = *++cur;
	if (yych == 'M') goto yy575;
	if (yych == 'm') goto yy575;
	goto yy55;
yy368:
	yych = *++cur;
	if (yych == 'P') goto yy577;
	if (yych == 'p') goto yy577;
	goto yy55;
yy369:
	yych = *++cur;
	if (yych == 'N') goto yy578;
	if (yych == 'n') goto yy578;
	goto yy55;
yy370:
	yych = *++cur;
	if (yych == 'E') goto yy580;
	if (yych == 'e') goto yy580;
	goto yy55;
yy371:
	yych = *++cur;
	if (yych == 'O') goto yy582;
	if (yych == 'o') goto yy582;
	goto yy55;
yy372:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 495 "tokenizer.re"
	{ KW(USE) }
#line 3373 "tokenizer.cpp"
yy374:
	yych = *++cur;
	if (yych == 'U') goto yy583;
	if (yych == 'u') goto yy583;
	goto yy55;
yy375:
	yych = *++cur;
	if (yych == 'A') goto yy584;
	if (yych == 'a') goto yy584;
	goto yy55;
yy376:
	yych = *++cur;
	if (yych == 'T') goto yy585;
	if (yych == 't') goto yy585;
	goto yy55;
yy377:
	yych = *++cur;
	if (yych == 'R') goto yy587;
	if (yych == 'r') goto yy587;
	goto yy55;
yy378:
	yych = *++cur;
	if (yych == 'L') goto yy588;
	if (yych == 'l') goto yy588;
	goto yy55;
yy379:
	yych = *++cur;
	if (yych == 'T') goto yy589;
	if (yych == 't') goto yy589;
	goto yy55;
yy380:
	yych = *++cur;
yy381:
	if (yybm[0+yych] & 128) {
		goto yy380;
	}
	if (yych == '"') goto yy590;
	goto yy79;
yy382:
	yych = *++cur;
yy383:
	if (yych <= '9') {
		if (yych == '\'') goto yy592;
		if (yych <= '/') goto yy79;
		goto yy382;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy79;
			goto yy382;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'f') goto yy382;
			goto yy79;
		}
	}
yy384:
	++cur;
	cur = ctxmar;
#line 523 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 3434 "tokenizer.cpp"
yy386:
	yych = *++cur;
yy387:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy386;
		} else {
			if (yych == '\v') goto yy386;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy386;
			if (yych <= 0x1F) goto yy79;
			goto yy386;
		} else {
			if (yych == ')') goto yy594;
			goto yy79;
		}
	}
yy388:
	++cur;
	cur -= 3;
#line 526 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 3461 "tokenizer.cpp"
yy390:
	yych = *++cur;
	if (yych <= '/') goto yy108;
	if (yych <= '9') goto yy390;
	goto yy108;
yy392:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy108;
		if (yych <= '9') goto yy392;
		if (yych <= '@') goto yy108;
		goto yy392;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy108;
			goto yy392;
		} else {
			if (yych <= '`') goto yy108;
			if (yych <= 'z') goto yy392;
			goto yy108;
		}
	}
yy394:
	yych = *++cur;
	if (yych == '.') goto yy596;
	goto yy111;
yy395:
	++cur;
#line 551 "tokenizer.re"
	{ WARN_REL(EQ) RET(TK_EQ) }
#line 3492 "tokenizer.cpp"
yy397:
	yych = *++cur;
	if (yych == '.') goto yy598;
	goto yy111;
yy398:
	yych = *++cur;
	if (yych == 'S') goto yy600;
	if (yych == 's') goto yy600;
	goto yy111;
yy399:
	++cur;
#line 566 "tokenizer.re"
	{ WARN_REL(GE) RET(TK_GE) }
#line 3506 "tokenizer.cpp"
yy401:
	++cur;
#line 563 "tokenizer.re"
	{ WARN_REL(GT) RET(TK_GT) }
#line 3511 "tokenizer.cpp"
yy403:
	++cur;
#line 560 "tokenizer.re"
	{ WARN_REL(LE) RET(TK_LE) }
#line 3516 "tokenizer.cpp"
yy405:
	++cur;
#line 557 "tokenizer.re"
	{ WARN_REL(LT) RET(TK_LT) }
#line 3521 "tokenizer.cpp"
yy407:
	++cur;
#line 554 "tokenizer.re"
	{ WARN_REL(NE) RET(TK_NE) }
#line 3526 "tokenizer.cpp"
yy409:
	yych = *++cur;
	if (yych == 'V') goto yy601;
	if (yych == 'v') goto yy601;
	goto yy111;
yy410:
	yych = *++cur;
	if (yych == '.') goto yy602;
	goto yy111;
yy411:
	++cur;
#line 572 "tokenizer.re"
	{ RET(TK_OR) }
#line 3540 "tokenizer.cpp"
yy413:
	yych = *++cur;
	if (yych == 'E') goto yy604;
	if (yych == 'e') goto yy604;
	goto yy111;
yy414:
	yych = *++cur;
	if (yych == '.') goto yy605;
	goto yy111;
yy415:
	++cur;
	cur = ctxmar;
#line 587 "tokenizer.re"
	{
                lex_int_large(al, tok, cur,
                    yylval.int_suffix.int_n,
                    yylval.int_suffix.int_kind);
                RET(TK_INTEGER)
            }
#line 3560 "tokenizer.cpp"
yy417:
	yych = *++cur;
	if (yych == '.') goto yy79;
	goto yy252;
yy418:
	yych = *++cur;
	if (yych == 'R') goto yy607;
	if (yych == 'r') goto yy607;
	goto yy55;
yy419:
	yych = *++cur;
	if (yych == 'C') goto yy608;
	if (yych == 'c') goto yy608;
	goto yy55;
yy420:
	yych = *++cur;
	if (yych == 'G') goto yy609;
	if (yych == 'g') goto yy609;
	goto yy55;
yy421:
	yych = *++cur;
	if (yych == 'C') goto yy610;
	if (yych == 'c') goto yy610;
	goto yy55;
yy422:
	yych = *++cur;
	if (yych == 'C') goto yy611;
	if (yych == 'c') goto yy611;
	goto yy55;
yy423:
	++cur;
#line 631 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 3594 "tokenizer.cpp"
yy425:
	++cur;
#line 632 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 3599 "tokenizer.cpp"
yy427:
	yych = *++cur;
	if (yych == 'S') goto yy612;
	if (yych == 's') goto yy612;
	goto yy55;
yy428:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 262 "tokenizer.re"
	{ KW(BIND) }
#line 3613 "tokenizer.cpp"
yy430:
	yych = *++cur;
	if (yych == 'K') goto yy613;
	if (yych == 'k') goto yy613;
	goto yy55;
yy431:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 264 "tokenizer.re"
	{ KW(CALL) }
#line 3627 "tokenizer.cpp"
yy433:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 265 "tokenizer.re"
	{ KW(CASE) }
#line 3636 "tokenizer.cpp"
yy435:
	yych = *++cur;
	if (yych == 'G') goto yy615;
	if (yych == 'g') goto yy615;
	goto yy55;
yy436:
	yych = *++cur;
	if (yych == 'A') goto yy616;
	if (yych == 'a') goto yy616;
	goto yy55;
yy437:
	yych = *++cur;
	if (yych == 'S') goto yy617;
	if (yych == 's') goto yy617;
	goto yy55;
yy438:
	yych = *++cur;
	if (yych == 'E') goto yy619;
	if (yych == 'e') goto yy619;
	goto yy55;
yy439:
	yych = *++cur;
	if (yych == 'M') goto yy621;
	if (yych == 'm') goto yy621;
	goto yy55;
yy440:
	yych = *++cur;
	if (yych == 'O') goto yy622;
	if (yych == 'o') goto yy622;
	goto yy55;
yy441:
	yych = *++cur;
	if (yych == 'L') goto yy623;
	if (yych == 'l') goto yy623;
	goto yy55;
yy442:
	yych = *++cur;
	if (yych == 'U') goto yy624;
	if (yych == 'u') goto yy624;
	goto yy55;
yy443:
	yych = *++cur;
	if (yych <= 'I') {
		if (yych == 'A') goto yy625;
		if (yych <= 'H') goto yy55;
		goto yy626;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy625;
		} else {
			if (yych == 'i') goto yy626;
			goto yy55;
		}
	}
yy444:
	yych = *++cur;
	if (yych == 'I') goto yy627;
	if (yych == 'i') goto yy627;
	goto yy55;
yy445:
	yych = *++cur;
	if (yych == 'E') goto yy628;
	if (yych == 'e') goto yy628;
	goto yy55;
yy446:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 280 "tokenizer.re"
	{ KW(DATA) }
#line 3710 "tokenizer.cpp"
yy448:
	yych = *++cur;
	if (yych == 'L') goto yy630;
	if (yych == 'l') goto yy630;
	goto yy55;
yy449:
	yych = *++cur;
	if (yych == 'U') goto yy631;
	if (yych == 'u') goto yy631;
	goto yy55;
yy450:
	yych = *++cur;
	if (yych == 'R') goto yy632;
	if (yych == 'r') goto yy632;
	goto yy55;
yy451:
	yych = *++cur;
	if (yych == 'N') goto yy633;
	if (yych == 'n') goto yy633;
	goto yy55;
yy452:
	yych = *++cur;
	if (yych <= '/') goto yy454;
	if (yych <= '9') goto yy452;
yy454:
	cur = ctxmar;
#line 285 "tokenizer.re"
	{
                // This is a label do statement, we have to match the
                // corresponding continue base "end do".
                uint64_t n = parse_int(cur);
                enddo_label_stack.push_back(n);
                KW(DO);
            }
#line 3745 "tokenizer.cpp"
yy455:
	yych = *++cur;
	if (yych == 'L') goto yy634;
	if (yych == 'l') goto yy634;
	goto yy55;
yy456:
	yych = *++cur;
	if (yych == 'I') goto yy635;
	if (yych == 'i') goto yy635;
	goto yy55;
yy457:
	yych = *++cur;
	if (yych == 'E') goto yy636;
	if (yych == 'e') goto yy636;
	goto yy55;
yy458:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'W') {
		if (yych <= '@') {
			if (yych <= '/') goto yy459;
			if (yych <= '9') goto yy54;
		} else {
			if (yych == 'I') goto yy637;
			if (yych <= 'V') goto yy54;
			goto yy638;
		}
	} else {
		if (yych <= 'i') {
			if (yych <= 'Z') goto yy54;
			if (yych <= '`') goto yy459;
			if (yych <= 'h') goto yy54;
			goto yy637;
		} else {
			if (yych == 'w') goto yy638;
			if (yych <= 'z') goto yy54;
		}
	}
yy459:
#line 298 "tokenizer.re"
	{ KW(ELSE) }
#line 3789 "tokenizer.cpp"
yy460:
	yych = *++cur;
yy461:
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':	goto yy460;
	case 'A':
	case 'a':	goto yy639;
	case 'B':
	case 'b':	goto yy640;
	case 'C':
	case 'c':	goto yy641;
	case 'D':
	case 'd':	goto yy642;
	case 'E':
	case 'e':	goto yy643;
	case 'F':
	case 'f':	goto yy644;
	case 'I':
	case 'i':	goto yy645;
	case 'M':
	case 'm':	goto yy646;
	case 'P':
	case 'p':	goto yy647;
	case 'S':
	case 's':	goto yy648;
	case 'T':
	case 't':	goto yy649;
	case 'W':
	case 'w':	goto yy650;
	default:	goto yy79;
	}
yy462:
	yych = *++cur;
	if (yych == 'F') goto yy651;
	if (yych == 'f') goto yy651;
	goto yy461;
yy463:
	yych = *++cur;
	if (yych == 'S') goto yy652;
	if (yych == 's') goto yy652;
	goto yy55;
yy464:
	yych = *++cur;
	if (yych == 'L') goto yy653;
	if (yych == 'l') goto yy653;
	goto yy55;
yy465:
	yych = *++cur;
	if (yych == 'R') goto yy654;
	if (yych == 'r') goto yy654;
	goto yy55;
yy466:
	yych = *++cur;
	if (yych == 'O') goto yy655;
	if (yych == 'o') goto yy655;
	goto yy55;
yy467:
	yych = *++cur;
	if (yych == 'N') goto yy657;
	if (yych == 'n') goto yy657;
	goto yy55;
yy468:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'I') goto yy658;
			goto yy55;
		} else {
			if (yych <= 'O') goto yy659;
			if (yych <= 'T') goto yy55;
			goto yy660;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'i') goto yy658;
			goto yy55;
		} else {
			if (yych <= 'o') goto yy659;
			if (yych == 'u') goto yy660;
			goto yy55;
		}
	}
yy469:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'F') goto yy661;
		if (yych <= 'M') goto yy55;
		goto yy663;
	} else {
		if (yych <= 'f') {
			if (yych <= 'e') goto yy55;
			goto yy661;
		} else {
			if (yych == 'n') goto yy663;
			goto yy55;
		}
	}
yy470:
	yych = *++cur;
	if (yych == 'O') goto yy664;
	if (yych == 'o') goto yy664;
	goto yy55;
yy471:
	yych = *++cur;
	if (yych == 'R') goto yy665;
	if (yych == 'r') goto yy665;
	goto yy55;
yy472:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'E') goto yy666;
		if (yych <= 'T') goto yy55;
		goto yy667;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy666;
		} else {
			if (yych == 'u') goto yy667;
			goto yy55;
		}
	}
yy473:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'E') goto yy668;
		if (yych <= 'X') goto yy55;
		goto yy669;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy668;
		} else {
			if (yych == 'y') goto yy669;
			goto yy55;
		}
	}
yy474:
	yych = *++cur;
	if (yych == 'H') goto yy670;
	if (yych == 'h') goto yy670;
	goto yy55;
yy475:
	yych = *++cur;
	if (yych == 'Y') goto yy671;
	if (yych == 'y') goto yy671;
	goto yy55;
yy476:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'E') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy477;
			if (yych <= 'D') goto yy54;
			goto yy673;
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'e') goto yy673;
			if (yych <= 'z') goto yy54;
		}
	}
yy477:
#line 379 "tokenizer.re"
	{ KW(ENUM) }
#line 3965 "tokenizer.cpp"
yy478:
	yych = *++cur;
	if (yych == 'V') goto yy674;
	if (yych == 'v') goto yy674;
	goto yy55;
yy479:
	yych = *++cur;
	if (yych == 'S') goto yy675;
	if (yych == 's') goto yy675;
	goto yy55;
yy480:
	yych = *++cur;
	if (yych == 'R') goto yy676;
	if (yych == 'r') goto yy676;
	goto yy55;
yy481:
	yych = *++cur;
	if (yych == 'T') goto yy678;
	if (yych == 't') goto yy678;
	goto yy55;
yy482:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 385 "tokenizer.re"
	{ KW(EXIT) }
#line 3994 "tokenizer.cpp"
yy484:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy680;
		if (yych <= 'Q') goto yy55;
		goto yy681;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy680;
		} else {
			if (yych == 'r') goto yy681;
			goto yy55;
		}
	}
yy485:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 388 "tokenizer.re"
	{ KW(FILE) }
#line 4018 "tokenizer.cpp"
yy487:
	yych = *++cur;
	if (yych == 'L') goto yy682;
	if (yych == 'l') goto yy682;
	goto yy55;
yy488:
	yych = *++cur;
	if (yych == 'H') goto yy684;
	if (yych == 'h') goto yy684;
	goto yy55;
yy489:
	yych = *++cur;
	if (yych == 'L') goto yy686;
	if (yych == 'l') goto yy686;
	goto yy55;
yy490:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '@') {
			if (yych <= '/') goto yy491;
			if (yych <= '9') goto yy54;
		} else {
			if (yych <= 'A') goto yy687;
			if (yych <= 'S') goto yy54;
			goto yy688;
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy687;
		} else {
			if (yych == 't') goto yy688;
			if (yych <= 'z') goto yy54;
		}
	}
yy491:
#line 405 "tokenizer.re"
	{ KW(FORM) }
#line 4060 "tokenizer.cpp"
yy492:
	yych = *++cur;
	if (yych == 'T') goto yy689;
	if (yych == 't') goto yy689;
	goto yy55;
yy493:
	yych = *++cur;
	if (yych == 'R') goto yy690;
	if (yych == 'r') goto yy690;
	goto yy55;
yy494:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 410 "tokenizer.re"
	{ KW(GOTO) }
#line 4079 "tokenizer.cpp"
yy496:
	yych = *++cur;
	if (yych == 'E') goto yy691;
	if (yych == 'e') goto yy691;
	goto yy55;
yy497:
	yych = *++cur;
	if (yych == 'I') goto yy692;
	if (yych == 'i') goto yy692;
	goto yy55;
yy498:
	yych = *++cur;
	if (yych == 'R') goto yy693;
	if (yych == 'r') goto yy693;
	goto yy55;
yy499:
	yych = *++cur;
	if (yych == 'R') goto yy694;
	if (yych == 'r') goto yy694;
	goto yy55;
yy500:
	yych = *++cur;
	if (yych == 'U') goto yy695;
	if (yych == 'u') goto yy695;
	goto yy79;
yy501:
	yych = *++cur;
	if (yych == 'U') goto yy696;
	if (yych == 'u') goto yy696;
	goto yy55;
yy502:
	yych = *++cur;
	if (yych == 'T') goto yy697;
	if (yych == 't') goto yy697;
	goto yy55;
yy503:
	yych = *++cur;
	if (yych == 'I') goto yy699;
	if (yych == 'i') goto yy699;
	goto yy55;
yy504:
	yych = *++cur;
	if (yych == 'A') goto yy700;
	if (yych == 'a') goto yy700;
	goto yy55;
yy505:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych <= 'M') {
			if (yych == 'G') goto yy701;
			goto yy55;
		} else {
			if (yych <= 'N') goto yy702;
			if (yych <= 'Q') goto yy55;
			goto yy703;
		}
	} else {
		if (yych <= 'm') {
			if (yych == 'g') goto yy701;
			goto yy55;
		} else {
			if (yych <= 'n') goto yy702;
			if (yych == 'r') goto yy703;
			goto yy55;
		}
	}
yy506:
	yych = *++cur;
	if (yych == 'I') goto yy704;
	if (yych == 'i') goto yy704;
	goto yy55;
yy507:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 427 "tokenizer.re"
	{ KW(KIND) }
#line 4159 "tokenizer.cpp"
yy509:
	yych = *++cur;
	if (yych == 'L') goto yy705;
	if (yych == 'l') goto yy705;
	goto yy55;
yy510:
	yych = *++cur;
	if (yych == 'C') goto yy707;
	if (yych == 'c') goto yy707;
	goto yy55;
yy511:
	yych = *++cur;
	if (yych == 'R') goto yy708;
	if (yych == 'r') goto yy708;
	goto yy55;
yy512:
	yych = *++cur;
	if (yych == 'L') goto yy709;
	if (yych == 'l') goto yy709;
	goto yy55;
yy513:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 434 "tokenizer.re"
	{ KW(MOLD) }
#line 4188 "tokenizer.cpp"
yy515:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'L') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy516;
			if (yych <= 'K') goto yy54;
			goto yy710;
		}
	} else {
		if (yych <= 'k') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'l') goto yy710;
			if (yych <= 'z') goto yy54;
		}
	}
yy516:
#line 435 "tokenizer.re"
	{ KW(NAME) }
#line 4214 "tokenizer.cpp"
yy517:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'I') {
			if (yych <= 'H') goto yy55;
			goto yy711;
		} else {
			if (yych == 'i') goto yy711;
			goto yy55;
		}
	}
yy518:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 442 "tokenizer.re"
	{ KW(NONE) }
#line 4239 "tokenizer.cpp"
yy520:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'O') {
		if (yych <= '\'') {
			if (yych == '"') goto yy82;
			if (yych <= '&') goto yy55;
			goto yy97;
		} else {
			if (yych == 'I') goto yy712;
			if (yych <= 'N') goto yy55;
			goto yy713;
		}
	} else {
		if (yych <= 'i') {
			if (yych == 'R') goto yy714;
			if (yych <= 'h') goto yy55;
			goto yy712;
		} else {
			if (yych <= 'o') {
				if (yych <= 'n') goto yy55;
				goto yy713;
			} else {
				if (yych == 'r') goto yy714;
				goto yy55;
			}
		}
	}
yy521:
	yych = *++cur;
	if (yych == 'S') goto yy715;
	if (yych == 's') goto yy715;
	goto yy55;
yy522:
	yych = *++cur;
	if (yych == 'I') goto yy716;
	if (yych == 'i') goto yy716;
	goto yy55;
yy523:
	++cur;
#line 633 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4282 "tokenizer.cpp"
yy525:
	++cur;
#line 634 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4287 "tokenizer.cpp"
yy527:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 444 "tokenizer.re"
	{ KW(ONLY) }
#line 4296 "tokenizer.cpp"
yy529:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 445 "tokenizer.re"
	{ KW(OPEN) }
#line 4305 "tokenizer.cpp"
yy531:
	yych = *++cur;
	if (yych == 'A') goto yy717;
	if (yych == 'a') goto yy717;
	goto yy55;
yy532:
	yych = *++cur;
	if (yych == 'O') goto yy718;
	if (yych == 'o') goto yy718;
	goto yy55;
yy533:
	yych = *++cur;
	if (yych == 'M') goto yy719;
	if (yych == 'm') goto yy719;
	goto yy55;
yy534:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 450 "tokenizer.re"
	{ KW(PASS) }
#line 4329 "tokenizer.cpp"
yy536:
	yych = *++cur;
	if (yych == 'T') goto yy720;
	if (yych == 't') goto yy720;
	goto yy55;
yy537:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 452 "tokenizer.re"
	{ KW(POST) }
#line 4343 "tokenizer.cpp"
yy539:
	yych = *++cur;
	if (yych == 'I') goto yy721;
	if (yych == 'i') goto yy721;
	goto yy55;
yy540:
	yych = *++cur;
	if (yych == 'T') goto yy722;
	if (yych == 't') goto yy722;
	goto yy55;
yy541:
	yych = *++cur;
	if (yych == 'A') goto yy724;
	if (yych == 'a') goto yy724;
	goto yy55;
yy542:
	yych = *++cur;
	if (yych == 'E') goto yy725;
	if (yych == 'e') goto yy725;
	goto yy55;
yy543:
	yych = *++cur;
	if (yych == 'R') goto yy726;
	if (yych == 'r') goto yy726;
	goto yy55;
yy544:
	yych = *++cur;
	if (yych == 'E') goto yy727;
	if (yych == 'e') goto yy727;
	goto yy55;
yy545:
	yych = *++cur;
	if (yych == 'I') goto yy728;
	if (yych == 'i') goto yy728;
	goto yy55;
yy546:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 460 "tokenizer.re"
	{ KW(PURE) }
#line 4387 "tokenizer.cpp"
yy548:
	yych = *++cur;
	if (yych == 'T') goto yy729;
	if (yych == 't') goto yy729;
	goto yy55;
yy549:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 462 "tokenizer.re"
	{ KW(RANK) }
#line 4401 "tokenizer.cpp"
yy551:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 463 "tokenizer.re"
	{ KW(READ) }
#line 4410 "tokenizer.cpp"
yy553:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 464 "tokenizer.re"
	{KW(REAL) }
#line 4419 "tokenizer.cpp"
yy555:
	yych = *++cur;
	if (yych == 'R') goto yy731;
	if (yych == 'r') goto yy731;
	goto yy55;
yy556:
	yych = *++cur;
	if (yych == 'C') goto yy732;
	if (yych == 'c') goto yy732;
	goto yy55;
yy557:
	yych = *++cur;
	if (yych == 'L') goto yy733;
	if (yych == 'l') goto yy733;
	goto yy55;
yy558:
	yych = *++cur;
	if (yych == 'R') goto yy734;
	if (yych == 'r') goto yy734;
	goto yy55;
yy559:
	yych = *++cur;
	if (yych == 'N') goto yy735;
	if (yych == 'n') goto yy735;
	goto yy55;
yy560:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 470 "tokenizer.re"
	{ KW(SAVE) }
#line 4453 "tokenizer.cpp"
yy562:
	yych = *++cur;
	if (yych == 'C') goto yy736;
	if (yych == 'c') goto yy736;
	goto yy55;
yy563:
	yych = *++cur;
	if (yych == 'E') goto yy737;
	if (yych == 'e') goto yy737;
	goto yy55;
yy564:
	yych = *++cur;
	if (yych == 'E') goto yy738;
	if (yych == 'e') goto yy738;
	goto yy55;
yy565:
	yych = *++cur;
	if (yych == 'C') goto yy739;
	if (yych == 'c') goto yy739;
	goto yy55;
yy566:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 478 "tokenizer.re"
	{ KW(STAT) }
#line 4482 "tokenizer.cpp"
yy568:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 479 "tokenizer.re"
	{ KW(STOP) }
#line 4491 "tokenizer.cpp"
yy570:
	yych = *++cur;
	if (yych == 'O') goto yy740;
	if (yych == 'o') goto yy740;
	goto yy55;
yy571:
	yych = *++cur;
	if (yych == 'O') goto yy741;
	if (yych == 'o') goto yy741;
	goto yy55;
yy572:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= 'H') {
			if (yych <= '9') {
				if (yych >= '0') goto yy54;
			} else {
				if (yych <= '@') goto yy573;
				if (yych <= 'A') goto yy742;
				goto yy54;
			}
		} else {
			if (yych <= 'L') {
				if (yych <= 'I') goto yy743;
				goto yy54;
			} else {
				if (yych <= 'M') goto yy744;
				if (yych <= 'S') goto yy54;
				goto yy745;
			}
		}
	} else {
		if (yych <= 'i') {
			if (yych <= '`') {
				if (yych <= 'Z') goto yy54;
			} else {
				if (yych <= 'a') goto yy742;
				if (yych <= 'h') goto yy54;
				goto yy743;
			}
		} else {
			if (yych <= 's') {
				if (yych == 'm') goto yy744;
				goto yy54;
			} else {
				if (yych <= 't') goto yy745;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy573:
#line 482 "tokenizer.re"
	{ KW(SYNC) }
#line 4548 "tokenizer.cpp"
yy574:
	yych = *++cur;
	if (yych == 'E') goto yy746;
	if (yych == 'e') goto yy746;
	goto yy55;
yy575:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy747;
#line 488 "tokenizer.re"
	{ KW(TEAM) }
#line 4562 "tokenizer.cpp"
yy577:
	yych = *++cur;
	if (yych == 'L') goto yy748;
	if (yych == 'l') goto yy748;
	goto yy55;
yy578:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 491 "tokenizer.re"
	{ KW(THEN) }
#line 4576 "tokenizer.cpp"
yy580:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 493 "tokenizer.re"
	{ KW(TYPE) }
#line 4585 "tokenizer.cpp"
yy582:
	yych = *++cur;
	if (yych == 'R') goto yy749;
	if (yych == 'r') goto yy749;
	goto yy55;
yy583:
	yych = *++cur;
	if (yych == 'E') goto yy750;
	if (yych == 'e') goto yy750;
	goto yy55;
yy584:
	yych = *++cur;
	if (yych == 'T') goto yy752;
	if (yych == 't') goto yy752;
	goto yy55;
yy585:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 498 "tokenizer.re"
	{ KW(WAIT) }
#line 4609 "tokenizer.cpp"
yy587:
	yych = *++cur;
	if (yych == 'E') goto yy753;
	if (yych == 'e') goto yy753;
	goto yy55;
yy588:
	yych = *++cur;
	if (yych == 'E') goto yy755;
	if (yych == 'e') goto yy755;
	goto yy55;
yy589:
	yych = *++cur;
	if (yych == 'E') goto yy757;
	if (yych == 'e') goto yy757;
	goto yy55;
yy590:
	++cur;
#line 635 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4629 "tokenizer.cpp"
yy592:
	++cur;
#line 636 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4634 "tokenizer.cpp"
yy594:
	++cur;
	cur = ctxmar;
#line 525 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 4640 "tokenizer.cpp"
yy596:
	++cur;
#line 571 "tokenizer.re"
	{ RET(TK_AND) }
#line 4645 "tokenizer.cpp"
yy598:
	++cur;
#line 574 "tokenizer.re"
	{ RET(TK_EQV) }
#line 4650 "tokenizer.cpp"
yy600:
	yych = *++cur;
	if (yych == 'E') goto yy759;
	if (yych == 'e') goto yy759;
	goto yy111;
yy601:
	yych = *++cur;
	if (yych == '.') goto yy760;
	goto yy111;
yy602:
	++cur;
#line 570 "tokenizer.re"
	{ RET(TK_NOT) }
#line 4664 "tokenizer.cpp"
yy604:
	yych = *++cur;
	if (yych == '.') goto yy762;
	goto yy111;
yy605:
	++cur;
#line 573 "tokenizer.re"
	{ RET(TK_XOR) }
#line 4673 "tokenizer.cpp"
yy607:
	yych = *++cur;
	if (yych == 'A') goto yy764;
	if (yych == 'a') goto yy764;
	goto yy55;
yy608:
	yych = *++cur;
	if (yych == 'A') goto yy765;
	if (yych == 'a') goto yy765;
	goto yy55;
yy609:
	yych = *++cur;
	if (yych == 'N') goto yy766;
	if (yych == 'n') goto yy766;
	goto yy55;
yy610:
	yych = *++cur;
	if (yych == 'I') goto yy768;
	if (yych == 'i') goto yy768;
	goto yy55;
yy611:
	yych = *++cur;
	if (yych == 'H') goto yy769;
	if (yych == 'h') goto yy769;
	goto yy55;
yy612:
	yych = *++cur;
	if (yych == 'P') goto yy770;
	if (yych == 'p') goto yy770;
	goto yy55;
yy613:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 263 "tokenizer.re"
	{ KW(BLOCK) }
#line 4712 "tokenizer.cpp"
yy615:
	yych = *++cur;
	if (yych == 'E') goto yy771;
	if (yych == 'e') goto yy771;
	goto yy55;
yy616:
	yych = *++cur;
	if (yych == 'C') goto yy773;
	if (yych == 'c') goto yy773;
	goto yy55;
yy617:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 269 "tokenizer.re"
	{ KW(CLASS) }
#line 4731 "tokenizer.cpp"
yy619:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 270 "tokenizer.re"
	{ KW(CLOSE) }
#line 4740 "tokenizer.cpp"
yy621:
	yych = *++cur;
	if (yych == 'E') goto yy774;
	if (yych == 'e') goto yy774;
	goto yy55;
yy622:
	yych = *++cur;
	if (yych == 'N') goto yy775;
	if (yych == 'n') goto yy775;
	goto yy55;
yy623:
	yych = *++cur;
	if (yych == 'E') goto yy777;
	if (yych == 'e') goto yy777;
	goto yy55;
yy624:
	yych = *++cur;
	if (yych == 'R') goto yy778;
	if (yych == 'r') goto yy778;
	goto yy55;
yy625:
	yych = *++cur;
	if (yych == 'I') goto yy779;
	if (yych == 'i') goto yy779;
	goto yy55;
yy626:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'G') goto yy780;
		if (yych <= 'M') goto yy55;
		goto yy781;
	} else {
		if (yych <= 'g') {
			if (yych <= 'f') goto yy55;
			goto yy780;
		} else {
			if (yych == 'n') goto yy781;
			goto yy55;
		}
	}
yy627:
	yych = *++cur;
	if (yych == 'C') goto yy782;
	if (yych == 'c') goto yy782;
	goto yy55;
yy628:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 279 "tokenizer.re"
	{ KW(CYCLE) }
#line 4794 "tokenizer.cpp"
yy630:
	yych = *++cur;
	if (yych == 'O') goto yy783;
	if (yych == 'o') goto yy783;
	goto yy55;
yy631:
	yych = *++cur;
	if (yych == 'L') goto yy784;
	if (yych == 'l') goto yy784;
	goto yy55;
yy632:
	yych = *++cur;
	if (yych == 'R') goto yy785;
	if (yych == 'r') goto yy785;
	goto yy55;
yy633:
	yych = *++cur;
	if (yych == 'S') goto yy786;
	if (yych == 's') goto yy786;
	goto yy55;
yy634:
	yych = *++cur;
	if (yych == 'E') goto yy787;
	if (yych == 'e') goto yy787;
	goto yy55;
yy635:
	yych = *++cur;
	if (yych == 'L') goto yy789;
	if (yych == 'l') goto yy789;
	goto yy55;
yy636:
	yych = *++cur;
	if (yych == 'N') goto yy790;
	if (yych == 'n') goto yy790;
	goto yy55;
yy637:
	yych = *++cur;
	if (yych == 'F') goto yy791;
	if (yych == 'f') goto yy791;
	goto yy55;
yy638:
	yych = *++cur;
	if (yych == 'H') goto yy793;
	if (yych == 'h') goto yy793;
	goto yy55;
yy639:
	yych = *++cur;
	if (yych == 'S') goto yy794;
	if (yych == 's') goto yy794;
	goto yy79;
yy640:
	yych = *++cur;
	if (yych == 'L') goto yy795;
	if (yych == 'l') goto yy795;
	goto yy79;
yy641:
	yych = *++cur;
	if (yych == 'R') goto yy796;
	if (yych == 'r') goto yy796;
	goto yy79;
yy642:
	yych = *++cur;
	if (yych == 'O') goto yy797;
	if (yych == 'o') goto yy797;
	goto yy79;
yy643:
	yych = *++cur;
	if (yych == 'N') goto yy799;
	if (yych == 'n') goto yy799;
	goto yy79;
yy644:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'O') goto yy800;
		if (yych <= 'T') goto yy79;
		goto yy801;
	} else {
		if (yych <= 'o') {
			if (yych <= 'n') goto yy79;
			goto yy800;
		} else {
			if (yych == 'u') goto yy801;
			goto yy79;
		}
	}
yy645:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'F') goto yy802;
		if (yych <= 'M') goto yy79;
		goto yy804;
	} else {
		if (yych <= 'f') {
			if (yych <= 'e') goto yy79;
			goto yy802;
		} else {
			if (yych == 'n') goto yy804;
			goto yy79;
		}
	}
yy646:
	yych = *++cur;
	if (yych == 'O') goto yy805;
	if (yych == 'o') goto yy805;
	goto yy79;
yy647:
	yych = *++cur;
	if (yych == 'R') goto yy806;
	if (yych == 'r') goto yy806;
	goto yy79;
yy648:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'E') goto yy807;
		if (yych <= 'T') goto yy79;
		goto yy808;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy79;
			goto yy807;
		} else {
			if (yych == 'u') goto yy808;
			goto yy79;
		}
	}
yy649:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'E') goto yy809;
		if (yych <= 'X') goto yy79;
		goto yy810;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy79;
			goto yy809;
		} else {
			if (yych == 'y') goto yy810;
			goto yy79;
		}
	}
yy650:
	yych = *++cur;
	if (yych == 'H') goto yy811;
	if (yych == 'h') goto yy811;
	goto yy79;
yy651:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'I') goto yy812;
			goto yy79;
		} else {
			if (yych <= 'O') goto yy800;
			if (yych <= 'T') goto yy79;
			goto yy801;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'i') goto yy812;
			goto yy79;
		} else {
			if (yych <= 'o') goto yy800;
			if (yych == 'u') goto yy801;
			goto yy79;
		}
	}
yy652:
	yych = *++cur;
	if (yych == 'S') goto yy813;
	if (yych == 's') goto yy813;
	goto yy55;
yy653:
	yych = *++cur;
	if (yych == 'O') goto yy814;
	if (yych == 'o') goto yy814;
	goto yy55;
yy654:
	yych = *++cur;
	if (yych == 'I') goto yy815;
	if (yych == 'i') goto yy815;
	goto yy55;
yy655:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 364 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    KW(CONTINUE)
                } else {
                    KW(ENDDO)
                }
            }
#line 4990 "tokenizer.cpp"
yy657:
	yych = *++cur;
	if (yych == 'U') goto yy816;
	if (yych == 'u') goto yy816;
	goto yy55;
yy658:
	yych = *++cur;
	if (yych == 'L') goto yy817;
	if (yych == 'l') goto yy817;
	goto yy55;
yy659:
	yych = *++cur;
	if (yych == 'R') goto yy818;
	if (yych == 'r') goto yy818;
	goto yy55;
yy660:
	yych = *++cur;
	if (yych == 'N') goto yy819;
	if (yych == 'n') goto yy819;
	goto yy55;
yy661:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 349 "tokenizer.re"
	{ KW(ENDIF) }
#line 5019 "tokenizer.cpp"
yy663:
	yych = *++cur;
	if (yych == 'T') goto yy820;
	if (yych == 't') goto yy820;
	goto yy55;
yy664:
	yych = *++cur;
	if (yych == 'D') goto yy821;
	if (yych == 'd') goto yy821;
	goto yy55;
yy665:
	yych = *++cur;
	if (yych == 'O') goto yy822;
	if (yych == 'o') goto yy822;
	goto yy55;
yy666:
	yych = *++cur;
	if (yych == 'L') goto yy823;
	if (yych == 'l') goto yy823;
	goto yy55;
yy667:
	yych = *++cur;
	if (yych == 'B') goto yy824;
	if (yych == 'b') goto yy824;
	goto yy55;
yy668:
	yych = *++cur;
	if (yych == 'A') goto yy825;
	if (yych == 'a') goto yy825;
	goto yy55;
yy669:
	yych = *++cur;
	if (yych == 'P') goto yy826;
	if (yych == 'p') goto yy826;
	goto yy55;
yy670:
	yych = *++cur;
	if (yych == 'E') goto yy827;
	if (yych == 'e') goto yy827;
	goto yy55;
yy671:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 378 "tokenizer.re"
	{ KW(ENTRY) }
#line 5068 "tokenizer.cpp"
yy673:
	yych = *++cur;
	if (yych == 'R') goto yy828;
	if (yych == 'r') goto yy828;
	goto yy55;
yy674:
	yych = *++cur;
	if (yych == 'A') goto yy829;
	if (yych == 'a') goto yy829;
	goto yy55;
yy675:
	yych = *++cur;
	if (yych == 'G') goto yy830;
	if (yych == 'g') goto yy830;
	goto yy55;
yy676:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 383 "tokenizer.re"
	{ KW(ERROR) }
#line 5092 "tokenizer.cpp"
yy678:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 384 "tokenizer.re"
	{ KW(EVENT) }
#line 5101 "tokenizer.cpp"
yy680:
	yych = *++cur;
	if (yych == 'D') goto yy832;
	if (yych == 'd') goto yy832;
	goto yy55;
yy681:
	yych = *++cur;
	if (yych == 'N') goto yy833;
	if (yych == 'n') goto yy833;
	goto yy55;
yy682:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 389 "tokenizer.re"
	{ KW(FINAL) }
#line 5120 "tokenizer.cpp"
yy684:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 390 "tokenizer.re"
	{ KW(FLUSH) }
#line 5129 "tokenizer.cpp"
yy686:
	yych = *++cur;
	if (yych == 'L') goto yy834;
	if (yych == 'l') goto yy834;
	goto yy55;
yy687:
	yych = *++cur;
	if (yych == 'T') goto yy836;
	if (yych == 't') goto yy836;
	goto yy55;
yy688:
	yych = *++cur;
	if (yych == 'E') goto yy838;
	if (yych == 'e') goto yy838;
	goto yy55;
yy689:
	yych = *++cur;
	if (yych == 'I') goto yy839;
	if (yych == 'i') goto yy839;
	goto yy55;
yy690:
	yych = *++cur;
	if (yych == 'I') goto yy840;
	if (yych == 'i') goto yy840;
	goto yy55;
yy691:
	yych = *++cur;
	if (yych == 'S') goto yy841;
	if (yych == 's') goto yy841;
	goto yy55;
yy692:
	yych = *++cur;
	if (yych == 'C') goto yy843;
	if (yych == 'c') goto yy843;
	goto yy55;
yy693:
	yych = *++cur;
	if (yych == 'T') goto yy844;
	if (yych == 't') goto yy844;
	goto yy55;
yy694:
	yych = *++cur;
	if (yych == 'E') goto yy846;
	if (yych == 'e') goto yy846;
	goto yy55;
yy695:
	yych = *++cur;
	if (yych == 'T') goto yy848;
	if (yych == 't') goto yy848;
	goto yy79;
yy696:
	yych = *++cur;
	if (yych == 'D') goto yy850;
	if (yych == 'd') goto yy850;
	goto yy55;
yy697:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 418 "tokenizer.re"
	{ KW(INOUT) }
#line 5193 "tokenizer.cpp"
yy699:
	yych = *++cur;
	if (yych == 'R') goto yy851;
	if (yych == 'r') goto yy851;
	goto yy55;
yy700:
	yych = *++cur;
	if (yych == 'N') goto yy852;
	if (yych == 'n') goto yy852;
	goto yy55;
yy701:
	yych = *++cur;
	if (yych == 'E') goto yy853;
	if (yych == 'e') goto yy853;
	goto yy55;
yy702:
	yych = *++cur;
	if (yych == 'T') goto yy854;
	if (yych == 't') goto yy854;
	goto yy55;
yy703:
	yych = *++cur;
	if (yych == 'F') goto yy856;
	if (yych == 'f') goto yy856;
	goto yy55;
yy704:
	yych = *++cur;
	if (yych == 'N') goto yy857;
	if (yych == 'n') goto yy857;
	goto yy55;
yy705:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy858;
#line 429 "tokenizer.re"
	{ KW(LOCAL) }
#line 5232 "tokenizer.cpp"
yy707:
	yych = *++cur;
	if (yych == 'A') goto yy859;
	if (yych == 'a') goto yy859;
	goto yy55;
yy708:
	yych = *++cur;
	if (yych == 'Y') goto yy860;
	if (yych == 'y') goto yy860;
	goto yy55;
yy709:
	yych = *++cur;
	if (yych == 'E') goto yy862;
	if (yych == 'e') goto yy862;
	goto yy55;
yy710:
	yych = *++cur;
	if (yych == 'I') goto yy864;
	if (yych == 'i') goto yy864;
	goto yy55;
yy711:
	yych = *++cur;
	if (yych == 'N') goto yy865;
	if (yych == 'n') goto yy865;
	goto yy55;
yy712:
	yych = *++cur;
	if (yych == 'N') goto yy866;
	if (yych == 'n') goto yy866;
	goto yy55;
yy713:
	yych = *++cur;
	if (yych == 'V') goto yy867;
	if (yych == 'v') goto yy867;
	goto yy55;
yy714:
	yych = *++cur;
	if (yych == 'E') goto yy868;
	if (yych == 'e') goto yy868;
	goto yy55;
yy715:
	yych = *++cur;
	if (yych == 'S') goto yy869;
	if (yych == 's') goto yy869;
	goto yy55;
yy716:
	yych = *++cur;
	if (yych == 'F') goto yy871;
	if (yych == 'f') goto yy871;
	goto yy55;
yy717:
	yych = *++cur;
	if (yych == 'T') goto yy872;
	if (yych == 't') goto yy872;
	goto yy55;
yy718:
	yych = *++cur;
	if (yych == 'N') goto yy873;
	if (yych == 'n') goto yy873;
	goto yy55;
yy719:
	yych = *++cur;
	if (yych == 'E') goto yy874;
	if (yych == 'e') goto yy874;
	goto yy55;
yy720:
	yych = *++cur;
	if (yych == 'E') goto yy875;
	if (yych == 'e') goto yy875;
	goto yy55;
yy721:
	yych = *++cur;
	if (yych == 'S') goto yy876;
	if (yych == 's') goto yy876;
	goto yy55;
yy722:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 454 "tokenizer.re"
	{ KW(PRINT) }
#line 5316 "tokenizer.cpp"
yy724:
	yych = *++cur;
	if (yych == 'T') goto yy877;
	if (yych == 't') goto yy877;
	goto yy55;
yy725:
	yych = *++cur;
	if (yych == 'D') goto yy878;
	if (yych == 'd') goto yy878;
	goto yy55;
yy726:
	yych = *++cur;
	if (yych == 'A') goto yy879;
	if (yych == 'a') goto yy879;
	goto yy55;
yy727:
	yych = *++cur;
	if (yych == 'C') goto yy880;
	if (yych == 'c') goto yy880;
	goto yy55;
yy728:
	yych = *++cur;
	if (yych == 'C') goto yy881;
	if (yych == 'c') goto yy881;
	goto yy55;
yy729:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 461 "tokenizer.re"
	{ KW(QUIET) }
#line 5350 "tokenizer.cpp"
yy731:
	yych = *++cur;
	if (yych == 'S') goto yy883;
	if (yych == 's') goto yy883;
	goto yy55;
yy732:
	yych = *++cur;
	if (yych == 'E') goto yy884;
	if (yych == 'e') goto yy884;
	goto yy55;
yy733:
	yych = *++cur;
	if (yych == 'T') goto yy886;
	if (yych == 't') goto yy886;
	goto yy55;
yy734:
	yych = *++cur;
	if (yych == 'N') goto yy888;
	if (yych == 'n') goto yy888;
	goto yy55;
yy735:
	yych = *++cur;
	if (yych == 'D') goto yy890;
	if (yych == 'd') goto yy890;
	goto yy55;
yy736:
	yych = *++cur;
	if (yych == 'T') goto yy892;
	if (yych == 't') goto yy892;
	goto yy55;
yy737:
	yych = *++cur;
	if (yych == 'N') goto yy894;
	if (yych == 'n') goto yy894;
	goto yy55;
yy738:
	yych = *++cur;
	if (yych == 'D') goto yy895;
	if (yych == 'd') goto yy895;
	goto yy55;
yy739:
	yych = *++cur;
	if (yych == 'E') goto yy897;
	if (yych == 'e') goto yy897;
	goto yy55;
yy740:
	yych = *++cur;
	if (yych == 'D') goto yy899;
	if (yych == 'd') goto yy899;
	goto yy55;
yy741:
	yych = *++cur;
	if (yych == 'U') goto yy900;
	if (yych == 'u') goto yy900;
	goto yy55;
yy742:
	yych = *++cur;
	if (yych == 'L') goto yy901;
	if (yych == 'l') goto yy901;
	goto yy55;
yy743:
	yych = *++cur;
	if (yych == 'M') goto yy902;
	if (yych == 'm') goto yy902;
	goto yy55;
yy744:
	yych = *++cur;
	if (yych == 'E') goto yy903;
	if (yych == 'e') goto yy903;
	goto yy55;
yy745:
	yych = *++cur;
	if (yych == 'E') goto yy904;
	if (yych == 'e') goto yy904;
	goto yy55;
yy746:
	yych = *++cur;
	if (yych == 'T') goto yy905;
	if (yych == 't') goto yy905;
	goto yy55;
yy747:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'N') {
			if (yych <= 'M') goto yy55;
			goto yy907;
		} else {
			if (yych == 'n') goto yy907;
			goto yy55;
		}
	}
yy748:
	yych = *++cur;
	if (yych == 'A') goto yy908;
	if (yych == 'a') goto yy908;
	goto yy55;
yy749:
	yych = *++cur;
	if (yych == 'M') goto yy909;
	if (yych == 'm') goto yy909;
	goto yy55;
yy750:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 496 "tokenizer.re"
	{ KW(VALUE) }
#line 5465 "tokenizer.cpp"
yy752:
	yych = *++cur;
	if (yych == 'I') goto yy910;
	if (yych == 'i') goto yy910;
	goto yy55;
yy753:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 499 "tokenizer.re"
	{ KW(WHERE) }
#line 5479 "tokenizer.cpp"
yy755:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 500 "tokenizer.re"
	{ KW(WHILE) }
#line 5488 "tokenizer.cpp"
yy757:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 501 "tokenizer.re"
	{ KW(WRITE) }
#line 5497 "tokenizer.cpp"
yy759:
	yych = *++cur;
	if (yych == '.') goto yy911;
	goto yy111;
yy760:
	++cur;
#line 575 "tokenizer.re"
	{ RET(TK_NEQV) }
#line 5506 "tokenizer.cpp"
yy762:
	yyaccept = 11;
	yych = *(mar = ++cur);
	if (yych == '_') goto yy913;
yy763:
#line 579 "tokenizer.re"
	{ RET(TK_TRUE) }
#line 5514 "tokenizer.cpp"
yy764:
	yych = *++cur;
	if (yych == 'C') goto yy914;
	if (yych == 'c') goto yy914;
	goto yy55;
yy765:
	yych = *++cur;
	if (yych == 'T') goto yy915;
	if (yych == 't') goto yy915;
	goto yy55;
yy766:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'M') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy767;
			if (yych <= 'L') goto yy54;
			goto yy916;
		}
	} else {
		if (yych <= 'l') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'm') goto yy916;
			if (yych <= 'z') goto yy54;
		}
	}
yy767:
#line 257 "tokenizer.re"
	{ KW(ASSIGN) }
#line 5550 "tokenizer.cpp"
yy768:
	yych = *++cur;
	if (yych == 'A') goto yy917;
	if (yych == 'a') goto yy917;
	goto yy55;
yy769:
	yych = *++cur;
	if (yych == 'R') goto yy918;
	if (yych == 'r') goto yy918;
	goto yy55;
yy770:
	yych = *++cur;
	if (yych == 'A') goto yy919;
	if (yych == 'a') goto yy919;
	goto yy55;
yy771:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy772;
			if (yych <= 'S') goto yy54;
			goto yy920;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy920;
			if (yych <= 'z') goto yy54;
		}
	}
yy772:
#line 266 "tokenizer.re"
	{ KW(CHANGE) }
#line 5591 "tokenizer.cpp"
yy773:
	yych = *++cur;
	if (yych == 'T') goto yy921;
	if (yych == 't') goto yy921;
	goto yy55;
yy774:
	yych = *++cur;
	if (yych == 'N') goto yy922;
	if (yych == 'n') goto yy922;
	goto yy55;
yy775:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 272 "tokenizer.re"
	{ KW(COMMON) }
#line 5610 "tokenizer.cpp"
yy777:
	yych = *++cur;
	if (yych == 'X') goto yy923;
	if (yych == 'x') goto yy923;
	goto yy55;
yy778:
	yych = *++cur;
	if (yych == 'R') goto yy925;
	if (yych == 'r') goto yy925;
	goto yy55;
yy779:
	yych = *++cur;
	if (yych == 'N') goto yy926;
	if (yych == 'n') goto yy926;
	goto yy55;
yy780:
	yych = *++cur;
	if (yych == 'U') goto yy927;
	if (yych == 'u') goto yy927;
	goto yy55;
yy781:
	yych = *++cur;
	if (yych == 'U') goto yy928;
	if (yych == 'u') goto yy928;
	goto yy55;
yy782:
	yych = *++cur;
	if (yych == 'A') goto yy929;
	if (yych == 'a') goto yy929;
	goto yy55;
yy783:
	yych = *++cur;
	if (yych == 'C') goto yy930;
	if (yych == 'c') goto yy930;
	goto yy55;
yy784:
	yych = *++cur;
	if (yych == 'T') goto yy931;
	if (yych == 't') goto yy931;
	goto yy55;
yy785:
	yych = *++cur;
	if (yych == 'E') goto yy933;
	if (yych == 'e') goto yy933;
	goto yy55;
yy786:
	yych = *++cur;
	if (yych == 'I') goto yy934;
	if (yych == 'i') goto yy934;
	goto yy55;
yy787:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'P') {
		if (yych <= '@') {
			if (yych <= '/') goto yy788;
			if (yych <= '9') goto yy54;
		} else {
			if (yych == 'C') goto yy935;
			if (yych <= 'O') goto yy54;
			goto yy936;
		}
	} else {
		if (yych <= 'c') {
			if (yych <= 'Z') goto yy54;
			if (yych <= '`') goto yy788;
			if (yych <= 'b') goto yy54;
			goto yy935;
		} else {
			if (yych == 'p') goto yy936;
			if (yych <= 'z') goto yy54;
		}
	}
yy788:
#line 294 "tokenizer.re"
	{ KW(DOUBLE) }
#line 5689 "tokenizer.cpp"
yy789:
	yych = *++cur;
	if (yych == 'E') goto yy937;
	if (yych == 'e') goto yy937;
	goto yy55;
yy790:
	yych = *++cur;
	if (yych == 'T') goto yy939;
	if (yych == 't') goto yy939;
	goto yy55;
yy791:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 299 "tokenizer.re"
	{ KW(ELSEIF) }
#line 5708 "tokenizer.cpp"
yy793:
	yych = *++cur;
	if (yych == 'E') goto yy940;
	if (yych == 'e') goto yy940;
	goto yy55;
yy794:
	yych = *++cur;
	if (yych == 'S') goto yy941;
	if (yych == 's') goto yy941;
	goto yy79;
yy795:
	yych = *++cur;
	if (yych == 'O') goto yy942;
	if (yych == 'o') goto yy942;
	goto yy79;
yy796:
	yych = *++cur;
	if (yych == 'I') goto yy943;
	if (yych == 'i') goto yy943;
	goto yy79;
yy797:
	++cur;
#line 357 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    KW(CONTINUE)
                } else {
                    KW(END_DO)
                }
            }
#line 5739 "tokenizer.cpp"
yy799:
	yych = *++cur;
	if (yych == 'U') goto yy944;
	if (yych == 'u') goto yy944;
	goto yy79;
yy800:
	yych = *++cur;
	if (yych == 'R') goto yy945;
	if (yych == 'r') goto yy945;
	goto yy79;
yy801:
	yych = *++cur;
	if (yych == 'N') goto yy946;
	if (yych == 'n') goto yy946;
	goto yy79;
yy802:
	++cur;
#line 348 "tokenizer.re"
	{ KW(END_IF) }
#line 5759 "tokenizer.cpp"
yy804:
	yych = *++cur;
	if (yych == 'T') goto yy947;
	if (yych == 't') goto yy947;
	goto yy79;
yy805:
	yych = *++cur;
	if (yych == 'D') goto yy948;
	if (yych == 'd') goto yy948;
	goto yy79;
yy806:
	yych = *++cur;
	if (yych == 'O') goto yy949;
	if (yych == 'o') goto yy949;
	goto yy79;
yy807:
	yych = *++cur;
	if (yych == 'L') goto yy950;
	if (yych == 'l') goto yy950;
	goto yy79;
yy808:
	yych = *++cur;
	if (yych == 'B') goto yy951;
	if (yych == 'b') goto yy951;
	goto yy79;
yy809:
	yych = *++cur;
	if (yych == 'A') goto yy952;
	if (yych == 'a') goto yy952;
	goto yy79;
yy810:
	yych = *++cur;
	if (yych == 'P') goto yy953;
	if (yych == 'p') goto yy953;
	goto yy79;
yy811:
	yych = *++cur;
	if (yych == 'E') goto yy954;
	if (yych == 'e') goto yy954;
	goto yy79;
yy812:
	yych = *++cur;
	if (yych == 'L') goto yy955;
	if (yych == 'l') goto yy955;
	goto yy79;
yy813:
	yych = *++cur;
	if (yych == 'O') goto yy956;
	if (yych == 'o') goto yy956;
	goto yy55;
yy814:
	yych = *++cur;
	if (yych == 'C') goto yy957;
	if (yych == 'c') goto yy957;
	goto yy55;
yy815:
	yych = *++cur;
	if (yych == 'T') goto yy958;
	if (yych == 't') goto yy958;
	goto yy55;
yy816:
	yych = *++cur;
	if (yych == 'M') goto yy959;
	if (yych == 'm') goto yy959;
	goto yy55;
yy817:
	yych = *++cur;
	if (yych == 'E') goto yy961;
	if (yych == 'e') goto yy961;
	goto yy55;
yy818:
	yych = *++cur;
	if (yych == 'A') goto yy963;
	if (yych == 'a') goto yy963;
	goto yy55;
yy819:
	yych = *++cur;
	if (yych == 'C') goto yy964;
	if (yych == 'c') goto yy964;
	goto yy55;
yy820:
	yych = *++cur;
	if (yych == 'E') goto yy965;
	if (yych == 'e') goto yy965;
	goto yy55;
yy821:
	yych = *++cur;
	if (yych == 'U') goto yy966;
	if (yych == 'u') goto yy966;
	goto yy55;
yy822:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy967;
		if (yych <= 'F') goto yy55;
		goto yy968;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy967;
		} else {
			if (yych == 'g') goto yy968;
			goto yy55;
		}
	}
yy823:
	yych = *++cur;
	if (yych == 'E') goto yy969;
	if (yych == 'e') goto yy969;
	goto yy55;
yy824:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy970;
		if (yych <= 'Q') goto yy55;
		goto yy971;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy970;
		} else {
			if (yych == 'r') goto yy971;
			goto yy55;
		}
	}
yy825:
	yych = *++cur;
	if (yych == 'M') goto yy972;
	if (yych == 'm') goto yy972;
	goto yy55;
yy826:
	yych = *++cur;
	if (yych == 'E') goto yy974;
	if (yych == 'e') goto yy974;
	goto yy55;
yy827:
	yych = *++cur;
	if (yych == 'R') goto yy976;
	if (yych == 'r') goto yy976;
	goto yy55;
yy828:
	yych = *++cur;
	if (yych == 'A') goto yy977;
	if (yych == 'a') goto yy977;
	goto yy55;
yy829:
	yych = *++cur;
	if (yych == 'L') goto yy978;
	if (yych == 'l') goto yy978;
	goto yy55;
yy830:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 382 "tokenizer.re"
	{ KW(ERRMSG) }
#line 5918 "tokenizer.cpp"
yy832:
	yych = *++cur;
	if (yych == 'S') goto yy979;
	if (yych == 's') goto yy979;
	goto yy55;
yy833:
	yych = *++cur;
	if (yych == 'A') goto yy981;
	if (yych == 'a') goto yy981;
	goto yy55;
yy834:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 391 "tokenizer.re"
	{ KW(FORALL) }
#line 5937 "tokenizer.cpp"
yy836:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy837;
			if (yych <= 'S') goto yy54;
			goto yy982;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy982;
			if (yych <= 'z') goto yy54;
		}
	}
yy837:
#line 392 "tokenizer.re"
	{
                if (last_token == yytokentype::TK_LABEL) {
                    unsigned char *start;
                    lex_format(cur, loc, start);
                    yylval.string.p = (char*) start;
                    yylval.string.n = cur-start-1;
                    RET(TK_FORMAT)
                } else {
                    token(yylval.string);
                    RET(TK_NAME)
                }
            }
#line 5974 "tokenizer.cpp"
yy838:
	yych = *++cur;
	if (yych == 'A') goto yy983;
	if (yych == 'a') goto yy983;
	goto yy55;
yy839:
	yych = *++cur;
	if (yych == 'O') goto yy984;
	if (yych == 'o') goto yy984;
	goto yy55;
yy840:
	yych = *++cur;
	if (yych == 'C') goto yy985;
	if (yych == 'c') goto yy985;
	goto yy55;
yy841:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 412 "tokenizer.re"
	{ KW(IMAGES) }
#line 5998 "tokenizer.cpp"
yy843:
	yych = *++cur;
	if (yych == 'I') goto yy987;
	if (yych == 'i') goto yy987;
	goto yy55;
yy844:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 414 "tokenizer.re"
	{ KW(IMPORT) }
#line 6012 "tokenizer.cpp"
yy846:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 415 "tokenizer.re"
	{ KW(IMPURE) }
#line 6021 "tokenizer.cpp"
yy848:
	++cur;
#line 419 "tokenizer.re"
	{ KW(IN_OUT) }
#line 6026 "tokenizer.cpp"
yy850:
	yych = *++cur;
	if (yych == 'E') goto yy988;
	if (yych == 'e') goto yy988;
	goto yy55;
yy851:
	yych = *++cur;
	if (yych == 'E') goto yy990;
	if (yych == 'e') goto yy990;
	goto yy55;
yy852:
	yych = *++cur;
	if (yych == 'T') goto yy992;
	if (yych == 't') goto yy992;
	goto yy55;
yy853:
	yych = *++cur;
	if (yych == 'R') goto yy993;
	if (yych == 'r') goto yy993;
	goto yy55;
yy854:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 423 "tokenizer.re"
	{ KW(INTENT) }
#line 6055 "tokenizer.cpp"
yy856:
	yych = *++cur;
	if (yych == 'A') goto yy995;
	if (yych == 'a') goto yy995;
	goto yy55;
yy857:
	yych = *++cur;
	if (yych == 'S') goto yy996;
	if (yych == 's') goto yy996;
	goto yy55;
yy858:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'I') {
			if (yych <= 'H') goto yy55;
			goto yy997;
		} else {
			if (yych == 'i') goto yy997;
			goto yy55;
		}
	}
yy859:
	yych = *++cur;
	if (yych == 'L') goto yy998;
	if (yych == 'l') goto yy998;
	goto yy55;
yy860:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 432 "tokenizer.re"
	{ KW(MEMORY) }
#line 6095 "tokenizer.cpp"
yy862:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 433 "tokenizer.re"
	{ KW(MODULE) }
#line 6104 "tokenizer.cpp"
yy864:
	yych = *++cur;
	if (yych == 'S') goto yy1000;
	if (yych == 's') goto yy1000;
	goto yy55;
yy865:
	yych = *++cur;
	if (yych == 'D') goto yy1001;
	if (yych == 'd') goto yy1001;
	goto yy55;
yy866:
	yych = *++cur;
	if (yych == 'T') goto yy1002;
	if (yych == 't') goto yy1002;
	goto yy55;
yy867:
	yych = *++cur;
	if (yych == 'E') goto yy1003;
	if (yych == 'e') goto yy1003;
	goto yy55;
yy868:
	yych = *++cur;
	if (yych == 'C') goto yy1004;
	if (yych == 'c') goto yy1004;
	goto yy55;
yy869:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 438 "tokenizer.re"
	{ KW(NOPASS) }
#line 6138 "tokenizer.cpp"
yy871:
	yych = *++cur;
	if (yych == 'Y') goto yy1005;
	if (yych == 'y') goto yy1005;
	goto yy55;
yy872:
	yych = *++cur;
	if (yych == 'O') goto yy1007;
	if (yych == 'o') goto yy1007;
	goto yy55;
yy873:
	yych = *++cur;
	if (yych == 'A') goto yy1008;
	if (yych == 'a') goto yy1008;
	goto yy55;
yy874:
	yych = *++cur;
	if (yych == 'T') goto yy1009;
	if (yych == 't') goto yy1009;
	goto yy55;
yy875:
	yych = *++cur;
	if (yych == 'R') goto yy1010;
	if (yych == 'r') goto yy1010;
	goto yy55;
yy876:
	yych = *++cur;
	if (yych == 'I') goto yy1012;
	if (yych == 'i') goto yy1012;
	goto yy55;
yy877:
	yych = *++cur;
	if (yych == 'E') goto yy1013;
	if (yych == 'e') goto yy1013;
	goto yy55;
yy878:
	yych = *++cur;
	if (yych == 'U') goto yy1015;
	if (yych == 'u') goto yy1015;
	goto yy55;
yy879:
	yych = *++cur;
	if (yych == 'M') goto yy1016;
	if (yych == 'm') goto yy1016;
	goto yy55;
yy880:
	yych = *++cur;
	if (yych == 'T') goto yy1018;
	if (yych == 't') goto yy1018;
	goto yy55;
yy881:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 459 "tokenizer.re"
	{ KW(PUBLIC) }
#line 6197 "tokenizer.cpp"
yy883:
	yych = *++cur;
	if (yych == 'I') goto yy1019;
	if (yych == 'i') goto yy1019;
	goto yy55;
yy884:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 466 "tokenizer.re"
	{ KW(REDUCE) }
#line 6211 "tokenizer.cpp"
yy886:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 467 "tokenizer.re"
	{ KW(RESULT) }
#line 6220 "tokenizer.cpp"
yy888:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 468 "tokenizer.re"
	{ KW(RETURN) }
#line 6229 "tokenizer.cpp"
yy890:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 469 "tokenizer.re"
	{ KW(REWIND) }
#line 6238 "tokenizer.cpp"
yy892:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= 'B') {
			if (yych <= '/') goto yy893;
			if (yych <= '9') goto yy54;
			if (yych >= 'A') goto yy54;
		} else {
			if (yych <= 'Q') {
				if (yych <= 'C') goto yy1020;
				goto yy54;
			} else {
				if (yych <= 'R') goto yy1021;
				if (yych <= 'S') goto yy54;
				goto yy1022;
			}
		}
	} else {
		if (yych <= 'q') {
			if (yych <= '`') {
				if (yych <= 'Z') goto yy54;
			} else {
				if (yych == 'c') goto yy1020;
				goto yy54;
			}
		} else {
			if (yych <= 's') {
				if (yych <= 'r') goto yy1021;
				goto yy54;
			} else {
				if (yych <= 't') goto yy1022;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy893:
#line 471 "tokenizer.re"
	{ KW(SELECT) }
#line 6280 "tokenizer.cpp"
yy894:
	yych = *++cur;
	if (yych == 'C') goto yy1023;
	if (yych == 'c') goto yy1023;
	goto yy55;
yy895:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 476 "tokenizer.re"
	{ KW(SHARED) }
#line 6294 "tokenizer.cpp"
yy897:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 477 "tokenizer.re"
	{ KW(SOURCE) }
#line 6303 "tokenizer.cpp"
yy899:
	yych = *++cur;
	if (yych == 'U') goto yy1024;
	if (yych == 'u') goto yy1024;
	goto yy55;
yy900:
	yych = *++cur;
	if (yych == 'T') goto yy1025;
	if (yych == 't') goto yy1025;
	goto yy55;
yy901:
	yych = *++cur;
	if (yych == 'L') goto yy1026;
	if (yych == 'l') goto yy1026;
	goto yy55;
yy902:
	yych = *++cur;
	if (yych == 'A') goto yy1028;
	if (yych == 'a') goto yy1028;
	goto yy55;
yy903:
	yych = *++cur;
	if (yych == 'M') goto yy1029;
	if (yych == 'm') goto yy1029;
	goto yy55;
yy904:
	yych = *++cur;
	if (yych == 'A') goto yy1030;
	if (yych == 'a') goto yy1030;
	goto yy55;
yy905:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 487 "tokenizer.re"
	{ KW(TARGET) }
#line 6342 "tokenizer.cpp"
yy907:
	yych = *++cur;
	if (yych == 'U') goto yy1031;
	if (yych == 'u') goto yy1031;
	goto yy55;
yy908:
	yych = *++cur;
	if (yych == 'T') goto yy1032;
	if (yych == 't') goto yy1032;
	goto yy55;
yy909:
	yych = *++cur;
	if (yych == 'A') goto yy1033;
	if (yych == 'a') goto yy1033;
	goto yy55;
yy910:
	yych = *++cur;
	if (yych == 'L') goto yy1034;
	if (yych == 'l') goto yy1034;
	goto yy55;
yy911:
	yyaccept = 12;
	yych = *(mar = ++cur);
	if (yych == '_') goto yy1035;
yy912:
#line 580 "tokenizer.re"
	{ RET(TK_FALSE) }
#line 6370 "tokenizer.cpp"
yy913:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1036;
		if (yych <= '@') goto yy79;
		goto yy1038;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1038;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1038;
			goto yy79;
		}
	}
yy914:
	yych = *++cur;
	if (yych == 'T') goto yy1040;
	if (yych == 't') goto yy1040;
	goto yy55;
yy915:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy1042;
		if (yych <= 'D') goto yy55;
		goto yy1043;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy1042;
		} else {
			if (yych == 'e') goto yy1043;
			goto yy55;
		}
	}
yy916:
	yych = *++cur;
	if (yych == 'E') goto yy1045;
	if (yych == 'e') goto yy1045;
	goto yy55;
yy917:
	yych = *++cur;
	if (yych == 'T') goto yy1046;
	if (yych == 't') goto yy1046;
	goto yy55;
yy918:
	yych = *++cur;
	if (yych == 'O') goto yy1047;
	if (yych == 'o') goto yy1047;
	goto yy55;
yy919:
	yych = *++cur;
	if (yych == 'C') goto yy1048;
	if (yych == 'c') goto yy1048;
	goto yy55;
yy920:
	yych = *++cur;
	if (yych == 'E') goto yy1049;
	if (yych == 'e') goto yy1049;
	goto yy55;
yy921:
	yych = *++cur;
	if (yych == 'E') goto yy1050;
	if (yych == 'e') goto yy1050;
	goto yy55;
yy922:
	yych = *++cur;
	if (yych == 'S') goto yy1051;
	if (yych == 's') goto yy1051;
	goto yy55;
yy923:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 273 "tokenizer.re"
	{ KW(COMPLEX) }
#line 6451 "tokenizer.cpp"
yy925:
	yych = *++cur;
	if (yych == 'E') goto yy1052;
	if (yych == 'e') goto yy1052;
	goto yy55;
yy926:
	yych = *++cur;
	if (yych == 'S') goto yy1053;
	if (yych == 's') goto yy1053;
	goto yy55;
yy927:
	yych = *++cur;
	if (yych == 'O') goto yy1055;
	if (yych == 'o') goto yy1055;
	goto yy55;
yy928:
	yych = *++cur;
	if (yych == 'E') goto yy1056;
	if (yych == 'e') goto yy1056;
	goto yy55;
yy929:
	yych = *++cur;
	if (yych == 'L') goto yy1058;
	if (yych == 'l') goto yy1058;
	goto yy55;
yy930:
	yych = *++cur;
	if (yych == 'A') goto yy1060;
	if (yych == 'a') goto yy1060;
	goto yy55;
yy931:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 282 "tokenizer.re"
	{ KW(DEFAULT) }
#line 6490 "tokenizer.cpp"
yy933:
	yych = *++cur;
	if (yych == 'D') goto yy1061;
	if (yych == 'd') goto yy1061;
	goto yy55;
yy934:
	yych = *++cur;
	if (yych == 'O') goto yy1063;
	if (yych == 'o') goto yy1063;
	goto yy55;
yy935:
	yych = *++cur;
	if (yych == 'O') goto yy1064;
	if (yych == 'o') goto yy1064;
	goto yy55;
yy936:
	yych = *++cur;
	if (yych == 'R') goto yy1065;
	if (yych == 'r') goto yy1065;
	goto yy55;
yy937:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 293 "tokenizer.re"
	{ KW(DOWHILE) }
#line 6519 "tokenizer.cpp"
yy939:
	yych = *++cur;
	if (yych == 'A') goto yy1066;
	if (yych == 'a') goto yy1066;
	goto yy55;
yy940:
	yych = *++cur;
	if (yych == 'R') goto yy1067;
	if (yych == 'r') goto yy1067;
	goto yy55;
yy941:
	yych = *++cur;
	if (yych == 'O') goto yy1068;
	if (yych == 'o') goto yy1068;
	goto yy79;
yy942:
	yych = *++cur;
	if (yych == 'C') goto yy1069;
	if (yych == 'c') goto yy1069;
	goto yy79;
yy943:
	yych = *++cur;
	if (yych == 'T') goto yy1070;
	if (yych == 't') goto yy1070;
	goto yy79;
yy944:
	yych = *++cur;
	if (yych == 'M') goto yy1071;
	if (yych == 'm') goto yy1071;
	goto yy79;
yy945:
	yych = *++cur;
	if (yych == 'A') goto yy1073;
	if (yych == 'a') goto yy1073;
	goto yy79;
yy946:
	yych = *++cur;
	if (yych == 'C') goto yy1074;
	if (yych == 'c') goto yy1074;
	goto yy79;
yy947:
	yych = *++cur;
	if (yych == 'E') goto yy1075;
	if (yych == 'e') goto yy1075;
	goto yy79;
yy948:
	yych = *++cur;
	if (yych == 'U') goto yy1076;
	if (yych == 'u') goto yy1076;
	goto yy79;
yy949:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy1077;
		if (yych <= 'F') goto yy79;
		goto yy1078;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy79;
			goto yy1077;
		} else {
			if (yych == 'g') goto yy1078;
			goto yy79;
		}
	}
yy950:
	yych = *++cur;
	if (yych == 'E') goto yy1079;
	if (yych == 'e') goto yy1079;
	goto yy79;
yy951:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy1080;
		if (yych <= 'Q') goto yy79;
		goto yy1081;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy79;
			goto yy1080;
		} else {
			if (yych == 'r') goto yy1081;
			goto yy79;
		}
	}
yy952:
	yych = *++cur;
	if (yych == 'M') goto yy1082;
	if (yych == 'm') goto yy1082;
	goto yy79;
yy953:
	yych = *++cur;
	if (yych == 'E') goto yy1084;
	if (yych == 'e') goto yy1084;
	goto yy79;
yy954:
	yych = *++cur;
	if (yych == 'R') goto yy1086;
	if (yych == 'r') goto yy1086;
	goto yy79;
yy955:
	yych = *++cur;
	if (yych == 'E') goto yy1087;
	if (yych == 'e') goto yy1087;
	goto yy79;
yy956:
	yych = *++cur;
	if (yych == 'C') goto yy1089;
	if (yych == 'c') goto yy1089;
	goto yy55;
yy957:
	yych = *++cur;
	if (yych == 'K') goto yy1090;
	if (yych == 'k') goto yy1090;
	goto yy55;
yy958:
	yych = *++cur;
	if (yych == 'I') goto yy1092;
	if (yych == 'i') goto yy1092;
	goto yy55;
yy959:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 331 "tokenizer.re"
	{ KW(ENDENUM) }
#line 6648 "tokenizer.cpp"
yy961:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 376 "tokenizer.re"
	{ KW(ENDFILE) }
#line 6657 "tokenizer.cpp"
yy963:
	yych = *++cur;
	if (yych == 'L') goto yy1093;
	if (yych == 'l') goto yy1093;
	goto yy55;
yy964:
	yych = *++cur;
	if (yych == 'T') goto yy1094;
	if (yych == 't') goto yy1094;
	goto yy55;
yy965:
	yych = *++cur;
	if (yych == 'R') goto yy1095;
	if (yych == 'r') goto yy1095;
	goto yy55;
yy966:
	yych = *++cur;
	if (yych == 'L') goto yy1096;
	if (yych == 'l') goto yy1096;
	goto yy55;
yy967:
	yych = *++cur;
	if (yych == 'E') goto yy1097;
	if (yych == 'e') goto yy1097;
	goto yy55;
yy968:
	yych = *++cur;
	if (yych == 'R') goto yy1098;
	if (yych == 'r') goto yy1098;
	goto yy55;
yy969:
	yych = *++cur;
	if (yych == 'C') goto yy1099;
	if (yych == 'c') goto yy1099;
	goto yy55;
yy970:
	yych = *++cur;
	if (yych == 'O') goto yy1100;
	if (yych == 'o') goto yy1100;
	goto yy55;
yy971:
	yych = *++cur;
	if (yych == 'O') goto yy1101;
	if (yych == 'o') goto yy1101;
	goto yy55;
yy972:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 343 "tokenizer.re"
	{ KW(ENDTEAM) }
#line 6711 "tokenizer.cpp"
yy974:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 355 "tokenizer.re"
	{ KW(ENDTYPE) }
#line 6720 "tokenizer.cpp"
yy976:
	yych = *++cur;
	if (yych == 'E') goto yy1102;
	if (yych == 'e') goto yy1102;
	goto yy55;
yy977:
	yych = *++cur;
	if (yych == 'T') goto yy1104;
	if (yych == 't') goto yy1104;
	goto yy55;
yy978:
	yych = *++cur;
	if (yych == 'E') goto yy1105;
	if (yych == 'e') goto yy1105;
	goto yy55;
yy979:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 386 "tokenizer.re"
	{ KW(EXTENDS) }
#line 6744 "tokenizer.cpp"
yy981:
	yych = *++cur;
	if (yych == 'L') goto yy1106;
	if (yych == 'l') goto yy1106;
	goto yy55;
yy982:
	yych = *++cur;
	if (yych == 'E') goto yy1108;
	if (yych == 'e') goto yy1108;
	goto yy55;
yy983:
	yych = *++cur;
	if (yych == 'M') goto yy1109;
	if (yych == 'm') goto yy1109;
	goto yy55;
yy984:
	yych = *++cur;
	if (yych == 'N') goto yy1111;
	if (yych == 'n') goto yy1111;
	goto yy55;
yy985:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 408 "tokenizer.re"
	{ KW(GENERIC) }
#line 6773 "tokenizer.cpp"
yy987:
	yych = *++cur;
	if (yych == 'T') goto yy1113;
	if (yych == 't') goto yy1113;
	goto yy55;
yy988:
	yyaccept = 13;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych >= '\t') goto yy1115;
		} else {
			if (yych == '\v') goto yy1115;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1115;
			if (yych >= ' ') goto yy1115;
		} else {
			if (yych == '_') goto yy144;
		}
	}
yy989:
#line 417 "tokenizer.re"
	{ KW(INCLUDE) }
#line 6802 "tokenizer.cpp"
yy990:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 420 "tokenizer.re"
	{ KW(INQUIRE) }
#line 6811 "tokenizer.cpp"
yy992:
	yych = *++cur;
	if (yych == 'I') goto yy1117;
	if (yych == 'i') goto yy1117;
	goto yy55;
yy993:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 422 "tokenizer.re"
	{ KW(INTEGER) }
#line 6825 "tokenizer.cpp"
yy995:
	yych = *++cur;
	if (yych == 'C') goto yy1118;
	if (yych == 'c') goto yy1118;
	goto yy55;
yy996:
	yych = *++cur;
	if (yych == 'I') goto yy1119;
	if (yych == 'i') goto yy1119;
	goto yy55;
yy997:
	yych = *++cur;
	if (yych == 'N') goto yy1120;
	if (yych == 'n') goto yy1120;
	goto yy55;
yy998:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 431 "tokenizer.re"
	{ KW(LOGICAL) }
#line 6849 "tokenizer.cpp"
yy1000:
	yych = *++cur;
	if (yych == 'T') goto yy1121;
	if (yych == 't') goto yy1121;
	goto yy55;
yy1001:
	yych = *++cur;
	if (yych == 'E') goto yy1123;
	if (yych == 'e') goto yy1123;
	goto yy55;
yy1002:
	yych = *++cur;
	if (yych == 'R') goto yy1124;
	if (yych == 'r') goto yy1124;
	goto yy55;
yy1003:
	yych = *++cur;
	if (yych == 'R') goto yy1125;
	if (yych == 'r') goto yy1125;
	goto yy55;
yy1004:
	yych = *++cur;
	if (yych == 'U') goto yy1126;
	if (yych == 'u') goto yy1126;
	goto yy55;
yy1005:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 443 "tokenizer.re"
	{ KW(NULLIFY) }
#line 6883 "tokenizer.cpp"
yy1007:
	yych = *++cur;
	if (yych == 'R') goto yy1127;
	if (yych == 'r') goto yy1127;
	goto yy55;
yy1008:
	yych = *++cur;
	if (yych == 'L') goto yy1129;
	if (yych == 'l') goto yy1129;
	goto yy55;
yy1009:
	yych = *++cur;
	if (yych == 'E') goto yy1131;
	if (yych == 'e') goto yy1131;
	goto yy55;
yy1010:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 451 "tokenizer.re"
	{ KW(POINTER) }
#line 6907 "tokenizer.cpp"
yy1012:
	yych = *++cur;
	if (yych == 'O') goto yy1132;
	if (yych == 'o') goto yy1132;
	goto yy55;
yy1013:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 455 "tokenizer.re"
	{ KW(PRIVATE) }
#line 6921 "tokenizer.cpp"
yy1015:
	yych = *++cur;
	if (yych == 'R') goto yy1133;
	if (yych == 'r') goto yy1133;
	goto yy55;
yy1016:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 457 "tokenizer.re"
	{ KW(PROGRAM) }
#line 6935 "tokenizer.cpp"
yy1018:
	yych = *++cur;
	if (yych == 'E') goto yy1134;
	if (yych == 'e') goto yy1134;
	goto yy55;
yy1019:
	yych = *++cur;
	if (yych == 'V') goto yy1135;
	if (yych == 'v') goto yy1135;
	goto yy55;
yy1020:
	yych = *++cur;
	if (yych == 'A') goto yy1136;
	if (yych == 'a') goto yy1136;
	goto yy55;
yy1021:
	yych = *++cur;
	if (yych == 'A') goto yy1137;
	if (yych == 'a') goto yy1137;
	goto yy55;
yy1022:
	yych = *++cur;
	if (yych == 'Y') goto yy1138;
	if (yych == 'y') goto yy1138;
	goto yy55;
yy1023:
	yych = *++cur;
	if (yych == 'E') goto yy1139;
	if (yych == 'e') goto yy1139;
	goto yy55;
yy1024:
	yych = *++cur;
	if (yych == 'L') goto yy1141;
	if (yych == 'l') goto yy1141;
	goto yy55;
yy1025:
	yych = *++cur;
	if (yych == 'I') goto yy1142;
	if (yych == 'i') goto yy1142;
	goto yy55;
yy1026:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 483 "tokenizer.re"
	{ KW(SYNC_ALL) }
#line 6984 "tokenizer.cpp"
yy1028:
	yych = *++cur;
	if (yych == 'G') goto yy1143;
	if (yych == 'g') goto yy1143;
	goto yy55;
yy1029:
	yych = *++cur;
	if (yych == 'O') goto yy1144;
	if (yych == 'o') goto yy1144;
	goto yy55;
yy1030:
	yych = *++cur;
	if (yych == 'M') goto yy1145;
	if (yych == 'm') goto yy1145;
	goto yy55;
yy1031:
	yych = *++cur;
	if (yych == 'M') goto yy1147;
	if (yych == 'm') goto yy1147;
	goto yy55;
yy1032:
	yych = *++cur;
	if (yych == 'E') goto yy1148;
	if (yych == 'e') goto yy1148;
	goto yy55;
yy1033:
	yych = *++cur;
	if (yych == 'T') goto yy1150;
	if (yych == 't') goto yy1150;
	goto yy55;
yy1034:
	yych = *++cur;
	if (yych == 'E') goto yy1151;
	if (yych == 'e') goto yy1151;
	goto yy55;
yy1035:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1153;
		if (yych <= '@') goto yy79;
		goto yy1155;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1155;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1155;
			goto yy79;
		}
	}
yy1036:
	yych = *++cur;
	if (yych <= '/') goto yy763;
	if (yych <= '9') goto yy1036;
	goto yy763;
yy1038:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy763;
		if (yych <= '9') goto yy1038;
		if (yych <= '@') goto yy763;
		goto yy1038;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy763;
			goto yy1038;
		} else {
			if (yych <= '`') goto yy763;
			if (yych <= 'z') goto yy1038;
			goto yy763;
		}
	}
yy1040:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 253 "tokenizer.re"
	{ KW(ABSTRACT) }
#line 7067 "tokenizer.cpp"
yy1042:
	yych = *++cur;
	if (yych == 'B') goto yy1157;
	if (yych == 'b') goto yy1157;
	goto yy55;
yy1043:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 256 "tokenizer.re"
	{ KW(ALLOCATE) }
#line 7081 "tokenizer.cpp"
yy1045:
	yych = *++cur;
	if (yych == 'N') goto yy1158;
	if (yych == 'n') goto yy1158;
	goto yy55;
yy1046:
	yych = *++cur;
	if (yych == 'E') goto yy1159;
	if (yych == 'e') goto yy1159;
	goto yy55;
yy1047:
	yych = *++cur;
	if (yych == 'N') goto yy1161;
	if (yych == 'n') goto yy1161;
	goto yy55;
yy1048:
	yych = *++cur;
	if (yych == 'E') goto yy1162;
	if (yych == 'e') goto yy1162;
	goto yy55;
yy1049:
	yych = *++cur;
	if (yych == 'A') goto yy1164;
	if (yych == 'a') goto yy1164;
	goto yy55;
yy1050:
	yych = *++cur;
	if (yych == 'R') goto yy1165;
	if (yych == 'r') goto yy1165;
	goto yy55;
yy1051:
	yych = *++cur;
	if (yych == 'I') goto yy1167;
	if (yych == 'i') goto yy1167;
	goto yy55;
yy1052:
	yych = *++cur;
	if (yych == 'N') goto yy1168;
	if (yych == 'n') goto yy1168;
	goto yy55;
yy1053:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 275 "tokenizer.re"
	{ KW(CONTAINS) }
#line 7130 "tokenizer.cpp"
yy1055:
	yych = *++cur;
	if (yych == 'U') goto yy1169;
	if (yych == 'u') goto yy1169;
	goto yy55;
yy1056:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 277 "tokenizer.re"
	{ KW(CONTINUE) }
#line 7144 "tokenizer.cpp"
yy1058:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 278 "tokenizer.re"
	{ KW(CRITICAL) }
#line 7153 "tokenizer.cpp"
yy1060:
	yych = *++cur;
	if (yych == 'T') goto yy1170;
	if (yych == 't') goto yy1170;
	goto yy55;
yy1061:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 283 "tokenizer.re"
	{ KW(DEFERRED) }
#line 7167 "tokenizer.cpp"
yy1063:
	yych = *++cur;
	if (yych == 'N') goto yy1171;
	if (yych == 'n') goto yy1171;
	goto yy55;
yy1064:
	yych = *++cur;
	if (yych == 'M') goto yy1173;
	if (yych == 'm') goto yy1173;
	goto yy55;
yy1065:
	yych = *++cur;
	if (yych == 'E') goto yy1174;
	if (yych == 'e') goto yy1174;
	goto yy55;
yy1066:
	yych = *++cur;
	if (yych == 'L') goto yy1175;
	if (yych == 'l') goto yy1175;
	goto yy55;
yy1067:
	yych = *++cur;
	if (yych == 'E') goto yy1177;
	if (yych == 'e') goto yy1177;
	goto yy55;
yy1068:
	yych = *++cur;
	if (yych == 'C') goto yy1179;
	if (yych == 'c') goto yy1179;
	goto yy79;
yy1069:
	yych = *++cur;
	if (yych == 'K') goto yy1180;
	if (yych == 'k') goto yy1180;
	goto yy79;
yy1070:
	yych = *++cur;
	if (yych == 'I') goto yy1182;
	if (yych == 'i') goto yy1182;
	goto yy79;
yy1071:
	++cur;
#line 330 "tokenizer.re"
	{ KW(END_ENUM) }
#line 7212 "tokenizer.cpp"
yy1073:
	yych = *++cur;
	if (yych == 'L') goto yy1183;
	if (yych == 'l') goto yy1183;
	goto yy79;
yy1074:
	yych = *++cur;
	if (yych == 'T') goto yy1184;
	if (yych == 't') goto yy1184;
	goto yy79;
yy1075:
	yych = *++cur;
	if (yych == 'R') goto yy1185;
	if (yych == 'r') goto yy1185;
	goto yy79;
yy1076:
	yych = *++cur;
	if (yych == 'L') goto yy1186;
	if (yych == 'l') goto yy1186;
	goto yy79;
yy1077:
	yych = *++cur;
	if (yych == 'E') goto yy1187;
	if (yych == 'e') goto yy1187;
	goto yy79;
yy1078:
	yych = *++cur;
	if (yych == 'R') goto yy1188;
	if (yych == 'r') goto yy1188;
	goto yy79;
yy1079:
	yych = *++cur;
	if (yych == 'C') goto yy1189;
	if (yych == 'c') goto yy1189;
	goto yy79;
yy1080:
	yych = *++cur;
	if (yych == 'O') goto yy1190;
	if (yych == 'o') goto yy1190;
	goto yy79;
yy1081:
	yych = *++cur;
	if (yych == 'O') goto yy1191;
	if (yych == 'o') goto yy1191;
	goto yy79;
yy1082:
	++cur;
#line 342 "tokenizer.re"
	{ KW(END_TEAM) }
#line 7262 "tokenizer.cpp"
yy1084:
	++cur;
#line 354 "tokenizer.re"
	{ KW(END_TYPE) }
#line 7267 "tokenizer.cpp"
yy1086:
	yych = *++cur;
	if (yych == 'E') goto yy1192;
	if (yych == 'e') goto yy1192;
	goto yy79;
yy1087:
	++cur;
#line 375 "tokenizer.re"
	{ KW(END_FILE) }
#line 7277 "tokenizer.cpp"
yy1089:
	yych = *++cur;
	if (yych == 'I') goto yy1194;
	if (yych == 'i') goto yy1194;
	goto yy55;
yy1090:
	yyaccept = 14;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= '/') {
		if (yych <= '\v') {
			if (yych == '\t') goto yy1195;
			if (yych >= '\v') goto yy1195;
		} else {
			if (yych <= '\r') {
				if (yych >= '\r') goto yy1195;
			} else {
				if (yych == ' ') goto yy1195;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') {
				if (yych <= '9') goto yy54;
			} else {
				if (yych == 'D') goto yy1197;
				goto yy54;
			}
		} else {
			if (yych <= 'c') {
				if (yych >= 'a') goto yy54;
			} else {
				if (yych <= 'd') goto yy1197;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy1091:
#line 314 "tokenizer.re"
	{ KW(ENDBLOCK) }
#line 7320 "tokenizer.cpp"
yy1092:
	yych = *++cur;
	if (yych == 'C') goto yy1198;
	if (yych == 'c') goto yy1198;
	goto yy55;
yy1093:
	yych = *++cur;
	if (yych == 'L') goto yy1199;
	if (yych == 'l') goto yy1199;
	goto yy55;
yy1094:
	yych = *++cur;
	if (yych == 'I') goto yy1201;
	if (yych == 'i') goto yy1201;
	goto yy55;
yy1095:
	yych = *++cur;
	if (yych == 'F') goto yy1202;
	if (yych == 'f') goto yy1202;
	goto yy55;
yy1096:
	yych = *++cur;
	if (yych == 'E') goto yy1203;
	if (yych == 'e') goto yy1203;
	goto yy55;
yy1097:
	yych = *++cur;
	if (yych == 'D') goto yy1205;
	if (yych == 'd') goto yy1205;
	goto yy55;
yy1098:
	yych = *++cur;
	if (yych == 'A') goto yy1206;
	if (yych == 'a') goto yy1206;
	goto yy55;
yy1099:
	yych = *++cur;
	if (yych == 'T') goto yy1207;
	if (yych == 't') goto yy1207;
	goto yy55;
yy1100:
	yych = *++cur;
	if (yych == 'D') goto yy1209;
	if (yych == 'd') goto yy1209;
	goto yy55;
yy1101:
	yych = *++cur;
	if (yych == 'U') goto yy1210;
	if (yych == 'u') goto yy1210;
	goto yy55;
yy1102:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 373 "tokenizer.re"
	{ KW(ENDWHERE) }
#line 7379 "tokenizer.cpp"
yy1104:
	yych = *++cur;
	if (yych == 'O') goto yy1211;
	if (yych == 'o') goto yy1211;
	goto yy55;
yy1105:
	yych = *++cur;
	if (yych == 'N') goto yy1212;
	if (yych == 'n') goto yy1212;
	goto yy55;
yy1106:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 387 "tokenizer.re"
	{ KW(EXTERNAL) }
#line 7398 "tokenizer.cpp"
yy1108:
	yych = *++cur;
	if (yych == 'D') goto yy1213;
	if (yych == 'd') goto yy1213;
	goto yy55;
yy1109:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 406 "tokenizer.re"
	{ KW(FORM_TEAM) }
#line 7412 "tokenizer.cpp"
yy1111:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 407 "tokenizer.re"
	{ KW(FUNCTION) }
#line 7421 "tokenizer.cpp"
yy1113:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 413 "tokenizer.re"
	{ KW(IMPLICIT) }
#line 7430 "tokenizer.cpp"
yy1115:
	yych = *++cur;
	if (yych <= '"') {
		if (yych <= '\f') {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy79;
				goto yy1115;
			} else {
				if (yych == '\v') goto yy1115;
				goto yy79;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy1115;
				goto yy79;
			} else {
				if (yych <= ' ') goto yy1115;
				if (yych <= '!') goto yy79;
				goto yy1215;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '\'') {
				if (yych <= '&') goto yy79;
				goto yy1217;
			} else {
				if (yych <= '/') goto yy79;
				if (yych <= '9') goto yy1219;
				goto yy79;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1221;
				if (yych <= '^') goto yy79;
				goto yy1221;
			} else {
				if (yych <= '`') goto yy79;
				if (yych <= 'z') goto yy1221;
				goto yy79;
			}
		}
	}
yy1117:
	yych = *++cur;
	if (yych == 'A') goto yy1223;
	if (yych == 'a') goto yy1223;
	goto yy55;
yy1118:
	yych = *++cur;
	if (yych == 'E') goto yy1224;
	if (yych == 'e') goto yy1224;
	goto yy55;
yy1119:
	yych = *++cur;
	if (yych == 'C') goto yy1226;
	if (yych == 'c') goto yy1226;
	goto yy55;
yy1120:
	yych = *++cur;
	if (yych == 'I') goto yy1228;
	if (yych == 'i') goto yy1228;
	goto yy55;
yy1121:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 436 "tokenizer.re"
	{ KW(NAMELIST) }
#line 7502 "tokenizer.cpp"
yy1123:
	yych = *++cur;
	if (yych == 'X') goto yy1229;
	if (yych == 'x') goto yy1229;
	goto yy55;
yy1124:
	yych = *++cur;
	if (yych == 'I') goto yy1231;
	if (yych == 'i') goto yy1231;
	goto yy55;
yy1125:
	yych = *++cur;
	if (yych == 'R') goto yy1232;
	if (yych == 'r') goto yy1232;
	goto yy55;
yy1126:
	yych = *++cur;
	if (yych == 'R') goto yy1233;
	if (yych == 'r') goto yy1233;
	goto yy55;
yy1127:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 446 "tokenizer.re"
	{ KW(OPERATOR) }
#line 7531 "tokenizer.cpp"
yy1129:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 447 "tokenizer.re"
	{ KW(OPTIONAL) }
#line 7540 "tokenizer.cpp"
yy1131:
	yych = *++cur;
	if (yych == 'R') goto yy1234;
	if (yych == 'r') goto yy1234;
	goto yy55;
yy1132:
	yych = *++cur;
	if (yych == 'N') goto yy1236;
	if (yych == 'n') goto yy1236;
	goto yy55;
yy1133:
	yych = *++cur;
	if (yych == 'E') goto yy1238;
	if (yych == 'e') goto yy1238;
	goto yy55;
yy1134:
	yych = *++cur;
	if (yych == 'D') goto yy1240;
	if (yych == 'd') goto yy1240;
	goto yy55;
yy1135:
	yych = *++cur;
	if (yych == 'E') goto yy1242;
	if (yych == 'e') goto yy1242;
	goto yy55;
yy1136:
	yych = *++cur;
	if (yych == 'S') goto yy1244;
	if (yych == 's') goto yy1244;
	goto yy55;
yy1137:
	yych = *++cur;
	if (yych == 'N') goto yy1245;
	if (yych == 'n') goto yy1245;
	goto yy55;
yy1138:
	yych = *++cur;
	if (yych == 'P') goto yy1246;
	if (yych == 'p') goto yy1246;
	goto yy55;
yy1139:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 475 "tokenizer.re"
	{ KW(SEQUENCE) }
#line 7589 "tokenizer.cpp"
yy1141:
	yych = *++cur;
	if (yych == 'E') goto yy1247;
	if (yych == 'e') goto yy1247;
	goto yy55;
yy1142:
	yych = *++cur;
	if (yych == 'N') goto yy1249;
	if (yych == 'n') goto yy1249;
	goto yy55;
yy1143:
	yych = *++cur;
	if (yych == 'E') goto yy1250;
	if (yych == 'e') goto yy1250;
	goto yy55;
yy1144:
	yych = *++cur;
	if (yych == 'R') goto yy1251;
	if (yych == 'r') goto yy1251;
	goto yy55;
yy1145:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 486 "tokenizer.re"
	{ KW(SYNC_TEAM) }
#line 7618 "tokenizer.cpp"
yy1147:
	yych = *++cur;
	if (yych == 'B') goto yy1252;
	if (yych == 'b') goto yy1252;
	goto yy55;
yy1148:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 490 "tokenizer.re"
	{ KW(TEMPLATE) }
#line 7632 "tokenizer.cpp"
yy1150:
	yych = *++cur;
	if (yych == 'T') goto yy1253;
	if (yych == 't') goto yy1253;
	goto yy55;
yy1151:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 497 "tokenizer.re"
	{ KW(VOLATILE) }
#line 7646 "tokenizer.cpp"
yy1153:
	yych = *++cur;
	if (yych <= '/') goto yy912;
	if (yych <= '9') goto yy1153;
	goto yy912;
yy1155:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy912;
		if (yych <= '9') goto yy1155;
		if (yych <= '@') goto yy912;
		goto yy1155;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy912;
			goto yy1155;
		} else {
			if (yych <= '`') goto yy912;
			if (yych <= 'z') goto yy1155;
			goto yy912;
		}
	}
yy1157:
	yych = *++cur;
	if (yych == 'L') goto yy1254;
	if (yych == 'l') goto yy1254;
	goto yy55;
yy1158:
	yych = *++cur;
	if (yych == 'T') goto yy1255;
	if (yych == 't') goto yy1255;
	goto yy55;
yy1159:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 259 "tokenizer.re"
	{ KW(ASSOCIATE) }
#line 7687 "tokenizer.cpp"
yy1161:
	yych = *++cur;
	if (yych == 'O') goto yy1257;
	if (yych == 'o') goto yy1257;
	goto yy55;
yy1162:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 261 "tokenizer.re"
	{ KW(BACKSPACE) }
#line 7701 "tokenizer.cpp"
yy1164:
	yych = *++cur;
	if (yych == 'M') goto yy1258;
	if (yych == 'm') goto yy1258;
	goto yy55;
yy1165:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 268 "tokenizer.re"
	{ KW(CHARACTER) }
#line 7715 "tokenizer.cpp"
yy1167:
	yych = *++cur;
	if (yych == 'O') goto yy1260;
	if (yych == 'o') goto yy1260;
	goto yy55;
yy1168:
	yych = *++cur;
	if (yych == 'T') goto yy1261;
	if (yych == 't') goto yy1261;
	goto yy55;
yy1169:
	yych = *++cur;
	if (yych == 'S') goto yy1263;
	if (yych == 's') goto yy1263;
	goto yy55;
yy1170:
	yych = *++cur;
	if (yych == 'E') goto yy1265;
	if (yych == 'e') goto yy1265;
	goto yy55;
yy1171:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 284 "tokenizer.re"
	{ KW(DIMENSION) }
#line 7744 "tokenizer.cpp"
yy1173:
	yych = *++cur;
	if (yych == 'P') goto yy1267;
	if (yych == 'p') goto yy1267;
	goto yy55;
yy1174:
	yych = *++cur;
	if (yych == 'C') goto yy1268;
	if (yych == 'c') goto yy1268;
	goto yy55;
yy1175:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 297 "tokenizer.re"
	{ KW(ELEMENTAL) }
#line 7763 "tokenizer.cpp"
yy1177:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 300 "tokenizer.re"
	{ KW(ELSEWHERE) }
#line 7772 "tokenizer.cpp"
yy1179:
	yych = *++cur;
	if (yych == 'I') goto yy1269;
	if (yych == 'i') goto yy1269;
	goto yy79;
yy1180:
	yyaccept = 15;
	yych = *(mar = ++cur);
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1270;
		} else {
			if (yych != '\f') goto yy1270;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1270;
		} else {
			if (yych <= 'D') goto yy1272;
			if (yych == 'd') goto yy1272;
		}
	}
yy1181:
#line 313 "tokenizer.re"
	{ KW(END_BLOCK) }
#line 7798 "tokenizer.cpp"
yy1182:
	yych = *++cur;
	if (yych == 'C') goto yy1273;
	if (yych == 'c') goto yy1273;
	goto yy79;
yy1183:
	yych = *++cur;
	if (yych == 'L') goto yy1274;
	if (yych == 'l') goto yy1274;
	goto yy79;
yy1184:
	yych = *++cur;
	if (yych == 'I') goto yy1276;
	if (yych == 'i') goto yy1276;
	goto yy79;
yy1185:
	yych = *++cur;
	if (yych == 'F') goto yy1277;
	if (yych == 'f') goto yy1277;
	goto yy79;
yy1186:
	yych = *++cur;
	if (yych == 'E') goto yy1278;
	if (yych == 'e') goto yy1278;
	goto yy79;
yy1187:
	yych = *++cur;
	if (yych == 'D') goto yy1280;
	if (yych == 'd') goto yy1280;
	goto yy79;
yy1188:
	yych = *++cur;
	if (yych == 'A') goto yy1281;
	if (yych == 'a') goto yy1281;
	goto yy79;
yy1189:
	yych = *++cur;
	if (yych == 'T') goto yy1282;
	if (yych == 't') goto yy1282;
	goto yy79;
yy1190:
	yych = *++cur;
	if (yych == 'D') goto yy1284;
	if (yych == 'd') goto yy1284;
	goto yy79;
yy1191:
	yych = *++cur;
	if (yych == 'U') goto yy1285;
	if (yych == 'u') goto yy1285;
	goto yy79;
yy1192:
	++cur;
#line 372 "tokenizer.re"
	{ KW(END_WHERE) }
#line 7853 "tokenizer.cpp"
yy1194:
	yych = *++cur;
	if (yych == 'A') goto yy1286;
	if (yych == 'a') goto yy1286;
	goto yy55;
yy1195:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1195;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy1195;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1195;
			goto yy79;
		} else {
			if (yych <= 'D') goto yy1287;
			if (yych == 'd') goto yy1287;
			goto yy79;
		}
	}
yy1197:
	yych = *++cur;
	if (yych == 'A') goto yy1288;
	if (yych == 'a') goto yy1288;
	goto yy55;
yy1198:
	yych = *++cur;
	if (yych == 'A') goto yy1289;
	if (yych == 'a') goto yy1289;
	goto yy55;
yy1199:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 346 "tokenizer.re"
	{ KW(ENDFORALL) }
#line 7897 "tokenizer.cpp"
yy1201:
	yych = *++cur;
	if (yych == 'O') goto yy1290;
	if (yych == 'o') goto yy1290;
	goto yy55;
yy1202:
	yych = *++cur;
	if (yych == 'A') goto yy1291;
	if (yych == 'a') goto yy1291;
	goto yy55;
yy1203:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 308 "tokenizer.re"
	{ KW(ENDMODULE) }
#line 7916 "tokenizer.cpp"
yy1205:
	yych = *++cur;
	if (yych == 'U') goto yy1292;
	if (yych == 'u') goto yy1292;
	goto yy55;
yy1206:
	yych = *++cur;
	if (yych == 'M') goto yy1293;
	if (yych == 'm') goto yy1293;
	goto yy55;
yy1207:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 334 "tokenizer.re"
	{ KW(ENDSELECT) }
#line 7935 "tokenizer.cpp"
yy1209:
	yych = *++cur;
	if (yych == 'U') goto yy1295;
	if (yych == 'u') goto yy1295;
	goto yy55;
yy1210:
	yych = *++cur;
	if (yych == 'T') goto yy1296;
	if (yych == 't') goto yy1296;
	goto yy55;
yy1211:
	yych = *++cur;
	if (yych == 'R') goto yy1297;
	if (yych == 'r') goto yy1297;
	goto yy55;
yy1212:
	yych = *++cur;
	if (yych == 'C') goto yy1299;
	if (yych == 'c') goto yy1299;
	goto yy55;
yy1213:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 404 "tokenizer.re"
	{ KW(FORMATTED) }
#line 7964 "tokenizer.cpp"
yy1215:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '"') goto yy1300;
	goto yy1215;
yy1217:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '\'') goto yy1302;
	goto yy1217;
yy1219:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '9') goto yy1219;
	if (yych == '_') goto yy1304;
	goto yy79;
yy1221:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1221;
		if (yych <= '@') goto yy79;
		goto yy1221;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1305;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1221;
			goto yy79;
		}
	}
yy1223:
	yych = *++cur;
	if (yych == 'T') goto yy1307;
	if (yych == 't') goto yy1307;
	goto yy55;
yy1224:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 424 "tokenizer.re"
	{ KW(INTERFACE) }
#line 8011 "tokenizer.cpp"
yy1226:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 425 "tokenizer.re"
	{ KW(INTRINSIC) }
#line 8020 "tokenizer.cpp"
yy1228:
	yych = *++cur;
	if (yych == 'T') goto yy1308;
	if (yych == 't') goto yy1308;
	goto yy55;
yy1229:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 437 "tokenizer.re"
	{ KW(NEW_INDEX) }
#line 8034 "tokenizer.cpp"
yy1231:
	yych = *++cur;
	if (yych == 'N') goto yy1310;
	if (yych == 'n') goto yy1310;
	goto yy55;
yy1232:
	yych = *++cur;
	if (yych == 'I') goto yy1311;
	if (yych == 'i') goto yy1311;
	goto yy55;
yy1233:
	yych = *++cur;
	if (yych == 'S') goto yy1312;
	if (yych == 's') goto yy1312;
	goto yy55;
yy1234:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 449 "tokenizer.re"
	{ KW(PARAMETER) }
#line 8058 "tokenizer.cpp"
yy1236:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 453 "tokenizer.re"
	{ KW(PRECISION) }
#line 8067 "tokenizer.cpp"
yy1238:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 456 "tokenizer.re"
	{ KW(PROCEDURE) }
#line 8076 "tokenizer.cpp"
yy1240:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 458 "tokenizer.re"
	{ KW(PROTECTED) }
#line 8085 "tokenizer.cpp"
yy1242:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 465 "tokenizer.re"
	{ KW(RECURSIVE) }
#line 8094 "tokenizer.cpp"
yy1244:
	yych = *++cur;
	if (yych == 'E') goto yy1313;
	if (yych == 'e') goto yy1313;
	goto yy55;
yy1245:
	yych = *++cur;
	if (yych == 'K') goto yy1315;
	if (yych == 'k') goto yy1315;
	goto yy55;
yy1246:
	yych = *++cur;
	if (yych == 'E') goto yy1317;
	if (yych == 'e') goto yy1317;
	goto yy55;
yy1247:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 480 "tokenizer.re"
	{ KW(SUBMODULE) }
#line 8118 "tokenizer.cpp"
yy1249:
	yych = *++cur;
	if (yych == 'E') goto yy1319;
	if (yych == 'e') goto yy1319;
	goto yy55;
yy1250:
	yych = *++cur;
	if (yych == 'S') goto yy1321;
	if (yych == 's') goto yy1321;
	goto yy55;
yy1251:
	yych = *++cur;
	if (yych == 'Y') goto yy1323;
	if (yych == 'y') goto yy1323;
	goto yy55;
yy1252:
	yych = *++cur;
	if (yych == 'E') goto yy1325;
	if (yych == 'e') goto yy1325;
	goto yy55;
yy1253:
	yych = *++cur;
	if (yych == 'E') goto yy1326;
	if (yych == 'e') goto yy1326;
	goto yy55;
yy1254:
	yych = *++cur;
	if (yych == 'E') goto yy1327;
	if (yych == 'e') goto yy1327;
	goto yy55;
yy1255:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 258 "tokenizer.re"
	{ KW(ASSIGNMENT) }
#line 8157 "tokenizer.cpp"
yy1257:
	yych = *++cur;
	if (yych == 'U') goto yy1329;
	if (yych == 'u') goto yy1329;
	goto yy55;
yy1258:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 267 "tokenizer.re"
	{ KW(CHANGE_TEAM) }
#line 8171 "tokenizer.cpp"
yy1260:
	yych = *++cur;
	if (yych == 'N') goto yy1330;
	if (yych == 'n') goto yy1330;
	goto yy55;
yy1261:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 274 "tokenizer.re"
	{ KW(CONCURRENT) }
#line 8185 "tokenizer.cpp"
yy1263:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 276 "tokenizer.re"
	{ KW(CONTIGUOUS) }
#line 8194 "tokenizer.cpp"
yy1265:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 281 "tokenizer.re"
	{ KW(DEALLOCATE) }
#line 8203 "tokenizer.cpp"
yy1267:
	yych = *++cur;
	if (yych == 'L') goto yy1332;
	if (yych == 'l') goto yy1332;
	goto yy55;
yy1268:
	yych = *++cur;
	if (yych == 'I') goto yy1333;
	if (yych == 'i') goto yy1333;
	goto yy55;
yy1269:
	yych = *++cur;
	if (yych == 'A') goto yy1334;
	if (yych == 'a') goto yy1334;
	goto yy79;
yy1270:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1270;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy1270;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1270;
			goto yy79;
		} else {
			if (yych <= 'D') goto yy1335;
			if (yych == 'd') goto yy1335;
			goto yy79;
		}
	}
yy1272:
	yych = *++cur;
	if (yych == 'A') goto yy1336;
	if (yych == 'a') goto yy1336;
	goto yy79;
yy1273:
	yych = *++cur;
	if (yych == 'A') goto yy1337;
	if (yych == 'a') goto yy1337;
	goto yy79;
yy1274:
	++cur;
#line 345 "tokenizer.re"
	{ KW(END_FORALL) }
#line 8253 "tokenizer.cpp"
yy1276:
	yych = *++cur;
	if (yych == 'O') goto yy1338;
	if (yych == 'o') goto yy1338;
	goto yy79;
yy1277:
	yych = *++cur;
	if (yych == 'A') goto yy1339;
	if (yych == 'a') goto yy1339;
	goto yy79;
yy1278:
	++cur;
#line 307 "tokenizer.re"
	{ KW(END_MODULE) }
#line 8268 "tokenizer.cpp"
yy1280:
	yych = *++cur;
	if (yych == 'U') goto yy1340;
	if (yych == 'u') goto yy1340;
	goto yy79;
yy1281:
	yych = *++cur;
	if (yych == 'M') goto yy1341;
	if (yych == 'm') goto yy1341;
	goto yy79;
yy1282:
	++cur;
#line 333 "tokenizer.re"
	{ KW(END_SELECT) }
#line 8283 "tokenizer.cpp"
yy1284:
	yych = *++cur;
	if (yych == 'U') goto yy1343;
	if (yych == 'u') goto yy1343;
	goto yy79;
yy1285:
	yych = *++cur;
	if (yych == 'T') goto yy1344;
	if (yych == 't') goto yy1344;
	goto yy79;
yy1286:
	yych = *++cur;
	if (yych == 'T') goto yy1345;
	if (yych == 't') goto yy1345;
	goto yy55;
yy1287:
	yych = *++cur;
	if (yych == 'A') goto yy1346;
	if (yych == 'a') goto yy1346;
	goto yy79;
yy1288:
	yych = *++cur;
	if (yych == 'T') goto yy1347;
	if (yych == 't') goto yy1347;
	goto yy55;
yy1289:
	yych = *++cur;
	if (yych == 'L') goto yy1348;
	if (yych == 'l') goto yy1348;
	goto yy55;
yy1290:
	yych = *++cur;
	if (yych == 'N') goto yy1350;
	if (yych == 'n') goto yy1350;
	goto yy55;
yy1291:
	yych = *++cur;
	if (yych == 'C') goto yy1352;
	if (yych == 'c') goto yy1352;
	goto yy55;
yy1292:
	yych = *++cur;
	if (yych == 'R') goto yy1353;
	if (yych == 'r') goto yy1353;
	goto yy55;
yy1293:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 305 "tokenizer.re"
	{ KW(ENDPROGRAM) }
#line 8337 "tokenizer.cpp"
yy1295:
	yych = *++cur;
	if (yych == 'L') goto yy1354;
	if (yych == 'l') goto yy1354;
	goto yy55;
yy1296:
	yych = *++cur;
	if (yych == 'I') goto yy1355;
	if (yych == 'i') goto yy1355;
	goto yy55;
yy1297:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 380 "tokenizer.re"
	{ KW(ENUMERATOR) }
#line 8356 "tokenizer.cpp"
yy1299:
	yych = *++cur;
	if (yych == 'E') goto yy1356;
	if (yych == 'e') goto yy1356;
	goto yy55;
yy1300:
	yyaccept = 16;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy1215;
yy1301:
#line 659 "tokenizer.re"
	{ continue; }
#line 8369 "tokenizer.cpp"
yy1302:
	yyaccept = 17;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy1217;
yy1303:
#line 660 "tokenizer.re"
	{ continue; }
#line 8377 "tokenizer.cpp"
yy1304:
	yych = *++cur;
	if (yych == '"') goto yy1215;
	if (yych == '\'') goto yy1217;
	goto yy79;
yy1305:
	yych = *++cur;
	if (yych <= '9') {
		if (yych <= '&') {
			if (yych == '"') goto yy1215;
			goto yy79;
		} else {
			if (yych <= '\'') goto yy1217;
			if (yych <= '/') goto yy79;
			goto yy1221;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy1221;
			goto yy79;
		} else {
			if (yych <= '_') goto yy1305;
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1221;
			goto yy79;
		}
	}
yy1307:
	yych = *++cur;
	if (yych == 'E') goto yy1358;
	if (yych == 'e') goto yy1358;
	goto yy55;
yy1308:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 430 "tokenizer.re"
	{ KW(LOCAL_INIT) }
#line 8419 "tokenizer.cpp"
yy1310:
	yych = *++cur;
	if (yych == 'S') goto yy1360;
	if (yych == 's') goto yy1360;
	goto yy55;
yy1311:
	yych = *++cur;
	if (yych == 'D') goto yy1361;
	if (yych == 'd') goto yy1361;
	goto yy55;
yy1312:
	yych = *++cur;
	if (yych == 'I') goto yy1362;
	if (yych == 'i') goto yy1362;
	goto yy55;
yy1313:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 472 "tokenizer.re"
	{ KW(SELECT_CASE) }
#line 8443 "tokenizer.cpp"
yy1315:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 473 "tokenizer.re"
	{ KW(SELECT_RANK) }
#line 8452 "tokenizer.cpp"
yy1317:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 474 "tokenizer.re"
	{ KW(SELECT_TYPE) }
#line 8461 "tokenizer.cpp"
yy1319:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 481 "tokenizer.re"
	{ KW(SUBROUTINE) }
#line 8470 "tokenizer.cpp"
yy1321:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 484 "tokenizer.re"
	{ KW(SYNC_IMAGES) }
#line 8479 "tokenizer.cpp"
yy1323:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 485 "tokenizer.re"
	{ KW(SYNC_MEMORY) }
#line 8488 "tokenizer.cpp"
yy1325:
	yych = *++cur;
	if (yych == 'R') goto yy1363;
	if (yych == 'r') goto yy1363;
	goto yy55;
yy1326:
	yych = *++cur;
	if (yych == 'D') goto yy1365;
	if (yych == 'd') goto yy1365;
	goto yy55;
yy1327:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 255 "tokenizer.re"
	{ KW(ALLOCATABLE) }
#line 8507 "tokenizer.cpp"
yy1329:
	yych = *++cur;
	if (yych == 'S') goto yy1367;
	if (yych == 's') goto yy1367;
	goto yy55;
yy1330:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 271 "tokenizer.re"
	{ KW(CODIMENSION) }
#line 8521 "tokenizer.cpp"
yy1332:
	yych = *++cur;
	if (yych == 'E') goto yy1369;
	if (yych == 'e') goto yy1369;
	goto yy55;
yy1333:
	yych = *++cur;
	if (yych == 'S') goto yy1370;
	if (yych == 's') goto yy1370;
	goto yy55;
yy1334:
	yych = *++cur;
	if (yych == 'T') goto yy1371;
	if (yych == 't') goto yy1371;
	goto yy79;
yy1335:
	yych = *++cur;
	if (yych == 'A') goto yy1372;
	if (yych == 'a') goto yy1372;
	goto yy79;
yy1336:
	yych = *++cur;
	if (yych == 'T') goto yy1373;
	if (yych == 't') goto yy1373;
	goto yy79;
yy1337:
	yych = *++cur;
	if (yych == 'L') goto yy1374;
	if (yych == 'l') goto yy1374;
	goto yy79;
yy1338:
	yych = *++cur;
	if (yych == 'N') goto yy1376;
	if (yych == 'n') goto yy1376;
	goto yy79;
yy1339:
	yych = *++cur;
	if (yych == 'C') goto yy1378;
	if (yych == 'c') goto yy1378;
	goto yy79;
yy1340:
	yych = *++cur;
	if (yych == 'R') goto yy1379;
	if (yych == 'r') goto yy1379;
	goto yy79;
yy1341:
	++cur;
#line 304 "tokenizer.re"
	{ KW(END_PROGRAM) }
#line 8571 "tokenizer.cpp"
yy1343:
	yych = *++cur;
	if (yych == 'L') goto yy1380;
	if (yych == 'l') goto yy1380;
	goto yy79;
yy1344:
	yych = *++cur;
	if (yych == 'I') goto yy1381;
	if (yych == 'i') goto yy1381;
	goto yy79;
yy1345:
	yych = *++cur;
	if (yych == 'E') goto yy1382;
	if (yych == 'e') goto yy1382;
	goto yy55;
yy1346:
	yych = *++cur;
	if (yych == 'T') goto yy1384;
	if (yych == 't') goto yy1384;
	goto yy79;
yy1347:
	yych = *++cur;
	if (yych == 'A') goto yy1385;
	if (yych == 'a') goto yy1385;
	goto yy55;
yy1348:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 340 "tokenizer.re"
	{ KW(ENDCRITICAL) }
#line 8605 "tokenizer.cpp"
yy1350:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 325 "tokenizer.re"
	{ KW(ENDFUNCTION) }
#line 8614 "tokenizer.cpp"
yy1352:
	yych = *++cur;
	if (yych == 'E') goto yy1387;
	if (yych == 'e') goto yy1387;
	goto yy55;
yy1353:
	yych = *++cur;
	if (yych == 'E') goto yy1389;
	if (yych == 'e') goto yy1389;
	goto yy55;
yy1354:
	yych = *++cur;
	if (yych == 'E') goto yy1391;
	if (yych == 'e') goto yy1391;
	goto yy55;
yy1355:
	yych = *++cur;
	if (yych == 'N') goto yy1393;
	if (yych == 'n') goto yy1393;
	goto yy55;
yy1356:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 381 "tokenizer.re"
	{ KW(EQUIVALENCE) }
#line 8643 "tokenizer.cpp"
yy1358:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 421 "tokenizer.re"
	{ KW(INSTANTIATE) }
#line 8652 "tokenizer.cpp"
yy1360:
	yych = *++cur;
	if (yych == 'I') goto yy1394;
	if (yych == 'i') goto yy1394;
	goto yy55;
yy1361:
	yych = *++cur;
	if (yych == 'A') goto yy1395;
	if (yych == 'a') goto yy1395;
	goto yy55;
yy1362:
	yych = *++cur;
	if (yych == 'V') goto yy1396;
	if (yych == 'v') goto yy1396;
	goto yy55;
yy1363:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 489 "tokenizer.re"
	{ KW(TEAM_NUMBER) }
#line 8676 "tokenizer.cpp"
yy1365:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 494 "tokenizer.re"
	{ KW(UNFORMATTED) }
#line 8685 "tokenizer.cpp"
yy1367:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 260 "tokenizer.re"
	{ KW(ASYNCHRONOUS) }
#line 8694 "tokenizer.cpp"
yy1369:
	yych = *++cur;
	if (yych == 'X') goto yy1397;
	if (yych == 'x') goto yy1397;
	goto yy55;
yy1370:
	yych = *++cur;
	if (yych == 'I') goto yy1399;
	if (yych == 'i') goto yy1399;
	goto yy55;
yy1371:
	yych = *++cur;
	if (yych == 'E') goto yy1400;
	if (yych == 'e') goto yy1400;
	goto yy79;
yy1372:
	yych = *++cur;
	if (yych == 'T') goto yy1402;
	if (yych == 't') goto yy1402;
	goto yy79;
yy1373:
	yych = *++cur;
	if (yych == 'A') goto yy1403;
	if (yych == 'a') goto yy1403;
	goto yy79;
yy1374:
	++cur;
#line 339 "tokenizer.re"
	{ KW(END_CRITICAL) }
#line 8724 "tokenizer.cpp"
yy1376:
	++cur;
#line 324 "tokenizer.re"
	{ KW(END_FUNCTION) }
#line 8729 "tokenizer.cpp"
yy1378:
	yych = *++cur;
	if (yych == 'E') goto yy1405;
	if (yych == 'e') goto yy1405;
	goto yy79;
yy1379:
	yych = *++cur;
	if (yych == 'E') goto yy1407;
	if (yych == 'e') goto yy1407;
	goto yy79;
yy1380:
	yych = *++cur;
	if (yych == 'E') goto yy1409;
	if (yych == 'e') goto yy1409;
	goto yy79;
yy1381:
	yych = *++cur;
	if (yych == 'N') goto yy1411;
	if (yych == 'n') goto yy1411;
	goto yy79;
yy1382:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 337 "tokenizer.re"
	{ KW(ENDASSOCIATE) }
#line 8758 "tokenizer.cpp"
yy1384:
	yych = *++cur;
	if (yych == 'A') goto yy1412;
	if (yych == 'a') goto yy1412;
	goto yy79;
yy1385:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 319 "tokenizer.re"
	{ KW(ENDBLOCKDATA) }
#line 8772 "tokenizer.cpp"
yy1387:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 352 "tokenizer.re"
	{ KW(ENDINTERFACE) }
#line 8781 "tokenizer.cpp"
yy1389:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 328 "tokenizer.re"
	{ KW(ENDPROCEDURE) }
#line 8790 "tokenizer.cpp"
yy1391:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 311 "tokenizer.re"
	{ KW(ENDSUBMODULE) }
#line 8799 "tokenizer.cpp"
yy1393:
	yych = *++cur;
	if (yych == 'E') goto yy1414;
	if (yych == 'e') goto yy1414;
	goto yy55;
yy1394:
	yych = *++cur;
	if (yych == 'C') goto yy1416;
	if (yych == 'c') goto yy1416;
	goto yy55;
yy1395:
	yych = *++cur;
	if (yych == 'B') goto yy1418;
	if (yych == 'b') goto yy1418;
	goto yy55;
yy1396:
	yych = *++cur;
	if (yych == 'E') goto yy1419;
	if (yych == 'e') goto yy1419;
	goto yy55;
yy1397:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 296 "tokenizer.re"
	{ KW(DOUBLE_COMPLEX) }
#line 8828 "tokenizer.cpp"
yy1399:
	yych = *++cur;
	if (yych == 'O') goto yy1421;
	if (yych == 'o') goto yy1421;
	goto yy55;
yy1400:
	++cur;
#line 336 "tokenizer.re"
	{ KW(END_ASSOCIATE) }
#line 8838 "tokenizer.cpp"
yy1402:
	yych = *++cur;
	if (yych == 'A') goto yy1422;
	if (yych == 'a') goto yy1422;
	goto yy79;
yy1403:
	++cur;
#line 318 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8848 "tokenizer.cpp"
yy1405:
	++cur;
#line 351 "tokenizer.re"
	{ KW(END_INTERFACE) }
#line 8853 "tokenizer.cpp"
yy1407:
	++cur;
#line 327 "tokenizer.re"
	{ KW(END_PROCEDURE) }
#line 8858 "tokenizer.cpp"
yy1409:
	++cur;
#line 310 "tokenizer.re"
	{ KW(END_SUBMODULE) }
#line 8863 "tokenizer.cpp"
yy1411:
	yych = *++cur;
	if (yych == 'E') goto yy1424;
	if (yych == 'e') goto yy1424;
	goto yy79;
yy1412:
	++cur;
#line 317 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8873 "tokenizer.cpp"
yy1414:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 322 "tokenizer.re"
	{ KW(ENDSUBROUTINE) }
#line 8882 "tokenizer.cpp"
yy1416:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 439 "tokenizer.re"
	{ KW(NON_INTRINSIC) }
#line 8891 "tokenizer.cpp"
yy1418:
	yych = *++cur;
	if (yych == 'L') goto yy1426;
	if (yych == 'l') goto yy1426;
	goto yy55;
yy1419:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 441 "tokenizer.re"
	{ KW(NON_RECURSIVE) }
#line 8905 "tokenizer.cpp"
yy1421:
	yych = *++cur;
	if (yych == 'N') goto yy1427;
	if (yych == 'n') goto yy1427;
	goto yy55;
yy1422:
	++cur;
#line 316 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8915 "tokenizer.cpp"
yy1424:
	++cur;
#line 321 "tokenizer.re"
	{ KW(END_SUBROUTINE) }
#line 8920 "tokenizer.cpp"
yy1426:
	yych = *++cur;
	if (yych == 'E') goto yy1429;
	if (yych == 'e') goto yy1429;
	goto yy55;
yy1427:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 295 "tokenizer.re"
	{ KW(DOUBLE_PRECISION) }
#line 8934 "tokenizer.cpp"
yy1429:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 440 "tokenizer.re"
	{ KW(NON_OVERRIDABLE) }
#line 8943 "tokenizer.cpp"
}
#line 667 "tokenizer.re"

    }
}

std::string token(unsigned char *tok, unsigned char* cur)
{
    return std::string((char *)tok, cur - tok);
}

void token_loc(Location &loc)
{
    loc.first = 1;
    loc.last = 1;
}

void lex_format(unsigned char *&cur, Location &loc,
        unsigned char *&start) {
    int num_paren = 0;
    for (;;) {
        unsigned char *tok = cur;
        unsigned char *mar;
        
#line 8968 "tokenizer.cpp"
{
	unsigned char yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm[] = {
		  0, 208, 208, 208, 208, 208, 208, 208, 
		208, 244, 176, 244, 208, 244, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		244, 208, 192, 208, 208, 208, 208,  80, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		216, 216, 216, 216, 216, 216, 216, 216, 
		216, 216, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
	};
	yych = *cur;
	if (yybm[0+yych] & 4) {
		goto yy1437;
	}
	switch (yych) {
	case 0x00:	goto yy1433;
	case '"':	goto yy1440;
	case '&':	goto yy1441;
	case '\'':	goto yy1442;
	case '(':	goto yy1443;
	case ')':	goto yy1445;
	case '*':	goto yy1447;
	case ',':	goto yy1448;
	case '/':
	case ':':	goto yy1450;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy1452;
	case 'A':
	case 'a':	goto yy1453;
	case 'B':
	case 'I':
	case 'O':
	case 'Z':
	case 'b':
	case 'i':
	case 'o':
	case 'z':	goto yy1456;
	case 'D':
	case 'F':
	case 'd':
	case 'f':	goto yy1457;
	case 'E':
	case 'e':	goto yy1458;
	case 'G':
	case 'g':	goto yy1459;
	case 'L':
	case 'l':	goto yy1460;
	case 'P':
	case 'p':	goto yy1461;
	case 'T':
	case 't':	goto yy1462;
	case 'X':
	case 'x':	goto yy1463;
	default:	goto yy1435;
	}
yy1433:
	++cur;
#line 760 "tokenizer.re"
	{
                token_loc(loc);
                std::string t = token(tok, cur);
                throw LFortran::parser_local::TokenizerError(
                    "End of file not expected in `format` statement '" + t + "'", loc);
            }
#line 9068 "tokenizer.cpp"
yy1435:
	++cur;
yy1436:
#line 726 "tokenizer.re"
	{
                token_loc(loc);
                std::string t = token(tok, cur);
                throw LFortran::parser_local::TokenizerError("Token '" + t
                    + "' is not recognized in `format` statement", loc);
            }
#line 9079 "tokenizer.cpp"
yy1437:
	yych = *++cur;
	if (yybm[0+yych] & 4) {
		goto yy1437;
	}
#line 766 "tokenizer.re"
	{ continue; }
#line 9087 "tokenizer.cpp"
yy1440:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy1436;
	goto yy1465;
yy1441:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= 0x08) goto yy1436;
		if (yych <= '\v') goto yy1470;
		goto yy1436;
	} else {
		if (yych <= '\r') goto yy1470;
		if (yych <= 0x1F) goto yy1436;
		if (yych <= '!') goto yy1470;
		goto yy1436;
	}
yy1442:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy1436;
	goto yy1477;
yy1443:
	++cur;
#line 732 "tokenizer.re"
	{
                if (num_paren == 0) {
                    num_paren++;
                    start = cur;
                    continue;
                } else {
                    cur--;
                    unsigned char *tmp;
                    lex_format(cur, loc, tmp);
                    continue;
                }
            }
#line 9126 "tokenizer.cpp"
yy1445:
	++cur;
#line 756 "tokenizer.re"
	{
                LFORTRAN_ASSERT(num_paren == 1);
                return;
            }
#line 9134 "tokenizer.cpp"
yy1447:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1436;
			goto yy1480;
		} else {
			if (yych == '\v') goto yy1480;
			goto yy1436;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1480;
			if (yych <= 0x1F) goto yy1436;
			goto yy1480;
		} else {
			if (yych == '(') goto yy1482;
			goto yy1436;
		}
	}
yy1448:
	++cur;
#line 767 "tokenizer.re"
	{ continue; }
#line 9160 "tokenizer.cpp"
yy1450:
	++cur;
yy1451:
#line 772 "tokenizer.re"
	{ continue; }
#line 9166 "tokenizer.cpp"
yy1452:
	yyaccept = 0;
	yych = *(mar = ++cur);
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':
	case '(':
	case '/':
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'A':
	case 'B':
	case 'D':
	case 'E':
	case 'F':
	case 'G':
	case 'I':
	case 'L':
	case 'O':
	case 'P':
	case 'X':
	case 'Z':
	case 'a':
	case 'b':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'i':
	case 'l':
	case 'o':
	case 'p':
	case 'x':
	case 'z':	goto yy1489;
	default:	goto yy1436;
	}
yy1453:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1453;
	}
yy1455:
#line 771 "tokenizer.re"
	{ continue; }
#line 9221 "tokenizer.cpp"
yy1456:
	yych = *++cur;
	if (yych <= '/') goto yy1436;
	if (yych <= '9') goto yy1495;
	goto yy1436;
yy1457:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '/') goto yy1436;
	if (yych <= '9') goto yy1497;
	goto yy1436;
yy1458:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 'W') {
		if (yych <= 'M') {
			if (yych <= '/') goto yy1436;
			if (yych <= '9') goto yy1499;
			goto yy1436;
		} else {
			if (yych <= 'N') goto yy1501;
			if (yych == 'S') goto yy1501;
			goto yy1436;
		}
	} else {
		if (yych <= 'r') {
			if (yych <= 'X') goto yy1501;
			if (yych == 'n') goto yy1501;
			goto yy1436;
		} else {
			if (yych <= 's') goto yy1501;
			if (yych == 'x') goto yy1501;
			goto yy1436;
		}
	}
yy1459:
	yych = *++cur;
	if (yych <= '/') goto yy1436;
	if (yych <= '9') goto yy1502;
	goto yy1436;
yy1460:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1453;
	}
	goto yy1436;
yy1461:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= 'D') goto yy1455;
	if (yych <= 'F') goto yy1491;
	if (yych <= 'd') goto yy1455;
	if (yych <= 'f') goto yy1491;
	goto yy1455;
yy1462:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 'Q') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1436;
			goto yy1504;
		} else {
			if (yych == 'L') goto yy1506;
			goto yy1436;
		}
	} else {
		if (yych <= 'l') {
			if (yych <= 'R') goto yy1506;
			if (yych <= 'k') goto yy1436;
			goto yy1506;
		} else {
			if (yych == 'r') goto yy1506;
			goto yy1436;
		}
	}
yy1463:
	++cur;
	goto yy1455;
yy1464:
	yych = *++cur;
yy1465:
	if (yybm[0+yych] & 16) {
		goto yy1464;
	}
	if (yych >= 0x01) goto yy1467;
yy1466:
	cur = mar;
	if (yyaccept <= 2) {
		if (yyaccept <= 1) {
			if (yyaccept == 0) {
				goto yy1436;
			} else {
				goto yy1455;
			}
		} else {
			goto yy1468;
		}
	} else {
		if (yyaccept == 3) {
			goto yy1473;
		} else {
			goto yy1479;
		}
	}
yy1467:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy1464;
yy1468:
#line 769 "tokenizer.re"
	{ continue; }
#line 9333 "tokenizer.cpp"
yy1469:
	yych = *++cur;
yy1470:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1466;
			goto yy1469;
		} else {
			if (yych <= '\n') goto yy1471;
			if (yych <= '\v') goto yy1469;
			goto yy1466;
		}
	} else {
		if (yych <= 0x1F) {
			if (yych <= '\r') goto yy1469;
			goto yy1466;
		} else {
			if (yych <= ' ') goto yy1469;
			if (yych <= '!') goto yy1474;
			goto yy1466;
		}
	}
yy1471:
	yyaccept = 3;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 32) {
		goto yy1471;
	}
	if (yych <= 0x1F) goto yy1473;
	if (yych <= '!') goto yy1474;
	if (yych == '&') goto yy1507;
yy1473:
#line 768 "tokenizer.re"
	{ continue; }
#line 9368 "tokenizer.cpp"
yy1474:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy1474;
	}
	if (yych <= 0x00) goto yy1466;
	goto yy1471;
yy1476:
	yych = *++cur;
yy1477:
	if (yybm[0+yych] & 128) {
		goto yy1476;
	}
	if (yych <= 0x00) goto yy1466;
	yyaccept = 4;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy1476;
yy1479:
#line 770 "tokenizer.re"
	{ continue; }
#line 9389 "tokenizer.cpp"
yy1480:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1466;
			goto yy1480;
		} else {
			if (yych == '\v') goto yy1480;
			goto yy1466;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1480;
			if (yych <= 0x1F) goto yy1466;
			goto yy1480;
		} else {
			if (yych != '(') goto yy1466;
		}
	}
yy1482:
	++cur;
#line 750 "tokenizer.re"
	{
                cur--;
                unsigned char *tmp;
                lex_format(cur, loc, tmp);
                continue;
            }
#line 9418 "tokenizer.cpp"
yy1484:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1466;
			goto yy1484;
		} else {
			if (yych == '\v') goto yy1484;
			goto yy1466;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1484;
			if (yych <= 0x1F) goto yy1466;
			goto yy1484;
		} else {
			if (yych != '(') goto yy1466;
		}
	}
yy1486:
	++cur;
#line 744 "tokenizer.re"
	{
                cur--;
                unsigned char *tmp;
                lex_format(cur, loc, tmp);
                continue;
            }
#line 9447 "tokenizer.cpp"
yy1488:
	yych = *++cur;
yy1489:
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':	goto yy1484;
	case '(':	goto yy1486;
	case '/':	goto yy1450;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy1488;
	case 'A':
	case 'a':	goto yy1453;
	case 'B':
	case 'I':
	case 'O':
	case 'Z':
	case 'b':
	case 'i':
	case 'o':
	case 'z':	goto yy1490;
	case 'D':
	case 'F':
	case 'd':
	case 'f':	goto yy1491;
	case 'E':
	case 'e':	goto yy1492;
	case 'G':
	case 'g':	goto yy1493;
	case 'L':
	case 'l':	goto yy1494;
	case 'P':
	case 'p':	goto yy1461;
	case 'X':
	case 'x':	goto yy1463;
	default:	goto yy1466;
	}
yy1490:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1495;
	goto yy1466;
yy1491:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1497;
	goto yy1466;
yy1492:
	yych = *++cur;
	if (yych <= 'W') {
		if (yych <= 'M') {
			if (yych <= '/') goto yy1466;
			if (yych <= '9') goto yy1499;
			goto yy1466;
		} else {
			if (yych <= 'N') goto yy1501;
			if (yych == 'S') goto yy1501;
			goto yy1466;
		}
	} else {
		if (yych <= 'r') {
			if (yych <= 'X') goto yy1501;
			if (yych == 'n') goto yy1501;
			goto yy1466;
		} else {
			if (yych <= 's') goto yy1501;
			if (yych == 'x') goto yy1501;
			goto yy1466;
		}
	}
yy1493:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1502;
	goto yy1466;
yy1494:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1453;
	}
	goto yy1466;
yy1495:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych == '.') goto yy1494;
	if (yych <= '/') goto yy1455;
	if (yych <= '9') goto yy1495;
	goto yy1455;
yy1497:
	yych = *++cur;
	if (yych == '.') goto yy1494;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1497;
	goto yy1466;
yy1499:
	yych = *++cur;
	if (yych == '.') goto yy1508;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1499;
	goto yy1466;
yy1501:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1499;
	goto yy1466;
yy1502:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych == '.') goto yy1508;
	if (yych <= '/') goto yy1455;
	if (yych <= '9') goto yy1502;
	goto yy1455;
yy1504:
	yych = *++cur;
	if (yych <= '/') goto yy1451;
	if (yych <= '9') goto yy1504;
	goto yy1451;
yy1506:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych <= '9') goto yy1504;
	goto yy1466;
yy1507:
	++cur;
	goto yy1473;
yy1508:
	yych = *++cur;
	if (yych <= '/') goto yy1466;
	if (yych >= ':') goto yy1466;
yy1509:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= 'D') {
		if (yych <= '/') goto yy1455;
		if (yych <= '9') goto yy1509;
		goto yy1455;
	} else {
		if (yych <= 'E') goto yy1494;
		if (yych == 'e') goto yy1494;
		goto yy1455;
	}
}
#line 773 "tokenizer.re"

    }
}

} // namespace LFortran
